#!/ebi/extserv/bin/perl/bin/perl

use strict;
use Getopt::Long;

# assignment.pl
# http://supfam.org
# Copyright (c) 2001 MRC and Julian Gough; see http://supfam.org/SUPERFAMILY/license.html
#This program is supposed to: Create SUPERFAMILY assignments from HMM scores on the model library
#Julian Gough 28.11.00 modified 13.9.01 and again 25.11.01
#
#Modified by Emmanuel Quevillon <tuco@ebi.ac.uk> on 5th April 2005
#Introduced usage of Getopt::Long module to treat command line options easier.
#Tried to improve speed of the script.

#ASIGN-VARIABLES-------------------------------------
#my $usage="assignment.pl [-arg argument ..] -db <database> -dir <scores directory (file for hmmpfam)> -out <outputfile> -p <prefix for score files (optioonal)>\n[-e Evaluecutoff]    -e 0.02 is default\n[-o \%overlap]    -o 35 is defualt 35%\n[-v y/n] verbose (print information)? y/n\n[-core residues] min. no. of residues in core of protein, default=15\n[-f sam/hmmer/hmmpfam] scores in SAM, HMMER or HMMER/hmmpfam (single file) format, default=sam\n[-a y/n]  are alignments available? y/n\n\nPlease specify a directory containing all of the HMM score results files, and no others. The filenames in the directory are assumed to be model names. SAM scores are assumed to be in 'mstat' file format. SAM alignments are assumed to be in 'mult' format. HMMER scores and alignments are assumed to be in individual files named 'res'\n";
my $usage = "assignment.pl <options> -db <database> -dir <score directory> -out <outputfile> -p <prefix>\n\n";
$usage .= "Options:\n--------\n\n";
$usage .= "\t-e      <number> Evalue cutoff (default is 0.02)\n";
$usage .= "\t-o      <number> Overlap percentage (default is 35%)\n";
$usage .= "\t-v      Verbose mode, prints informations\n";
$usage .= "\t-core   <number> Core residues, min number of residues in core protein (default is 15)\n";
$usage .= "\t-f      <sam|hmmer|hmmpfam> Scores in SAM, HMMER(.hmm) or HMMER (hmmpfam single file) (default is sam)\n";
$usage .= "\t-a      <y|n> Are alignments available? (default is y)\n";
$usage .= "\t-h      Displays this help and exit\n\n";
$usage .= "Mandatory:\n----------\n\n";
$usage .= "\t-db     <database> Query file (mandatory)\n";
$usage .= "\t-dir    <directory> Scores directory (file for hmmpfam) (mandatory)\n";
$usage .= "\t-out    <outputfile> Output file where to write results (mandatory)\n";
$usage .= "\t-p      <prefix> Prefix for score files (optional)\n\n";
$usage .= "Please specify a directory containing all of the HMM score results files, and no others.\n";
$usage .= "The filenames in the directory are assumed to be model names. SAM scores are assumed to be model names.\n";
$usage .= "SAM scores are assumed to be in 'mstat' file format. SAM alignments are assumed to be in 'mult' format\n";
$usage .= "SAM alignments are assumed to be in 'mult' format. HMMER scores and alignments are assumed to be in individual\n";
$usage .= "files named 'res'\n\n";

my $name;
my @hits;
my $hit;
my $i;
my @regionfrom;
my @regionto;
my @scores;
my @models;
my @order;
my @assignfrom=();
my @assignto=();
my @assignscore=();
my @assignmod=();
my $j;
my $conflict;
my $k;
my $seq;
my $count;
my $l;
my @match1;
my @match2;
my $leftover;
my %output;
my $seq2;
my @assignseq;
my $comments;
my $alloverlap;
my %dbseqs;
my $format;
my $suffix;

#----------------------------------------------------


#ARGS------------------------------------------------

#Command line options
my ($percentsame, $cutoff, $alignsyes, $coresize, $verbose, $format, $database, $mstatdir, $outputfile, $prefix, $help);

GetOptions(
	   'o:i'    => \$percentsame,
	   'e:i'    => \$cutoff,
	   'a:s'    => \$alignsyes,
	   'core:i' => \$coresize,
	   'v'      => \$verbose,
	   'f:s'    => \$format,
	   'db=s'   => \$database,
	   'dir=s'  => \$mstatdir,
	   'out=s'  => \$outputfile,
	   'p:s'    => \$prefix,
           'h'      => \$help
	   );

if($help){
    print $usage;
    exit 0;
}

$percentsame = '35' unless($percentsame);
$cutoff = 0.02 unless($cutoff);
$coresize = 15 unless($coresize);
$prefix = '' unless($prefix);

if($alignsyes){
    die "ERROR: Alignments option value can only be 'y' or 'n', not '$alignsyes'." if($alignsyes ne 'y' && $alignsyes ne 'n');
}else{
    $alignsyes = 'y';
}

if($format){
    die "ERROR: Format option can only be 'sam', 'hmmer' or 'hmmpfam', not '$format'." if($format ne 'sam' && $format ne 'hmmer' && $format ne 'hmmpfam');
}else{
    $format = 'sam';
}

die "ERROR: Missing something (database or directory or output file)\n\nUsage : $usage\n\n" unless($database && $mstatdir && $outputfile);
$outputfile .= ".ass";

print "\nRunning assignment.pl....\n\nScores directory: $mstatdir\noutputfile: $outputfile\nPercent overlap required to be same domain: $percentsame\nEvalue cut-off:  $cutoff\ndatabase:  $database \nMinimum coresize:  $coresize residues\nScore files are in $format format.\nAlignments are available?  y/n:   $alignsyes\n\n" if($verbose);
#----------------------------------------------------

#SUPPORTED-OPTIONS-----------------------------------
unless ($alignsyes eq 'y'){
  die "ERROR: Currently only formats including alignments are supported. Other formats coming soon (25.11.01)\n\n";
}
#----------------------------------------------------

#MAIN-LOOP-------------------------------------------

if ($format eq 'sam'){
  &AllHitsFromMstatDir($mstatdir,$cutoff,$prefix);
}
elsif ($format eq 'hmmer'){
  &AllHitsFromHmmerDir($mstatdir,$cutoff,$prefix);
}
elsif ($format eq 'hmmpfam'){
  &AllHitsFromHmmpfam($mstatdir,$cutoff,$prefix);
}
else{
  die "ERROR: Format must be specified as either 'sam', 'hmmpfam'  or 'hmmer'\n";
}

%dbseqs=&ReadAllSeqFasta($database);

unless(open OUT,(">$outputfile")){
    close OUT;
    die "ERROR: Could not open [$outputfile] : $!";
}

$percentsame=($percentsame/100);
unless(open DATA,("$database")){
    close DATA;
    die "ERROR: Could not open [$database] : $!";
}

while (<DATA>){
  if (/^>/){
    if (/^>(\S+)$/){
      $name=$1;
      $comments='';
    }
    elsif(/^>(\S+)\,\s(.*)$/){
      $name=$1;
      $comments=$2;
    }
    elsif(/^>(\S+)\s(.*)$/){
      $name=$1;
      $comments=$2;
    }
    print ".";
    if (exists($output{$name})){
      $seq=$dbseqs{$name};
      $seq =~ s/\n//g;
      @hits=split /\:/,$output{$name};
      $i=0;
      @models=();
      @scores=();
      @regionfrom=();
      @regionto=();
      foreach $hit (@hits){
	$hit =~ /^(\S+)\,(\S+)\,(\S+)\,(\S+)$/;
	$models[$i]=$1;
	$regionfrom[$i]=$2;
	$regionto[$i]=$3;
	$scores[$i]=$4;
	$models[$i] =~ s/^\w\w_//;
	$i++;
      }
      @order=&OrderArray(@scores);
      @assignfrom=();
      @assignto=();
      @assignscore=();
      @assignmod=();
      @assignseq=();
      $leftover=length($seq);
      foreach $i (@order){
	$conflict=0;
	if (scalar(@assignmod) == 0){
	  $assignfrom[0]=$regionfrom[$i];
	  $assignto[0]=$regionto[$i];
	  $assignscore[0]=$scores[$i];
	  $assignmod[0]=$models[$i];
	  if ($format eq 'sam' and $alignsyes eq 'y'){
	    $assignseq[0]=GetMultSeq("$mstatdir/$prefix$assignmod[0].mult","$name\_$regionfrom[$i]:$regionto[$i]");
	  }
	  elsif ($format eq 'hmmer' and $alignsyes eq 'y'){
	    $assignseq[0]=GetHmmerSeq("$mstatdir/$prefix$assignmod[0].res","$name\_$regionfrom[$i]:$regionto[$i]");
	  }
	  elsif ($format eq 'hmmpfam' and $alignsyes eq 'y'){
	    $assignseq[0]=GetHmmpfamSeq("$mstatdir/$prefix.res","$name\_$regionfrom[$i]:$regionto[$i]",$assignmod[0]);
	  }
	  else{
	    die "ERROR: Support currently requires alignments\n";
	  }
	  $leftover=$leftover-($assignseq[0] =~ tr/[A-Z]//);
	  next;
	}
	else{
	  $alloverlap=0;
	  if ($format eq 'sam' and $alignsyes eq 'y'){
	    $seq2=GetMultSeq("$mstatdir/$prefix$models[$i].mult","$name\_$regionfrom[$i]:$regionto[$i]");
	  }
	  elsif($format eq 'hmmer' and $alignsyes eq 'y'){
	    $seq2=GetHmmerSeq("$mstatdir/$prefix$models[$i].res","$name\_$regionfrom[$i]:$regionto[$i]");
	  }
	  elsif($format eq 'hmmpfam' and $alignsyes eq 'y'){
	    $seq2=GetHmmpfamSeq("$mstatdir/$prefix.res","$name\_$regionfrom[$i]:$regionto[$i]",$models[$i]);
	  }
	  
	  $count=0;
	  @match2=split //,$seq2;
	  foreach $l (0 .. scalar(@match2)-1){
	    if (uc($match2[$l]) eq $match2[$l]){
	      $count++;
	    }
	  }
	  foreach $k (0 .. scalar(@assignmod)-1){
	    if ($assignto[$k] < $regionfrom[$i] or $assignfrom[$k] > $regionto[$i]){
	    }
	    else{
	      @match1=split //,$assignseq[$k];
	      foreach $l (0 .. scalar(@match1)-1){
		if (uc($match2[$l]) eq $match2[$l] and uc($match1[$l]) eq $match1[$l]){
		  $alloverlap++;
		}
	      }
	      if ($alloverlap/$count > $percentsame){
		$conflict=1;
		last;
	      }
	    }
	  }
	}
	if ($alloverlap/$count > $percentsame){
	  $conflict=1;
	}
	if ($conflict == 0){
	  push @assignfrom,$regionfrom[$i];
	  push @assignto,$regionto[$i];
	  push @assignscore,$scores[$i];
	  push @assignmod,$models[$i];
	  push @assignseq,$seq2;
	  $leftover=$leftover-($seq2 =~ tr/[A-Z]//)+$alloverlap;
	}
	if ($leftover < $coresize){
	  last;
	}
      }
      for $j (0 .. scalar(@assignmod)-1){
	print OUT "$name\t$assignmod[$j]\t$assignfrom[$j]\t$assignto[$j]\t$assignscore[$j]\t$assignseq[$j]\n";
      }
    }
  }
}
close DATA;
print "......................Finished assigning\n";
close OUT;
#----------------------------------------------------


#SUB-ROUTINE-----------------------------------------------
sub AllHitsFromMstatDir{
  #  This reads in a fasta name and a directory of mstat files
#  and returns an array of model,region,score
  my $mstatdir=$_[0];
  my $cutoff=$_[1];
  my $pre=$_[2];
  my $model;
  
  unless(open HFMDDIR,("ls $mstatdir |")){
      close HFMDDIR;
      die "ERROR: Could not open [$mstatdir] : $!";
  }

  while (<HFMDDIR>){
    if (/^$pre(\S+)\.mstat$/){
      $model=$1;
      unless(open(HITSFROMMSTAT, "$mstatdir/$pre$model.mstat")){
	  close HFMDDIR;
	  close HITSFROMMSTAT;
	  die "ERROR: Could not open [$mstatdir/$pre$model.mstat] : $!";
      }

      while(<HITSFROMMSTAT>){
	unless (/^\%/){
	  if (/^(\S+)_(\d+):(\d+)\s+\S+\s+\S+\s+\S+\s+(-?\d\.\de\S\d\d\d?)\s.+$/ or /^(\S+)_(\d+):(\d+)\s+\S+\s+\S+\s+\S+\s+(-?\d\.\de\S\d\d\d?)$/){
	    if($4 <= $cutoff){
	      if (exists($output{$1})){
		$output{$1}=$output{$1}.":$model,$2,$3,$4";
	      }
	      else{
		$output{$1}="$model,$2,$3,$4";
	      }
	    }
	  }
	}
      }
      close HITSFROMMSTAT;
    }
  }
  close HFMDDIR;
  return;
}
#----------------------------------------------------------

#SUB-ROUTINE-----------------------------------------------
sub AllHitsFromHmmerDir{
  #  This reads in a fasta name and a directory of mstat files
  #  and returns an array of model,region,score
  my $mstatdir=$_[0];
  my $cutoff=$_[1];
  my $pre=$_[2];
  my $model;
  my $parsedflag=0;
  
  unless(open HFMDDIR,("ls $mstatdir |")){
      close HFMDDIR;
      die "ERROR: Could not open [$mstatdir] : $!";
  }

  while (<HFMDDIR>){
    if (/^$pre(\S+)\.(\w+)$/){
      $model=$1;
      $suffix=$2;
      unless(open(HITSFROMMSTAT, "$mstatdir/$pre$model.$2")){
	  close HFMDDIR;
	  close HTISFROMMSTAT;
	  die "ERROR: Could not open [$mstatdir/$pre$model.$2] : $!";
      }
      while(<HITSFROMMSTAT>){
	if (/Parsed\sfor\sdomains:/){
	  $parsedflag=1;
	}
	elsif ($parsedflag == 1 and /^\s*$/){
	  $parsedflag=0;
	  last;
	}
	elsif ($parsedflag == 1){
	  if (/^(\S+)\s+\d+\/\d+\s+(\d+)\s+(\d+)\s.*\s(\S+)\s*$/){
	    if($4 <= $cutoff){
	      if (exists($output{$1})){
		$output{$1}=$output{$1}.":$model,$2,$3,$4";
	      }
	      else{
		$output{$1}="$model,$2,$3,$4";
	      }
	    }
	  }
	}
      }
      close HITSFROMMSTAT;
    }
  }
  close HFMDDIR;
  return;
}
#----------------------------------------------------------

#SUB-ROUTINE-----------------------------------------------
sub AllHitsFromHmmpfam{
  my $mstatdir=$_[0];
  my $cutoff=$_[1];
  my $pre=$_[2];
  my $model;
  my $seqid;
  my $start;
  my $end;
  my $E;
  my $parsedflag=0;
  
  unless(open(HITSFROMMSTAT, "$mstatdir/$pre.res")){
      close HITSFROMMSTAT;
      die "ERROR: Could not open [$mstatdir/$pre.res] : $!";
  }

  while(<HITSFROMMSTAT>){
    if(/^Query\ssequence\:\s+(\S+)/){
      $seqid=$1;
    }
    elsif (/Parsed\sfor\sdomains:/){
      $parsedflag=1;
    }
    elsif ($parsedflag == 1 and /^\s*$/){
      $parsedflag=0;
    }
    elsif ($parsedflag == 1){
      if (/^(\d+)\s+\d+\s+\d+\/\d+\s+(\d+)\s+(\d+)\s.*\s(\S+)\s*$/){
	$model = $1;
	$start = $2;
	$end   = $3;
	$E     = $4;
	
	if($E <= $cutoff){
	  if (exists($output{$seqid})){
	    $output{$seqid}=$output{$seqid}.":$model,$start,$end,$E";
	  }
	  else{
	    $output{$seqid}="$model,$start,$end,$E";
	  }
	}
      }
    }
  }
  close HITSFROMMSTAT;
  return;
}
#----------------------------------------------------------

#SUB-ROUTINE-----------------------------------------------
#This should read all of the sequences in a fasta file into
#an array and return that array ready for a hash
sub ReadAllSeqFasta{
  my $filename;
  my @sequences;
  my $flag=1;
  my $seq;
  my $name;
  
  $filename=$_[0];
  unless(open(GETSEQREGS2, "$filename")){
      close GETSEQREGS2;
      die "ERROR: Could not open [$filename] : $!";
  }

  while(<GETSEQREGS2>){
    if (/^>(\S+)/){
      unless ($flag == 1){
	if (exists($output{$name})){
	  push @sequences,lc $seq;
	}
      }
      $flag=0;
      $name=$1;
      if (exists($output{$name})){
	push @sequences,$name;
      }
      $seq='';
    }
    else{
      $seq=$seq.$_;
      chomp $seq;
    }
  }
  if (exists($output{$name})){
    push @sequences,lc $seq;
  }
  close GETSEQREGS2;
  return (@sequences);
}
#----------------------------------------------------------

#get the sequence from mult file
sub GetMultSeq{
  my $multfile=$_[0];
  my $seqid=$_[1];
  my $flag=0;
  my $seq='';
  my $line;
  
  unless(open MULT,("$multfile")){
      close MULT;
      die "ERROR: Could not open [$multfile] : $!";
  }

  while (<MULT>){
    if (/^>(\S+)/){
      if ($1 eq $seqid){
	$flag=1;
      }
      elsif($flag == 1){
	$flag=0;
	last;
      }
    }
    elsif ($flag == 1){
      $line=$_;
      chomp $line;
      $seq=$seq.$line;
    }
  }
  close MULT;
  $seq =~ s/\.//g;
  $seq =~ s/\-//g;
  return ($seq);
}

#get the sequence from mult file
sub GetHmmerSeq{
  my $multfile=$_[0];
  my $seqid=$_[1];
  my $flag=0;
  my $seq='';
  my $start;
  my $stop;
  my $begin='';
  my $end='';
  my @tempseq;
  my $seqidt;
  my @temp;
  
  if ($seqid =~ /(\S+)\_(\d+):(\d+)/){
    $seqid=$1;
    $start=$2;
    $stop=$3;
  }
  else {
    die "ERROR:  failed to parse the sequence ID and regions";
  }
  if (length($seqid) > 10){
    @temp=split //,$seqid;
    $seqidt=join '',@temp[0 .. 9];
  }
  else{
    $seqidt=$seqid;
  }
  
  unless(open MULT,("$multfile")){
      close MULT;
      die "ERROR: Could not open [$multfile] : $!";
  }
  while (<MULT>){
    if (/^$seqid:\sdomain\s+\d+\sof\s+\d+,\s+from\s+$start\s+to\s+$stop\:/){
      $flag=1;
    }
    elsif($flag != 0 and /^\S+:\sdomain\s+\d+\sof\s+\d+,\s+from\s+\d+\s+to\s+\d+\:/){
      last;
    }
    elsif ($flag == 1 and (/^\s+\*->(\w+)/ or /^\s+([\w\.]+)/ or /^\s+(\w+)<-\*/)){
      $flag=2;
    }
    elsif ($flag == 2 and /^\s+$seqidt\S*\s+([-\d]+)\s+(\S+)\s+([-\d]+)/){
      unless ($3 eq '-'){
	$end=$3;
      }
      if ($begin eq '' and $1 ne '-'){
	$begin=$1;
      }
      $flag=3;
      $seq=$seq.$2;
      $flag=1;
    }
  }
  close MULT;
  $seq =~ s/\.//g;
  $seq =~ s/\-//g;
  @tempseq=split //, $dbseqs{$seqid};
  until ($begin <= 1){
    $seq=$tempseq[$begin-2].$seq;
    $begin=$begin-1;
  }
  until ($end >= scalar(@tempseq)){
    $seq=$seq.$tempseq[$end];
    $end=$end+1;
  }
  return ($seq);
}


sub GetHmmpfamSeq
  {
    my $multfile=$_[0];
    my $seqid=$_[1];
    my $model=$_[2];
    
    my $flag=0;
    my $seq='';
    my $start;
    my $stop;
    my $begin='';
    my $end='';
    my @tempseq;
    
    if ($seqid =~ /(\S+)\_(\d+):(\d+)/){
      $seqid=$1;
      $start=$2;
      $stop=$3;
    }
    else {
      die "ERROR:  failed to parse the sequence ID and regions";
    }

    # vsi@ebi.ac.uk: HMMPfam result has just 10 characters of the
    # sequence ID in the alignments. Also if the ID contains any
    # special characters, it will break the regular expression, so
    # quote all metacharacters.
    my $shortid = substr($seqid, 0, 10);
    $shortid = quotemeta $shortid;
    
    unless(open MULT,("$multfile")){
	close MULT;
	die "ERROR: Could not open [$multfile] : $!";
    }

    while (<MULT>){
      if (/^Query\s+sequence:\s+$seqid/)
	{
	  $flag=1;
	}
      elsif ($flag == 1 and /^$model\s+\d+:\sdomain\s+\d+\sof\s+\d+,\s+from\s+$start\s+to\s+$stop\:/){
	$flag=2;
      }
      elsif(($flag==2 or $flag==3) and 
	    (/^\d+\s+\d+:\sdomain\s+\d+\sof\s+\d+,\s+from\s+\d+\s+to\s+\d+\:/ or /^\/\//)){
	last;
      }
      elsif ($flag == 2 and (/^\s+\*->([\w\.]+)/ or /^\s+([\w\.]+)/ or /^\s+([\w\.]+)<-\*/)){
	$flag=3;
      }
      elsif ($flag == 3 and /^\s+$shortid\s+([-\d]+)\s+(\S+)\s+([-\d]+)/){
	unless ($3 eq '-'){
	  $end=$3;
	}
	if ($begin eq ''){
	  $begin=$1;
	  if($begin eq '-'){
	    $begin=0;
	  }
	}
	$seq=$seq.$2;
	$flag=2;
      }
    }
    close MULT;
    
    $seq =~ s/\.//g;
    $seq =~ s/\-//g;
    @tempseq=split //, $dbseqs{$seqid};
    until ($begin <= 1){
      $seq=$tempseq[$begin-2].$seq;
      $begin=$begin-1;
    }
    until ($end >= scalar(@tempseq)-1){
      $seq=$seq.$tempseq[$end-1];
      $end=$end+1;
    }
    
    return ($seq);
  }

#SUB-ROUTINE-----------------------------------------------
sub Args{
  # This reads in ARGV and sets all -n  type options and sets 
  # the global hash array %args
  my $arguement;
  my $type="rumplestiltskin";
  my $i=0;
  my @args;
  
  foreach $arguement (@ARGV){
    if ($arguement =~/^-(\S+)$/){
      $type=$1;
    }
    else{
      if ($type eq "rumplestiltskin"){
	$i++;
	push @args,"arg$i",$arguement;
      }
      else{
	push @args,$type,$arguement;
	$type="rumplestiltskin";
      }
    }
  }
  return (@args);
}
#----------------------------------------------------------

#SUB-ROUTINE-----------------------------------------------
sub OrderArray{
  #  This reads in an array of values and orders them 
  #returning the list of the order
  
  my @values=@_;
  my %map;
  my $i;
  my @sorted;
  my @listout;
  my @temp;
  my $flag;
  my $j=0;
  my $old='rumplestiltsin';
  
  for $i(0..scalar(@values)-1){
    if (exists($map{$values[$i]})){
      $map{$values[$i]}=join ',',$map{$values[$i]},$i;
    }
    else{
      $map{$values[$i]}=$i;
    }
  }
  @sorted= sort NumericallySort @values;
  for $i (0..scalar(@sorted)-1){
    if ($map{$sorted[$i]} =~ /,/){
      unless ($old eq $map{$sorted[$i]}){
	$j=0;
	$old=$map{$sorted[$i]};
      }
      @temp=split /,/ , $map{$sorted[$i]};
      $listout[$i]=$temp[$j];
      $j++;
    }
    else{
      $j=0;
      $listout[$i]=$map{$sorted[$i]};
    }
  }
  return @listout;
}
#----------------------------------------------------------


#SUB-ROUTINE-----------------------------------------
sub NumericallySort {$a <=> $b;}
#----------------------------------------------------
