#!/ebi/extserv/bin/perl/bin/perl

=head1 NAME

index_data.pl - Indexer and converter (ascii to binary) script for InterProScan datafiles.

=head1 SYNOPSIS

  index_data.pl -f file1 [-f fileN] -inx {to index file} -iforce {force reindexing} -bin {to convert ascii datafile to binary}
                -bforce {force reconversion ascii to bin}  -v {verbose(optional)} -h {display this help}

	Options:
		-f       The name of the file you want to index. See description below. Each file should have a -f flag.
		-inx     To tell the program you want to index one or several files for InterProScan.
		-iforce  Force the reindex of the file if already indexed (update the index).
		-bin     To tell the program to convert a hmm library file (ascii) to a binary file.
		         This action will speed up the hmm analysis up to 40%.
		-bforce  Force the reconversion of a ascii file to a binary if this file already exists.
		-v       Be verbose during actions.
		-h       Display a help and exit.

=head1 DESCRIPTION

  This script allow user to index or reindex (-iforce option) some files needed by InterProScan.
  You can give a list of files to be indexed.

  Supported files for InterProScan indexing (.inx) are :
  - interpro.xml
  - match.xml
  - Gene3D.hmm
  - Pfam
  - Pfam-A.seed
  - Pfam-C
  - prints.pval
  - prodom.ipr
  - sf.seq
  - sf_hmm
  - smart.HMMs
  - smart.desc (not public)
  - smart.thresholds (not public)
  - superfamily.hmm
  - TIGRFAMs_HMM.LIB

  If the files are already indexed and up to date, the indexing is skipped unless you use -iforce option.

  You can also use this script to convert or reconvert (-bforce) your ascii HMM datafile to binary file to speed up (~40%) the 
  application using hmmer (Gene3D, Pfam, Tigr, Smart, PIR and Superfamily).
  Supported files for conversion are :
  - Pfam (Pfam)
  - TIGRFAMs_HMM.LIB (Tigr)
  - sf_hmm (PIR)
  - sf_hmm_sub (PIR)
  - smart.HMMs (Smart)
  - superfamily.hmm (Superfamily)
  - Gene3D.hmm (Gene3D)

  Using the -bin option, the new binary file created will be named with a new extension '.bin' (e.g. Pfam -> Pfam.bin)

  Finally, any fasta files should also be indexed for blast.
  - sf.seq
  - prodom.ipr

=head1 VERSIONS

$Id: index_data.pl,v 1.3 2005/11/16 16:19:34 hunter Exp $

Copyright (c) European Bioinformatics Institute 2002

=head1 AUTHORS / ACKNOWLEDGMENTS

  Emmanuel Quevillon <tuco@ebi.ac.uk>

=cut


BEGIN {
$ENV{IPRSCAN_HOME} ='/scratch/tmp/iprscan';
  $ENV{IPRSCAN_LIB} = $ENV{IPRSCAN_HOME} . "/lib";
}


use lib "$ENV{IPRSCAN_LIB}";
use Index::InterPro;
use Index::IprMatches;
use Index::Prints;
use Index::SmartDesc;
use Index::SmartThresholds;
use Index::Hmm;
use Index::PfamClan;
use Index::PfamA;
use Dispatcher::Tool;
use Getopt::Long;
use strict;

$| = 1;

my($verbose, $help, $iforce, $binary, $bforce, $index, $bin);
my @files;

my $module = { 'interpro.xml'     => { 'pack'   => 'Index::InterPro',
				       'regrec' => 'recdel.interpro',
				       'list'   => ['ac', 'name', 'short_name', 'type', 'meth_ac', 'meth_name', 'class', 'taxonomy']},
	       'match.xml'        => { 'pack'   => 'Index::IprMatches',
				       'regrec' => 'recdel.match',
				       'list'   => ['crc', 'id']},
	       'smart.desc'       => { 'pack'   => 'Index::SmartDesc',
				       'regrec' => 'recdel.smart.desc',
				       'list'   => ['ac']},
	       'smart.thresholds' => { 'pack'   => 'Index::SmartThresholds',
				       'regrec' => 'recdel.smart.thresholds',
				       'list'   => ['ac']},
	       'Pfam'             => { 'pack'   => 'Index::Hmm',
				       'regrec' => 'recdel.hmm',
				       'list'   => ['ac', 'name']},
	       'prints.pval'      => { 'pack'   => 'Index::Prints',
				       'recdel' => 'recdel.prints',
				       'list'   => ['ac', 'name']},
	       'superfamily.hmm'  => { 'pack'   => 'Index::Hmm',
				       'regrec' => 'recdel.hmm',
				       'list'   => ['ac', 'name', 'id']},
	       'sf_hmm'           => { 'pack'   => 'Index::Hmm',
				       'regrec' => 'recdel.hmm',
				       'list'   => ['ac', 'name']},
	       'sf_hmm_sub'       => { 'pack'   => 'Index::Hmm',
				       'regrec' => 'recdel.hmm',
				       'list'   => ['ac', 'name']},
	       'TIGRFAMs_HMM.LIB' => { 'pack'   => 'Index::Hmm',
				       'regrec' => 'recdel.hmm',
				       'list'   => ['ac', 'name']},
	       'smart.HMMs'       => { 'pack'   => 'Index::Hmm',
				       'regrec' => 'recdel.hmm',
				       'list'   => ['ac', 'name']},
	       'prodom.ipr'       => { },
               'sf.seq'           => { },
               'Gene3D.hmm'       => { 'pack'   => 'Index::Hmm',
                                       'regrec' => 'recdel.hmm',
                                       'list'   => ['ac', 'name']},
	       'Pfam-C'           => { 'pack'   => 'Index::PfamClan',
				       'regrec' => 'recdel.hmm',
				       'list'   => ['clan', 'id', 'desc', 'memb']},
	       'Pfam-A.seed'      => { 'pack'   => 'Index::PfamA',
				       'regrec' => 'recdel.hmm',
				       'list'   => ['ac', 'id', 'desc', 'nested']}
	     };

unless(&GetOptions('f=s'    => \@files,
		   'v'      => \$verbose,
		   'inx'    => \$index,
		   'iforce' => \$iforce,
		   'h'      => \$help,
		   'bin'    => \$bin,
		   'bforce' => \$bforce
		  )){
    usage();
}

if($help){
    usage();
}

unless($index || $bin){
    print "\nERROR: No action precised to do (indexing or binary conversion?)\n\n";
    usage();
}


unless(scalar(@files)){
    $binary =  [ ];
    for(keys %$module){

	if($index){
	    push(@files, $_) unless($_ =~ /sf_hmm_sub/);
	}
	
	if($bin){
	    push(@$binary, $_) if($module->{$_}->{pack} =~ /Hmm/);
	}
    }
    print "\nNo specific file(s) given, indexing all files:\n\n",join("\n", @files),"\n\n" if($index);
    print "\nNo specific file(s) given, converting all files:\n\n",join("\n", @$binary), "\n\n" if($bin);
    @files = @$binary unless $index;
}

if($index && grep { /sf_hmm_sub/ } @files){
    print STDERR "Sorry, file [sf_hmm_sub] can only be bin converted, not indexed.\nPlease remove it from your list of files.\nIndexing exited.\n";
    exit(1);
}

my ($res, $msg);

#We create a Tool object to retrieve specific informations for each index file
print "Creating Tool Object..." if($verbose);
my $tool = new Dispatcher::Tool();
print "Done\n" if($verbose);

$tool->setToolName('index');

print "Parsing Configuration file..." if($verbose);
($res, $msg) = $tool->parseConfig(1, {});
die $msg unless($res);
print "Done\n" if($verbose);

my $path = $tool->getConfigValue('iprscan.home') || $ENV{IPRSCAN_HOME};
die "ERROR: No path set for configuration file" unless($path);
die "ERROR: [$path] is not readable" unless(-d $path);

my $inx;
my $caller;

foreach my $file (@files){

    my $next = 0;

    my $f = $path . "/data/" . $file;

    if($file eq 'prodom.ipr' || $file eq 'sf.seq'){
	die "ERROR: File [$f] not found : $!" unless(-f $f);
	($res, $msg) = FormatBlastDB("$f", $verbose);
	die "Blast indexing ERROR on $file : $msg" unless($res);
	next;
    }

    unless($module->{$file}){
	print STDERR "\nERROR: This file [$file] is not supported by InterProScan\n";
	print STDERR "Please check supported files in the list below :\n\n";
	usage();
    }
    
    #These files are not public.
    next if($f =~ /smart\.thresholds|smart\.desc/ && !-f $f);
    die "ERROR: File $f does not exist" unless(-f $f or -l $f);
  
    if($index){

	if($verbose){
	    print "\n========================================================\n";
	    print "\tIndexing for $file";
	    print "\n========================================================\n";
	}

	$caller = $module->{$file}->{pack};
	
	print "Creating $caller module ... " if $verbose;
	($res, $inx) = $caller->new($f, 1);
	die "ERROR: $inx" unless($res);
	print "DONE\n\n" if $verbose;
	
	if($res){
	    if($iforce){
		print "... File $f being reindexed ...\n" if($verbose);
		if(-f $inx->{inx}){
		    print "removing ", $inx->{inx}, " ... " if($verbose);
		    $res = unlink($inx->{inx});
		    unless($res){
			print "Unable to remove $f.inx : $!\nNext file\n";
			$next = 1;
		    }
		    print "DONE\n" if($verbose);
		}else{
		    print "*****\nWARNING !! : Index file ", $inx->{inx}, " not here. Index file will be created during indexing.\n*****\n";
		}
	    }else{
		if(-f $inx->{inx}){
		    print "Cheking indexes for file $f ... " if($verbose);
		    ($res, $msg) = $inx->checkIndex();
		    die "ERROR: $f $msg" unless $res;
		    print "DONE\n" if($verbose);
		}else{
		    print "*****\nWARNING !! Index file ", $inx->{inx}, " not here. Index file will be created during indexing.\n*****\n";
		    $res = 0;
		}
		if($res){
		    print "File $f already indexed ... next one\n" if($verbose);
		    $next = 1;
		}
	    }
	}
	
	unless($next){
	    print "Building index from file $f, may take long time ... " if($verbose);
	    ($res, $msg) = $inx->buildIndex($module->{$file}->{list});
	    die "ERROR: Failed to build index from file $f : $msg\n" unless($res);
	    print "DONE\n\n" if($verbose);
	    unless(ref($msg) eq 'HASH'){ print "table not a hash\n"; exit;}
	    
	    print "Creating Index file for datafile $f ... " if($verbose);
	    ($res, $msg) = $inx->indexOut($msg);
	    die "ERROR: Problem during indexation of datafile $f : $msg\n" unless($res);
	    print "DONE\n\n\n" if($verbose);;
	}
    }


    #Binary conversion
    if($bin && $module->{$file}->{pack} =~ /Hmm/){

	if($verbose){
	    print "\n========================================================\n";
	    print "\tBinary Conversion for $file";
	    print "\n========================================================\n";
	}
	print "Converting ascii file $file to binary file ... " if $verbose;

	if(-e "$f.bin"){
	    unless($bforce){
		print "\nFile $f.bin already exists. No need to convert it. Skipped\n" if $verbose;
		print "DONE\n" if$verbose;
		next;
	    }else{
		print "\nReconversion\n" if $verbose;
		unless(unlink("$f.bin")){
		    die "ERROR: Unable to remove $f.bin : $!";
		}
	    }
	}
	
	unless(-e "$path/bin/binaries/hmmconvert" && -x "$path/bin/binaries/hmmconvert"){
	    print STDERR "\n\nERROR: Cannot convert $file, hmmconvert binary does not exist or not executable : $!\n";
	    print STDERR "\nConversion skipped\n";
	    exit 1;
	}else{
	    #We don't want to be bored by the hmmconvert output, so we redirect the standard output and error to /dev/null
	    if(system("$path/bin/binaries/hmmconvert -b $f $f.bin >& /dev/null")){
		print STDERR "ERROR: Problem during the conversion of file $f : $!";
		exit 1;
	    }
	}
	print "DONE\n" if $verbose;
    }

} #end for loop


print "*********************************************************";
print "\n- Indexing for ".join(', ', @files), " DONE" if($index);
print "\n- Binary conversion for ". join(', ', grep { $module->{$_}->{pack} =~ /Hmm/ } @files), " DONE" if($bin);
print "\n*********************************************************\n";


#################
# SubRoutines
#################


#################
# usage
#################

sub usage {

    print "\n================";
    print "\n INDEX_DATA.PL";
    print "\n================\n";
    print "Usage : $0 -f file1 [-f fileN] -inx -iforce -bin -bforce -v -h\n";
    print "Options:
                -f       The name(s) of the file(s) you want to index. See description below. Each file should be
                         preceeded by a -f flag.
                         NOTE: DO NOT GIVE FULL PATH! JUST THE NAME OF THE FILE(S) TO UPDATE!!

                -inx     To tell the program you want to index one or several files for InterProScan.

                -iforce  Force reindexing the file(s) if already indexed (update the index).

                -bin     To tell the program to convert a hmm library file (ascii) to a binary file.
                         This action will speed up the hmm analysis up to 40%.

                -bforce  Force the reconversion of a ascii file to a binary if this file already exists.

                -v       Be verbose during actions (optional)

                -h       Displays a help and exit.\n\n";

    print "Supported Files for InterProScan indexing:\n";
    for(sort keys %$module){
	print "- $_\n" unless($_ =~ /sf_hmm_sub/);
    }
    print "\nUsing -bin option to convert ascii datafile to binary\n";
    print "Supported files are :\n";
    for(sort keys %$module){
	print "- $_\n" if($module->{$_}->{pack} =~ /Hmm/);
    }
    print "\n";
    exit;

}

#################
# FormatBlastDB
#################

sub FormatBlastDB ($) {

    my($file, $verbose) = @_;

    return(0, "FormatBlastDB: Need a file to index.") unless($file);

    my($mtime_data, $mtime_inx);

    $mtime_data = (stat($file))[9];
    $mtime_inx  = (stat("$file.psq"))[9] if(-f "$file.psq");

    my $options=''; # default is no options
    $options = '-p T -o T' if ($file =~ /prodom.ipr/);

    my $formatdb = "$path/bin/binaries/blast/formatdb $options -i $file -l $path/data/formatdb.log";

    if(($mtime_inx && $mtime_inx < $mtime_data) || ($mtime_data && !$mtime_inx) || ($mtime_inx && ($mtime_data < $mtime_inx) && $iforce)){
	print "Indexing $file ... " if($verbose);
	if(system("$formatdb")){
	    return(0, "FormatBlastDB: Formatdb command [$formatdb] failed: $!");
	}
	print "DONE\n" if($verbose);
	return(1, '');
    }elsif($mtime_inx && $mtime_data < $mtime_inx){
	print "$file indexes look fresher than data, indexing skipped\n" if($verbose);
	return(1, '');
    }else{
	return(0, "FormatBlastDB: MISSING SOMETHING? $file not indexed.");
    }
}
