#!/ebi/extserv/bin/perl/bin/perl

=head1 NAME

iprscan_wrapper.pl - wrapper for InterProScan.

=head1 SYNOPSIS

iprscan_wrapper.pl < parameter-file

=head1 DESCRIPTION

Parameter file should contain parameters written with
Dispatcher::Tool writeParamFile function.

=head1 VERSIONS

$Id: iprscan_wrapper.pl,v 1.3 2005/11/16 16:44:52 hunter Exp $

Copyright (c) European Bioinformatics Institute 2002

=head1 AUTHORS / ACKNOWLEDGEMENTS

Ville Silventoinen <vsi@ebi.ac.uk>
Emmanuel Quevillon <tuco@ebi.ac.uk>

=cut

BEGIN {

$ENV{IPRSCAN_HOME}='/scratch/tmp/iprscan';

  unless($ENV{IPRSCAN_LIB}){
    $ENV{IPRSCAN_LIB} = "$ENV{IPRSCAN_HOME}/lib";
  }
  #ckeck if these path are directories.
  unless(-d $ENV{IPRSCAN_HOME}){
    print STDERR "$0 : $ENV{IPRSCAN_HOME} is not a directory!\n";
    exit;
  }
  unless(-d $ENV{IPRSCAN_LIB}){
      print STDERR "$0 : $ENV{IPRSCAN_LIB} is not a directory!\n";
      exit;
    }
}

use Data::Dumper;
use English;
use lib $ENV{IPRSCAN_LIB};
use Dispatcher::Tool::InterProScan;
use Index::Fasta;
use strict;

my $tool = new Dispatcher::Tool::InterProScan();
my ($res, $msg, $seqs, $jobs);

# parse the input parameters
my $params = { };

($res, $msg) = $tool->readParamFile(\*STDIN, $params);
handleError($tool, $msg, undef, undef, 1) unless $res;

handleError($tool, "iprscan tool name not defined in input parameters",
	    undef, undef, 1) unless $params->{tool};
handleError($tool, "iprscan jobid not defined in input parameters",
	    undef, undef, 1) unless $params->{jobid};
handleError($tool, "iprscan applications not defined in input parameters",
	    undef, undef, 1) unless $params->{appl};

$tool->setParams($params);
$tool->setToolName($tool->getParam('tool'));

($res, $msg) = $tool->parseConfig(1, $tool->getParams());
handleError($tool, $msg, undef, undef, 1) unless $res;

my $tname = $tool->getToolName();

my $efile = $tool->getConfigValue('toolerrors');
handleError($tool, "$tname toolerrors not configured", undef, undef, 1) unless $efile;

my $ifile = $tool->getConfigValue('toolinput');
handleError($tool, "$tname toolinput not configured", $efile, undef, 1) unless $ifile;

my $nfile = $tool->getConfigValue('toolnocrc');
handleError($tool, "$tname toolnocrc not configured", $efile, undef, 1) unless $nfile;

my $sfile = $tool->getConfigValue('toolstatus');
handleError($tool, "$tname toolstatus not configured", $efile, undef, 1) unless $sfile;

my $rfile = $tool->getConfigValue('toolraw');
handleError($tool, "$tname toolraw not configured", $efile, undef, 1) unless $rfile;

my $xfile = $tool->getConfigValue('toolxml');
handleError($tool, "$tname toolxml not configured", $efile, undef, 1) unless $xfile;

my $tfile = $tool->getConfigValue('tooltxt');
handleError($tool, "$tname tooltxt not configured", $efile, undef, 1) unless $tfile;

my $hfile = $tool->getConfigValue('toolhtml');
handleError($tool, "$tname toolhtml not configured", $efile, undef, 1) unless $hfile;

my $cfile = $tool->getConfigValue('mailconfirm');
handleError($tool, "$tname mailconfirm not configured", $efile, undef, 1) unless $cfile;

my $qname = $tool->getConfigValue('queue');
handleError($tool, "$tname queue not configured", $efile, undef, 1) unless $qname;

my $debug = $tool->getConfigValue('debug');

my $jobid = $tool->getParam('jobid');

my $applist = $tool->getParam('appl');
my @apps = split(',', $applist) if $applist;
handleError($tool, "$tname submission failed, no applications defined",
	    $efile, undef, 1) unless $applist && @apps > 0;

my $checkcrc  = $tool->getParam('crc');
my $goterms   = $tool->getParam('goterms');
my $iprlookup = $tool->getParam('iprlookup');

# parse id, length and crc64 for each sequence and creates chunk if needed
($res, $seqs) = $tool->checkSequences($ifile, $checkcrc, $nfile, \@apps, $iprlookup, $goterms); #iprlookup and goterms were 1
handleError($tool, "$tname submission failed: $seqs", $efile) unless $res;

my $h_chunk = { };

$h_chunk->{nbchunk} = $seqs->{nchunk};
delete $seqs->{nchunk}; #avoid to compromise the next command

#We index the input sequences file here instead of in the cgi script (iprscan). Like that we don't need to read once again the input sequence file. 
#We do it only once.
my $index;
($res, $index) = new Index::Fasta($ifile);
handleError($tool, $index, $efile, undef, 1) unless $res;

($res, $msg) = $index->indexOut($seqs->{inx});
handleError($tool, $msg, $efile, undef, 1) unless $res;

delete $seqs->{inx}; #avoid to compromise the next command

my $nseqs = keys %$seqs;
handleError($tool, "no input sequences", $efile) unless $nseqs > 0;

#
# Set the sequence info to the parameters file so the CGI script doesn't need
# to parse the sequences again when showing the result page.
# This also enables us to show the results in the order of the input.
#
$tool->setParam('nseqs', $nseqs);
my ($seq, $seqn, $cnk);

foreach $seq (keys %$seqs) {
    $seqn = $seqs->{$seq}->{seqn};
    $cnk  = $seqs->{$seq}->{cnk};
    next unless $seqn;
    $tool->setParam("seq.$cnk.$seqn.id",    $seq);
    $tool->setParam("seq.$cnk.$seqn.crc64", $seqs->{$seq}->{crc64});
    $tool->setParam("seq.$cnk.$seqn.len",   $seqs->{$seq}->{len});
}

($res, $msg) = $tool->writeParamFile($tool->getConfigValue('toolparams'), $tool->getParams());
handleError($tool, "$tname submission failed: $msg", $efile, undef, 1) unless $res;


my $email = $tool->getParam('email');

unless ($tool->isInteractive()) {
    # create and send a confirmation mail to the user
    ($res, $msg) = $tool->createConfirmationMail(undef, $cfile);
    handleError($tool, "failed to create confirmation mail: $msg", $efile, undef, 1) unless $res;

    ($res, $msg) = $tool->sendMail($email,
				   $tool->getConfigValue('mailconfsubject'),
				   undef, $cfile,
				   $tool->getConfigValue('mailfromaddr'),
				   $tool->getConfigValue('mailerrorto'));
    handleError($tool, "failed to send confirmation mail: $msg", $efile, undef, 1) unless $res;
}


#User wants to run jobs in parallel?
my $parallel = $tool->getParam('onego');

unless($parallel){

  my $out = $tool->getConfigValue('nostarted');

  for (my $i = 1; $i <= $h_chunk->{nbchunk}; $i++){

    ($res, $msg) = $tool->addChunk("$i", \$out);
    handleError($tool, $msg, $efile, undef, 1) unless($res);

    ($res, $msg) = $tool->getNoJobsStarted($out);
    handleError($tool, $msg, $efile, undef, 1) unless($res);

  }
}



for(my $chunk = 1; $chunk <= $h_chunk->{nbchunk}; $chunk++){

    ($res, $msg) = $tool->addChunk("$chunk", \$nfile);
    handleError($tool, $msg, $efile, undef, 1) unless($res);

    if (-e $nfile && -s $nfile > 0) {
	
	($res, $jobs, $msg) = $tool->submitJobs($nfile);
	handleError($tool, $msg, $efile, undef, 1) unless($res);

	$h_chunk->{$chunk}->{jobs}  = $jobs;
	$h_chunk->{$chunk}->{status} = { };
	
	for(keys %$jobs){
	    #Set each job status to 'running'
	    $h_chunk->{$chunk}->{status}->{$_} = Dispatcher::Queue::JOB_RUNNING;
	}
	
	unless($parallel){
	    if($tool->getConfigValue('nostarted')){
		my $nostarted = $tool->getConfigValue('nostarted');
		($res, $msg) = $tool->addChunk("$chunk", \$nostarted);
		if(-f $nostarted){
		    unlink($nostarted);
		}
	    }

	    ($res, $msg) = supervise($tool, $h_chunk, $chunk);
	    handleError($tool, $msg, $efile, undef, 1) unless($res);

	} #end unless parallel
	
    }elsif($checkcrc){
	
	#We introduce chunk directory in the path
	($res, $msg) = $tool->addChunk($chunk, \$rfile, \$xfile);
	handleError($tool, $msg, $efile, undef, 1) unless($res);
	
	($res, $msg) = $tool->createRawResult($seqs, $h_chunk, $chunk);
	unless ($res) {
	    # something went wrong, terminate any unfinished jobs
	    #$tool->terminateJobs($jobs) if (keys %$jobs > 0);
	    handleError($tool, "failed to create raw result: $msg", $efile, undef, 1);
	}
	
	#
	# create the XML result file
	#
	# NOTE: always use a temporary output file first, it prevents the client
	# reading a zero size file. It can be renamed later, so when the new file
	# appears in the NFS file system, it is complete and readable.
	#
	
	#For all situations (CGI, command line) we create xml output...
	my $xmlcmd = $tool->getConfigValue('iprscan.xml.converter');
	handleError($tool, "$tname iprscan.xml.converter not configured", $efile, undef, 1) unless $xmlcmd;

	$xmlcmd =~ s/\%format/xml/;
	$xmlcmd =~ s/\%toolraw/$rfile/;
	
	system("$xmlcmd > $xfile.tmp 2> $efile");
	handleError($tool, "failed to create XML output", $efile, $efile, 1) if -s $efile > 0;

	unless (rename("$xfile.tmp", $xfile)) {
	    handleError($tool, "failed to rename $xfile.tmp: $!", $efile, undef, 1);
	}
	
	if (-s $xfile == 0 && -s $efile > 0) {
	    handleError($tool, "failed to create XML output", $efile, undef, 1);
	}
	
	#We need to create txt output if the submission is CGI.
	if($tool->getParam('submission') eq 'cgi'){
	    my $txtcmd = $tool->getConfigValue('iprscan.xml.converter');
	    my $txtfile = $tool->getConfigValue('tooltxt');
	    
	    ($res, $msg) = $tool->addChunk("$chunk", \$txtfile);
	    return(0, $msg) unless($res);

	    $txtcmd =~ s/\%format/txt/;
	    $txtcmd =~ s/\%toolraw/$rfile/;

	    system("$txtcmd > $txtfile.tmp 2> $efile");
	    handleError($tool, "failed to create TXT output", $efile, $efile, 1) if -s $efile > 0;

	    unless (rename("$txtfile.tmp", $txtfile)) {
		handleError($tool, "failed to rename $txtfile.tmp: $!", $efile, undef, 1);
	    }

	    if (-s $txtfile == 0 && -s $efile > 0) {
		handleError($tool, "failed to create TXT output", $efile, undef, 1);
	    }
	}
	next;
    }  #end elsif $checkcrc
    else{
	handleError($tool, "Could not find $nfile: $!", $efile, undef, 1);
    }
}  # end for loop

#we launch all the job in one go and wait all finished
if($parallel){
    ($res, $msg) = supervise($tool, $h_chunk);
    handleError($tool, $msg, $efile, undef, 1) unless($res);
}


#############  now we need to merge all the raw file in a single one in the root directory
#############  and then create the xml file to be displayed when using command line.
#############  If user specified a specific format we create output for this format
#############  otherwise we create xml output by default.

if($tool->getParam('submission') eq "cli"){

  my $dir = $tool->getConfigValue('workdir');
  handleError($tool, "Workdir not defined or does not exist", $efile, undef, 1) unless($dir && -d $dir);

  my $toolraw = $tool->getConfigValue('toolraw');
  handleError($tool, "Toolraw not defined", $efile, undef, 1) unless($toolraw);

  my $format = $tool->getParam('format');
  my $toolxml = ($format =~ /htm[l]?/ ? $tool->getConfigValue('toolxml') : $tool->getConfigValue("tool$format"));
  handleError($tool, "tool$format not defined", $efile, undef, 1) unless($toolxml);

  #Merge all the .raw files to one
  if(system("find $dir -name \\*.raw -exec cat {} >> $toolraw.tmp \\;")){
    handleError($tool, "System command [find] failed : $!", $efile, undef, 1);
  }

  unless(rename("$toolraw.tmp", "$toolraw")){
    handleError($tool, "Unable to rename $toolraw.tmp : $!", $efile, undef, 1);
  }

  my $xcmd = $tool->getConfigValue('iprscan.xml.converter');
  handleError($tool, "converter command not configured", $efile, undef, 1) unless($xcmd);

  my $bacform = $format;

  if($format =~ /htm[l]?/){
      #We create output in xml format, converter uses Dispatcher::InterProScan::createResultPicutre/View to create html output.
      $xcmd =~ s/\%format/xml/;
      $bacform = 'xml';
  }else{
      $xcmd =~ s/\%format/$format/;
  }
  $xcmd =~ s/\%toolraw/$toolraw/;


  if($format !~ /raw/i){ #We do nothing for raw format, already exists.
    if(system("$xcmd > $toolxml.tmp 2> $efile") || -s $efile > 0){
      handleError($tool, "Failed to create $bacform output : $!", $efile, undef, 1);
    }

    unless(rename("$toolxml.tmp", "$toolxml")){
      handleError($tool, "Failed to rename $bacform output : $!", $efile, undef, 1);
    }

    if(-s $toolxml == 0 && -s $efile > 0){
      handleError($tool, "Failed to create $bacform output : $!", $efile, undef, 1);
    }
  }

  #Now we check if the format asked by the user is html to convert the xml page to html.
  if($format =~ /htm[l]?/){

    $xcmd = $tool->getConfigValue('iprscan.xml.converter');
    handleError($tool, "converter command not configured", $efile, undef, 1) unless($xcmd);

    $xcmd =~ s/\%format/html/;
    $xcmd =~ s/\%toolraw/$toolraw/;

    $toolxml = $tool->getConfigValue('toolhtml');
    handleError($tool, "toolhtml not configured", $efile, undef, 1) unless($toolxml);

    if(system("$xcmd > $toolxml.tmp 2> $efile")){
      handleError($tool, "Failed to create HTML output : $!", $efile, undef, 1);
    }

    unless(rename("$toolxml.tmp", "$toolxml")){
      handleError($tool, "Failed to rename HTML output : $!", $efile, undef, 1);
    }
  }
}


if ($email && $email =~ /\@/) {

    # send the results also via email
    my $url = $tool->getConfigValue('workurl');
    $url .= '1';

    my $title = $tool->getParam('title') || $jobid;
    my $rfile = $tool->getConfigValue('toolraw');
    my $dir = $tool->getConfigValue('workdir');

    unless(-f $rfile){
      if(system("find $dir -name \\*.raw -exec cat {} >> $rfile.tmp \\;")){
	handleError($tool, "System command [find] failed : $!", $efile, undef, 1);
      }
      unless(rename("$rfile.tmp", "$rfile")){
	handleError($tool, "Unable to rename $rfile.tmp : $!", $efile, undef, 1);
      }
    }

    # mail link to the results to user
    $msg  = "**  Your results of \"$title\" are in:\n";
    $msg .= "**\n**\n";
    $msg .= "**  $url\n";
    $msg .= "**\n**\n";
    $msg .= "**  Please make sure the above URL is passed to your browser\n";
    $msg .= "**   in one line, as some mail clients may wrap long lines.\n";
    $msg .= "**\n**\n";
    $msg .= "**\n**\n\n";

    my $subject = $tool->getParam('title') || $tool->getConfigValue('mailresultsubject');

    ($res, $msg) = $tool->sendMail($email, $subject, $msg, $rfile,
				   $tool->getConfigValue('mailfromaddr'),
				   $tool->getConfigValue('mailerrorto'));

    handleError($tool, "failed to send result email: $msg", $efile, undef, 1) unless $res;
    unlink($rfile) unless($tool->getParam('submission') eq 'cli');
}

# cleanup temporary files
unlink($cfile, $sfile);
unlink($efile) if -e $efile && -s $efile == 0;
($res, $msg) = $tool->cleanup($h_chunk);
handleError($msg, $efile, 1) unless $res;

exit;




=head2 handleError

 Description: Convenience method that reports the error and exits. This method
              also writes errors to the $efile so the CGI can show errors to
              the end-user.
 Arguments:   $tool       Dispatcher::Tool object.
              $msg        Error message.
              $efile      InterProScan error file (optional).
              $jfile      Job error file (optional).
              $addinfo    1: add contact details, 0: does not
 Returns:     exits

=cut

sub handleError($$$;$$) {
    my ($tool, $msg, $efile, $jfile, $addinfo) = @_;

    my $html = $tool->isInteractive();

    if ($addinfo) {
	$msg .= "\n\nTo help us solve the problem, please copy and paste ";
	$msg .= "this error\nand email it to the administrator ";
	my $admin = $tool->getConfigValue('adminaddr');

	if ($html) {
	    $msg .= "<a href=\"mailto:$admin?subject=Error in InterProScan job $jobid\">$admin</a>";
	}
	else {
	    $msg .= $admin;
	}

	$msg .= ".\n\nSorry about the inconvenience and thank you for using InterProScan.\n\n";
    }

    if ($efile && -f $efile && open(ERRORS, "> $efile")) {
	$msg =~ s/\n/<br>/g if $html;
	print ERRORS $msg;

	if ($jfile && -f $jfile && open(JOBERR, "< $jfile")) {
	    print ERRORS "<pre>" if $html;

	    while (<JOBERR>) {
		print ERRORS $_;
	    }

	    print ERRORS "</pre>" if $html;
	    close JOBERR;
	}
	close ERRORS;

	$tool->errorExit(undef, $efile);
    }
    else {
	$tool->errorExit($msg);
    }
}


=head2 supervise

 Description: Checks status of the jobs untill they are finished. Then parses
              results, creates raw and xml files in each chunks.

 Arguments:   $tool       Dispatcher::Tool object.
              $h_chunk    Hash table containing info about jobs in chunks.
              $chunk      Chunk number to run supervise on (optional). If no
                          variable is given, supervise run on every chunks.
 Returns:     1, ''       on success
              0, message  on error

=cut

sub supervise {

  my ($tool, $h_chunk, $chunk) = @_;
  #to be removed if we move this function to InterProScan.pm package

  return(0, "supervise: Not a Tool reference!") unless($tool->isa("Dispatcher::Tool"));
  return(0, "supervise: Need a reference to a hash table (h_chunk)") unless($h_chunk && ref($h_chunk) eq 'HASH');

  my $interval  = $tool->getConfigValue('pollinterval')  || '60';
  my $maxrounds = $tool->getConfigValue('maxpollrounds') || '10000';
  my ($jobid, $status, $jobs, $report);
  my ($start, $stop, $tstart, $tstop);
  my %joberrs;

  $start = $chunk ? $chunk : '';
  $stop  = $chunk ? $chunk : $h_chunk->{nbchunk};

  my $round = 0;

  while (1) {

    sleep($interval);

    my $alldone = 1;  # be optimistic
    $round++;

    #Get all the status (number) for all jobs
    $h_chunk = getUnfinishedJobs($h_chunk, $start);


    ($res, $h_chunk) = $tool->checkJobs($h_chunk, $start);
    unless ($res) {
      # something went wrong, terminate any unfinished jobs
      return(0, "supervise: failed to check jobs: $h_chunk");
    }

    if ($tool->getConfigValue('updatestatus')) {
      #
      # update the job status file for CGI/CLI
      #

      ($res, $h_chunk) = $tool->updateStatus($h_chunk, $start);
      return(0, "supervise: $h_chunk") unless($res);

    } #end if

    $tstart  = $start ? $start : 1;
    $tstop   = $stop;

    for($tstart; $tstart <= $tstop; $tstart++){

      $jobs = $h_chunk->{$tstart}->{jobs};
      $report = $h_chunk->{$tstart}->{status};
      my $cnkjobdone = 1; #optimistic, all jobs for this chunk a done.

      foreach $jobid (keys %$report) {

	$status = $report->{$jobid};

	if ($status =~ /failed/ || $status == Dispatcher::Queue::JOB_FAILED) {
	    $joberrs{$jobid} = 1 unless(exists($joberrs{$jobid}));
	}
	elsif ($status == Dispatcher::Queue::JOB_PENDING) {
	  #		$log->debug("$jobid job $jobname pending") if $log;
	  $alldone = 0;
	  $cnkjobdone = 0;
	}
	elsif ($status == Dispatcher::Queue::JOB_RUNNING) {
	  #		$log->debug("$jobid job $jobname running") if $log;
	  $alldone = 0;
	  $cnkjobdone = 0;
	}
	elsif ($status == Dispatcher::Queue::JOB_DONE) {

	  #We need to check if there is no error. Sometime the main program exits
	  #reporting some errors and the exitcode in the .exitcode file is 0.

	  my $exfile = $jobs->{$jobid}->{exitcode};
	  my $erfile = $jobs->{$jobid}->{errors};
	  my $oufile = $jobs->{$jobid}->{output};

	  if(-f $exfile && -s $exfile > 0 ){
	    unless(open(EX, "<$exfile")){
	      return(0, "supervise: could not open $exfile : $!");
	    }
	    local $INPUT_RECORD_SEPARATOR;
	    my $errcode = <EX>;
	    close EX;

	    $INPUT_RECORD_SEPARATOR = '\n';

	    unless(defined($errcode) && $errcode eq "0"){
	      $h_chunk->{$tstart}->{status}->{$jobid} = Dispatcher::Queue::JOB_FAILED;

	      if(-f $erfile && -s $erfile > 0){
	        #In the case of failed job, we check if there is an error file to be displayed later in the report.
		$joberrs{$jobid} = 1 unless(exists($joberrs{$jobid}));
	      }
	    }else{
	      if(-f $erfile && -s $erfile > 0 && -s $oufile == 0){

		$h_chunk->{$tstart}->{status}->{$jobid} = Dispatcher::Queue::JOB_FAILED;
		$joberrs{$jobid} = 1 unless(exists($joberrs{$jobid}));

	      }
	    }
	  }else{
            #We consider that the jobs is still running.
	    $h_chunk->{$tstart}->{status}->{$jobid} = Dispatcher::Queue::JOB_RUNNING;
	    $alldone = 0;
	    $cnkjobdone = 0;
	  }
	}
	elsif ($status == Dispatcher::Queue::JOB_NOT_FOUND) {
	  # LSF keeps the job information only for an hour or so.
	  # When one job takes considerably longer than others, this
	  # can happen and is perfectly normal.
	  #		$log->debug("$jobid job $jobname not found (done?)") if $log;
	}

      } #end foreach $jobid (keys %$report) loop

      #All the jobs are done for this chunk?
      if($cnkjobdone){
	($res, $msg) = doRawResults($tool, $h_chunk, $tstart);
	return(0, "supervise: $msg") unless($res);
      }

    } #end for loop (tstart)

    last if $alldone;

    #Uncomment this if you want to apply a job time limit to your jobs.
    #This implementation only works with queueing system.
    #Check the FAQ for to know more about it.

    if($tool->getConfigValue('job.time.limit') && $tool->getConfigValue('queue') ne 'local'){
	if ($round > $maxrounds) {
	    # to avoid infinite loop, give up!
	    my $hours = int($round * $interval / 3600);

	    #This loop can only work with a queuing system.
	    for(my $i = 1; $i <= $h_chunk->{nbchunk}; $i++){
		$jobs = $h_chunk->{$i}->{jobs};
		$tool->terminateJobs($jobs);
	    }
	    $msg = "InterProScan job $jobid didn't finish in $hours hours, job terminated.\n";
	    return(0, "supervise: $msg");
	}
    }
  } #end while(1)

  #EQ for each failed jobs we create a report
  foreach my $job (sort keys %joberrs){
    ($res, $msg) = $tool->createFailedReport($job, $h_chunk);
    return(0, "supervise: $msg") unless($res);
  }

  return(1, '');
}


=head2 doRawResults

  Description: Method to create raw result for a specific chunk when all the jobs are done in this particular chunk.
  Arguments:   $tool    Reference to a Dispatcher::Tool object.
               $h_chunk Reference to hash containing info of chunks.
               $chunk   Number of chunk to work with.
  Returns:     1, ''    on success.
               0, msg   on error.

=cut

sub doRawResults($$){

  my($tool, $href, $cnk) = @_;

  return(0, "doRawResults: Need a reference to a Dispatcher::Tool object") unless($tool && $tool->isa("Dispatcher::Tool"));
  return(0, "doRawResults: Need a reference to a hash table") unless($href && ref($href) eq 'HASH');
  return(0, "doRawResults: Need a number representing the chunk to work with") unless($cnk && $cnk =~ /^\d+$/);

  my $rfile = $tool->getConfigValue('toolraw');
  return(0, "doRawResults: rawfile not configured") unless($rfile);

  my $xfile = $tool->getConfigValue('toolxml');
  return(0, "doRawResults: xmlfile not configured") unless($xfile);

  ($res, $msg) = $tool->addChunk("$cnk", \$rfile, \$xfile);
  return(0, $msg) unless($res);

  #If the raw file has already been done
  return(1, '') if(-f $rfile);

  ($res, $msg) = $tool->createRawResult($seqs, $href, $cnk);
  unless ($res) {
  # something went wrong, terminate any unfinished jobs
  #$tool->terminateJobs($jobs) if (keys %$jobs > 0);
  return(0, "doRawResults: failed to create raw result: $msg");
  #handleError($tool, "failed to create raw result: $msg", $efile, undef, 1);
  }

  #
  # create the XML result file
  #
  # NOTE: always use a temporary output file first, it prevents the client
  # reading a zero size file. It can be renamed later, so when the new file
  # appears in the NFS file system, it is complete and readable.
  #

  #For all situations (CGI, command line) we create xml output...
  my $xmlcmd = $tool->getConfigValue('iprscan.xml.converter');
  return(0, "doRawResults: iprscan.xml.converter not configured") unless($xmlcmd);

  $xmlcmd =~ s/\%format/xml/;
  #To introduce the chunks path
  $xmlcmd =~ s/\%toolraw/$rfile/;

  system("$xmlcmd > $xfile.tmp 2> $efile");
  return(0, "doRawResults: failed to create XML output : $!") if -s $efile > 0;

  unless (rename("$xfile.tmp", $xfile)) {
    return(0, "doRawResults: failed to rename $xfile.tmp: $!");
  }

  if (-s $xfile == 0 && -s $efile > 0) {
    return(0, "doRawResults: failed to create XML output");
  }

  #If we have CGI submission, we need to create txt output.
  if($tool->getParam('submission') eq "cgi"){
    my $txtcmd = $tool->getConfigValue('iprscan.xml.converter');
    my $txtfile = $tool->getConfigValue('tooltxt');

    ($res, $msg) = $tool->addChunk("$cnk", \$txtfile);
    return(0, "doRawResults: $msg") unless($res);

    $txtcmd =~ s/\%format/txt/;
    $txtcmd =~ s/\%toolraw/$rfile/;

    system("$txtcmd > $txtfile.tmp 2> $efile");
    return(0, "doRawResults: failed to create TXT output") if -s $efile > 0;

    unless (rename("$txtfile.tmp", $txtfile)) {
     return(0, "doRawResults: failed to rename $txtfile.tmp: $!");
    }

    if (-s $txtfile == 0 && -s $efile > 0) {
      return(0, "doRawResults: failed to create TXT output");
    }
  }

  return(1, '');
}



=head2 getUnfinishedJobs

 Description: Convenience method that returns the job IDs that are not finished.
 Arguments:   $h_chunk    Reference to hash containing info of chunks.
              $chunk      Number of chunk to work with (optional). If $chunk not
                          given, run on every chunks.
 Returns:     modified    $h_chunk in consequences.

=cut

sub getUnfinishedJobs($$) {
    my $h_chunk = shift;
    my $chunk = shift;
    my ($jobid, $status, $start, $stop);

    $start = $chunk ? $chunk : 1;
    $stop  = $chunk ? $chunk : $h_chunk->{nbchunk};

    for($start; $start <= $stop; $start++){

      foreach $jobid (keys %{$h_chunk->{$start}->{status}}) {
        $status = $h_chunk->{$start}->{status}->{$jobid};

        unless ($status) {
	  # bootstrap value for the first call
	  $status = Dispatcher::Queue::JOB_PENDING;
	  $h_chunk->{$start}->{status}->{$jobid} = $status;
        }

        if ($status != Dispatcher::Queue::JOB_DONE) {
	  $h_chunk->{$start}->{status}->{$jobid} = $status;
        }else{
	  delete $h_chunk->{$start}->{status}->{$jobid};
	}
      }
    }

    return $h_chunk;
}
