#!/ebi/extserv/bin/perl/bin/perl
#
# go through a .ass file and remove overlaps between hits to the same sequence
#
use strict;

my $usage = "Usage: remove_overlaps.pl <input.ass> <output.ass>";
my ($infile, $outfile) = @ARGV;

die "$usage\n ... " unless defined $outfile;
unless(-e $infile){
  print STDERR "File $infile doesn't exist!\n$usage\n ... ";
  exit -1;
}

# read in the input file
#
open F, $infile or die "Cannot open $infile for reading: $!\n ... ";

my (%hits);

while(<F>)
{
    chomp;

    my ($seqid, $model, $start, $end, $E, $align) = split /\s+/;

    push @{$hits{$seqid}}, { seqid  => $seqid,
			     model  => $model,
			     start  => $start,
			     end    => $end,
			     E      => $E,
			     align  => $align  };
  };

close F;


# sort out overlaps between hits, print out the result
#
unless(open F, ">$outfile"){
  print STDERR "Cannot open $outfile for writing: $!\n ... ";
  exit -1;
}

foreach my $seqid (sort keys %hits)
{
    &remove_overlaps_tuco( @{$hits{$seqid}} );

    foreach my $hit (@{ $hits{$seqid} })
    {
	# careful -- a hit may get completely removed
	next unless exists $hit->{start};

	# write the hit to the new file
	print F join( "\t",
		      $hit->{seqid},
		      $hit->{model},
		      $hit->{start},
		      $hit->{end},
		      $hit->{E},
		      $hit->{align} ), "\n";
    };
};

close F;


# remove overlaps between hits to a particular sequence
#
# input: 
#
#    ( $hit1, $hit2, $hit3, .... ) = all hits to a sequence
# 
# where each hit must include something like:
#
#    $hit->{start} = 1
#    $hit->{end}   = 100
#
# the sub alters the starts and ends so that there are no overlaps
# between the hits 
#
sub remove_overlaps_tuco
{
    my ($hit1, $hit2, $i1, $i2);

    # sort out overlaps between contigs
    for $i1 (0 .. scalar(@_)-1)
    {
	for $i2 ($i1+1 .. scalar(@_)-1)
	{
	    &remove_overlap_contigs($_[$i1], $_[$i2]);
	};
    };
};

# remove an overlap between two contigs
#
# input:  ( { start => $start_contig1, end => $end_contig1 },
#           { start => $start_contig2, end => $end_contig2 }  )
#
# alters the {start}s and {end}s so that there is no overlap
#
sub remove_overlap_contigs
{
    my ($c1, $c2) = @_;

    # undefined {start} = contig deleted
    return unless exists $c1->{start} and exists $c2->{start};

    # do the contigs overlap? if no, return now
    return if $c1->{end} < $c2->{start} or $c2->{end} < $c1->{start};

    # if we got here, there is an overlap between the contigs 
    # that needs to be resolved 

    if( $c1->{start} < $c2->{start} and $c1->{end} > $c2->{end} )
    {
	# c1: |-------------|
	# c2:    |-------|
	#
	delete $c2->{start};
	delete $c2->{end};
    }
    elsif( $c2->{start} < $c1->{start} and $c2->{end} > $c1->{end} )
    {
	# c1:    |-------|
	# c2: |-------------|
	#
	delete $c1->{start};
	delete $c1->{end};
    }
    else
    {
	# can assign 'left' and 'right' contig
	my ($l, $r, $c);

	if( $c1->{start} > $c2->{start} or $c1->{end} > $c2->{end} )
	{
	    $r = $c1; $l = $c2;
	}
	else
	{
	    $l = $c1; $r = $c2;
	};

	# get the centre of the overlap
	$c = $r->{start} + int( ($l->{end} - $r->{start})/2 );

	$l->{end}   = $c;
	$r->{start} = $c+1;
    };
};
