#!/ebi/extserv/bin/perl/bin/perl
use Data::Dumper;
use DirHandle;
use English;
use File::Path;
use Getopt::Long;
use Sys::Hostname;
use strict;

#
# http://supfam.org
# Copyright (c) 2001 MRC and Julian Gough; see http://supfam.org/SUPERFAMILY/license.html
#This program is supposed to:  run SUPERFAMILY in stand-alone mode
#Julian Gough 22.2.02
#

$| = 1;

#ASIGN-VARIABLES-------------------------------------
my $usage="$0 -i <*file.fa> -o <output> -t <scratchdir> -m <modelsdir> -d <models.tab> -p <hmmprog> -c <cpu num> -e <E-value threshold> -r <remove temporary files y/n> -s <scriptsdir>\n*=required";
my $queryfile;
my $unique = hostname() . time() . $PID;

#defaults options
my $threshold='0.02';
my $scratchdir='/scatch/tmp';
my $modelsdir='/scratch/InterPro/SCOP/models/hmmermodlib';
my $modelsacc;
my $modelstab='/scratch/InterPro/SCOP/models/model.tab';
my $output="$unique.ass";
my $hmm="/scratch/InterPro/SCOP/hmmer-2.3.1.bin.intel-linux/binaries/hmmsearch";
my $scriptdir="./";
my $cpu = 1;
#--------


my $seqs=0;
my $models=0;
my @mods;
my $mod;
my $flag;
my $modsuffix;
my @files;

# MM: new -- if set to one, expect model library as one biiig file
# (hmmpfam requires this)
my $libsinglefile = 0;

my %mstat;
my %sf;
my %des;
my $count=0;
my $remove='y'; #added 21st July
#EQ was my $remove='n'; 
my $assignment=0;
my $fasta_checker=0;
my $outsuffix;
my $verbose = 0;

#--------
#Parsing the command line

unless (&GetOptions('i:s' => \$queryfile,     #input sequence file
		    'o:s' => \$output,        #output file
		    't:s' => \$scratchdir,    #temporary directory
		    'm:s' => \$modelsdir,     #model directory or library file
		    'a:s' => \$modelsacc,     #model acc file
		    'd:s' => \$modelstab,     #model tab
		    'p:s' => \$hmm,           #path to hmm(search-score) binary
		    'e:f' => \$threshold,     #e-value,
		    'r:s' => \$remove,        #remove temporary file (y|n)
		    'v'   => \$verbose,       #be verbose
		    'c:i' => \$cpu,           #number of cpu to use (default 1)
		    's:s' => \$scriptdir)) {  #path to scriptdir (for use of other scripts)
  print "usage: $usage\n";
  exit;
}
  


#----------------------------------------------------

#ARGS------------------------------------------------

#EQ was die "Usage: $usage\n" unless (@ARGV > 0);

print "\nSUPERFAMILY running\n" if $verbose;
print "ID of this run: $unique\n" if $verbose;

$queryfile = $ARGV[0] unless $queryfile;
print "using query file of fasta sequence(s): $queryfile\n" if $verbose;

$output = $ARGV[1] if defined @$ARGV[1];
print "using output file: $output\n" if $verbose;

$scratchdir = $ARGV[2] if defined @$ARGV[2];
print "using scratch directory: $scratchdir\n" if $verbose;

$modelsdir = $ARGV[3] if defined @$ARGV[3];
print "using models directory: $modelsdir\n" if $verbose;

$modelstab = $ARGV[4] if defined @$ARGV[4];
print "using models table: $modelstab\n" if $verbose;

$hmm = $ARGV[5] if defined @$ARGV[5];
print "using HMM scoring program: $hmm\n" if $verbose;

$threshold = $ARGV[6] if defined @$ARGV[6];
print "using E-value threshold of $threshold\n" if $verbose;

$scriptdir = $ARGV[8] if defined $ARGV[8];
print "using scripts directory: $scriptdir\n" if $verbose;
print "\n" if $verbose;

#FS Added by Florence----------------------------------
# Create a directory in the user defined scratch dir
# so remove can be performed on the entire directory
# which speed up the process
# vsi 2003-11-13: uses mkpath instead of system call

unless (-d $scratchdir) {
    unless (mkpath($scratchdir)) {
	die "could not create directory $scratchdir: $!";
    }
}

unless ($hmm =~ /hmmpfam$/){
    unless (mkpath("$scratchdir/$unique")) {
	die "could not create directory $scratchdir/$unique: $!";
    }
    $scratchdir = "$scratchdir/$unique";
}

if ($remove && $remove eq 'y') {
    print "temporary files will be removed from $scratchdir\n" if $verbose;
}
else {
    print "temporary files will not be removed from $scratchdir\n" if $verbose;
}

#----------------------------------------------------

#CHecks----------------------------------------------
print "Checks\n" if $verbose;

#CHeck-scratch-dir--
print "Checking scratch directory\n" if $verbose;

unless (-w "$scratchdir"){
    die "Unable to write to scratch directory $scratchdir\n";
}
else {
    print "Scratch directory checking finished\n" if $verbose;
}

#CHeck-scripts-dir--

my $fscript = "$scriptdir/fasta_checker.pl";
my $ascript = "$scriptdir/assignment.pl";
my $rscript = "$scriptdir/remove_overlaps.pl";

print "Checking scripts directory\n" if $verbose;

if (!-e $fscript) {
    die "$fscript not found\n";
}
elsif (!-x $fscript) {
    die "$fscript not an executable\n";
}

if (!-e $ascript) {
    die "$ascript not found\n";
}
elsif (!-x $ascript) {
    die "$ascript not an executable\n";
}


if (!-e $rscript) {
    die "$rscript not found\n";
}
elsif (!-x $rscript) {
    die "$rscript not an executable\n";
}

print "Scripts checking finished\n" if $verbose;

#
# output files
#
my $fastaout = "$scratchdir/$unique.fa";
my $logfile  = "$scratchdir/$unique.log";
my $resfile  = "$scratchdir/$unique.res";
my $assfile  = "$scratchdir/$unique.ass";
my $remfile  = "$scratchdir/$unique.rem";

push @files, $fastaout, $logfile, $resfile, $assfile, $remfile;

print "Checking and fixing sanity and formatting of fasta file\n" if $verbose;
system ("$fscript $queryfile > $fastaout");

unless (open(FASTASEQS, "< $queryfile")) {
    cleanup(\@files);
    die "could not read input file $queryfile: $!\n";
}
  
while (<FASTASEQS>) {
    if ($_ =~ /^>/) {
	$seqs++;
    }
}
close FASTASEQS;

if ($seqs == 0){
    cleanup(\@files);
    die "No query sequences found to run\n";
}
else{
    print "query file checking done: $seqs sequences to run\n" if $verbose;
    $queryfile = $fastaout;
}

#Check-hmm-prog-----
print "Checking HMM program\n" if $verbose;
$flag=0;

if ($hmm =~ /hmmscore$/) {
    $flag=1;
    $modsuffix='mod';
    $outsuffix='mstat';
}
elsif ($hmm =~ /hmmsearch$/) {
    $flag=1;
    $modsuffix='hmm';
    $outsuffix='res';
}
elsif ($hmm =~ /hmmpfam$/) {
    $flag=1;
    $modsuffix='hmm';
    $libsinglefile = 1;
    $outsuffix='res';
}

if ($flag == 0) {
    cleanup(\@files);
    die "HMM program '$hmm' does not look good\n";
}
else{
    print "HMM program checking done\n" if $verbose;
}

#Check-models-------
print "Checking models\n" if $verbose;

if ($libsinglefile) {
    # MM new: model library as a single big HMMER file
    if (not -e $modelsdir ) {
	cleanup(\@files);
	die "The model library file $modelsdir doesn't exist.\n";
    }

    my $re;

    if ($modelsacc && -e $modelsacc) {
	print "Using $modelsacc to read the ACC fields (faster)\n" if $verbose;
	$re = '(\d+)';
    }
    else {
	$modelsacc = $modelsdir;
	print "Using $modelsdir to read the ACC fields (slower)\n" if $verbose;
	$re = '^ACC\s+(\S+)';
    }

    unless (open(MODELS, "< $modelsacc")) {
	cleanup(\@files);
	die "could not read models file $modelsacc: $!\n";
    }

    while (<MODELS>) {
	if ($_ =~ /$re/) {
	    $models++;
	    push @mods, $1;
	}
    }
    close MODELS;

    if ($models == 0) {
	cleanup(\@files);
	die "No SUPERFAMILY models found in the file $modelsdir.\n";
    }
}
else {
    my $dh = DirHandle->new($modelsdir);
    unless ($dh) {
	cleanup(\@files);
	die "can't read directory $modelsdir: $!";
    }

    my @modelfiles = $dh->read();
    my $modelfile;

    foreach $modelfile (@modelfiles) {
        if ($modelfile =~ /(\d+)\.$modsuffix$/) {
	    $models++;
	    push @mods,$1;
        }
    }

    undef @modelfiles;  # save memory

    if ($models == 0){
	cleanup(\@files);
	die "No SUPERFAMILY models found in $modelsdir\nOnly models with suffix .$modsuffix accepted with HMM program $hmm\n";
    }
}

print "models checking done: $models models\n" if $verbose;

#Check-models-table-
print "Checking table of models\n" if $verbose;

unless (open(MODTAB, "<$modelstab")) {
    cleanup(\@files);
    die "Unable to read $modelstab: $!\n";
}

while (<MODTAB>){
    if (/^(\S+)\t(\S+)\t\S+\t\S+\t([^\n]+)/){
	$sf{$1}=$2;
	$des{$1}=$3;
    }
}
close MODTAB;

$flag=0;

foreach $mod (@mods){
    unless (exists($sf{$mod})){
	$flag=1;
	print STDERR "model: $mod not found in $modelstab\n";
    }
}

if ($flag == 1){
    cleanup(\@files);
    die "Some models not accounted for in models table: $modelstab\n";
}
else {
    print "Models table checking done\n" if $verbose;
}

#Check-evalue-threshold
print "Checking E-value threshold\n" if $verbose;

if ($threshold < 1.0e-7) {
    print "WARNING: E-value threshold is very conservative, expect to miss several true hits\n" if $verbose;
}
elsif ($threshold < 1.0e-4) {
    print "WARNING: E-value threshold is conservative, expect to miss some true hits\n" if $verbose;
}
elsif ($threshold < 0.01) {
    print "E-value threshold is acceptable, but 0.02 is recommended\n" if $verbose;
}
elsif ($threshold >= 0.01 and $threshold <= 0.02) {
    print "E-value threshold is acceptable\n" if $verbose;
}
elsif ($threshold < 0.1) {
    print "WARNING: E-value threshold is high, expect some false hits\n" if $verbose;
}
elsif ($threshold < 5) {
    print "WARNING: E-value threshold is very high, expect lots of false hits\n" if $verbose;
}
else {
    cleanup(\@files);
    die "E-value threshold is too high, now come on, be reasonable!\nTry 0.02\n";
}

print "E-value threshold checking done\n" if $verbose;
print "\n" if $verbose;

#------------------------------------------------------------

#RUN-HMMs----------------------------------------------------
print "Commencing running of HMMs " if $verbose;

if ($libsinglefile) {
    print "(sorry, no progress indicator for hmmpfam)\n" if $verbose;
    print "[ Please be patient, the run will take several minutes at the very least! ]\n" if $verbose;
}
else {
    print "(. per 100 models)\n" if $verbose;
}

open SAVEERR, ">&STDERR";
unless (open(STDERR, ">$logfile")) {
    cleanup(\@files);
    die "can't redirect standard error\n";
}
    

if ($libsinglefile) {
    system("$hmm --cpu $cpu -E $threshold -Z $models $modelsdir $queryfile > $resfile");
}
else {
    foreach $mod (@mods){
	if ($hmm =~ /hmmscore/){
	    system("$hmm $scratchdir/$unique\_$mod -i $modelsdir/$mod.$modsuffix -select_mdalign 8 -mdEmax $threshold -db $queryfile -dbsize $models  -sw 2");
	    system ("rm $unique\_$mod.dist");
	}
	elsif ($hmm =~ /hmmsearch/){
	    system("$hmm --cpu $cpu -E $threshold -Z $models $modelsdir/$mod.$modsuffix $queryfile > $scratchdir/$unique\_$mod.res");
	}
	$count++;
	if ($count == 100*int($count/100)){
	    print '.' if $verbose;
	}
    }
}
close STDERR;
open STDERR , ">&SAVEERR";
print "\nHMM runs complete\n\n" if $verbose;
#------------------------------------------------------------

#CHeck-HMM-runs---------------------------------------------
print "Checking HMM runs completed\n" if $verbose;

if ($libsinglefile) {
    if ( not -e "$resfile" ) {
	cleanup(\@files);
	die "Runs finished, but the result file $resfile doesn't exist!\n";
    }
  
    if ( -z "$resfile" ) {
	cleanup(\@files);
	die "Runs finished, but the result file $resfile has zero size!\n";
    }
  
    unless(`tail -1 $resfile` =~ /\/\//){
	#cleanup(\@files);
	die "Runs finished, but the result file $resfile is incomplete!\n";
    }
}
else {
    my $dh = DirHandle->new($scratchdir);
    unless ($dh) {
	cleanup(\@files);
	die "can't read directory $scratchdir: $!";
    }

    my @results = $dh->read();
    my $result;

    foreach $result (@results) {
        if ($result =~ /_(\d+)\.$outsuffix$/) {
	    $mstat{$1} = 1;
        }
    }

    undef @results;  # save memory

    $flag=0;
    foreach $mod (@mods){
	if (exists($mstat{$mod})){
	    unless ($mstat{$mod} > 0){
		print STDERR "model $mod not completed successfully\n" ;
		$flag=1;
	    }
	}
	else {
	    print STDERR "model $mod not completed successfully\n" ;
	    $flag=1;
	}
    }

    if ($flag == 1) {
	cleanup(\@files);
	die "all runs not completed successfully\n";
    }
    else {
	print "HMM runs checked, all completed successfully\n\n" if $verbose;
    }
}
#------------------------------------------------------------

#Parse-HITS-from-models--------------------------------------
print "parsing hits from models\n" if $verbose ;
print "running assignment program\n" if $verbose;

if ($outsuffix eq 'mstat') {
    system ("$ascript -f sam -e $threshold -v -db $queryfile -dir $scratchdir -out $scratchdir/$unique -p $unique\_");
}
elsif ($outsuffix eq 'res') {
    if ($libsinglefile) { 	
	system ("$ascript -f hmmpfam -e $threshold -v -db $queryfile -dir $scratchdir -out $scratchdir/$unique -p $unique");
    }
    else {
	system ("$ascript -f hmmer -e $threshold -v -db $queryfile -dir $scratchdir -out $scratchdir/$unique -p $unique\_");
    }
}

print "parsing completed\n\n" if $verbose;

print "removing overlaps\n" if $verbose;

system("$rscript $assfile $remfile");

unless(rename($remfile, $assfile)){
    cleanup(\@files);
    die "Unable to rename file $remfile: $!";
}


print "adding SCOP information from models\n" if $verbose;

#If chunk implementation for iprscan v4.0, maybe change the open option to  ">> $output".
unless (open(ASS, "> $output")) {
    cleanup(\@files);
    die "could not write file $output: $!\n";
}

unless (open(AS, "< $assfile")) {
    cleanup(\@files);
    die "could not read file $assfile: $!\n";
}

while (<AS>){
    if (/^(\S*)\t(\S*)\t(\S*)\t(\S*)\t(\S*)\t(\S*)$/){
	#EQ was print ASS "$1\t$sf{$2}\t$2\t$3\t$4\t$5\t$6\t$des{$2}\n"; #Added '//' at the end to give a separation delimiter for results.Easier to parse.
	print ASS "$1\t$sf{$2}\t$2\t$3\t$4\t$5\t$6\t$des{$2}\n\/\/\n";
    }
}

close AS;
close ASS;


#------------------------------------------------------------

#Clean-up-files----------------------------------------------
if ($remove =~ /[yY]/){
    print "Cleaning up and removing files from $scratchdir\n" if $verbose;
    cleanup(\@files);
}
else {
    print "temporary files not deleted:\n" if $verbose;
    for (@files) {
	print "$_\n" if $verbose;
    }

    if ($scratchdir =~ /$unique$/) {
	print "temporary files not deleted in $scratchdir\n" if $verbose;
    }
}

#------------------------------------------------------------

#Bye-bye-----------------------------------------------------
print "SUPERFAMILY run $unique complete!\nSee $output for results\n\n" if $verbose;
print "Output is a tab-delimited table of domains, there could be more than one line per sequence and there may be no lines for any given sequence.\n
The columns are as follows:\n
SequenceID\tSCOPsuperfamilyID\tSUPERFAMILYmodelID\tMatchStart\tMatchEnd\tEvalueScore\tAlignmentToModel\tSCOPsuperfamilyDescription\n\/\/\n" if $verbose;
#------------------------------------------------------------

sub cleanup($) {
    my $files = shift;
    if ($files && ref($files) eq 'ARRAY') {
	for (@$files) {
	    print "removing $_\n" if $verbose;
	    unlink($_);
	}
    }

    if ($scratchdir =~ /$unique$/) {
	print "Removing temporary directory $scratchdir\n" if $verbose;
	rmtree($scratchdir);
    }

    print "all temporary files removed\n\n" if $verbose;	
}



