=head1 NAME

Dispatcher::Tool - base class for tools.

=head1 SYNOPSIS

  # examples here
  use Dispatcher::Tool;

=head1 DESCRIPTION

=head1 VERSIONS

$Id: Tool.pm,v 1.1.1.1 2005/08/18 13:18:26 hunter Exp $

Copyright (c) European Bioinformatics Institute 2002

=head1 AUTHORS / ACKNOWLEDGEMENTS

Ville Silventoinen <vsi@ebi.ac.uk>

=cut

package Dispatcher::Tool;

BEGIN{
  unless($ENV{IPRSCAN_HOME}){
    print STDERR "$0 : IPRSCAN_HOME environment variable not set!";
    exit();
  }
  unless($ENV{IPRSCAN_LIB}){
    $ENV{IPRSCAN_LIB} = "$ENV{IPRSCAN_HOME}/lib";
  }
}

use CGI qw(:standard :shotcuts);
use Data::Dumper;
use English;
use File::Basename;
use File::Path;
use File::Spec::Functions;
use FileHandle;
use IO::Scalar;
use Mail::Send;
use Sys::Hostname;
use URI::Escape;

use lib $ENV{IPRSCAN_LIB};
use Dispatcher::Config;
use strict;

#my $WGetzURL = 'http://srs.ebi.ac.uk/srsbin/cgi-bin/wgetz';

use constant MAX_INPUT_SEQUENCE_SIZE   => 10240000;  # 10 MB
use constant MAX_CONFIRM_SEQUENCE_SIZE => 10240;     # 10 kB

my @Months = ('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec');


=head2 new

 Description: Constructor allocates an anonymous hash, which
              is tied to the class.
 Arguments:   -
 Returns:     $self object

=cut

sub new() {
    my $class = shift;
    my $self = { };
    bless $self, $class;
    return $self;
}


=head2 getToolName

 Description: Returns tool name.
 Arguments:   -
 Returns:     Tool name.

=cut

sub getToolName() {
    my $self = shift;
    return $self->{tool};
}


=head2 setToolName

 Description: Sets tool name.
 Arguments:   $name    Tool name.
 Returns:     -

=cut

sub setToolName($) {
    my ($self, $name) = @_;
    $self->{tool} = $name;
}


=head2 getParam

 Description: Returns tool input parameter value.
 Arguments:   $key    Parameter key.
 Returns:     Value or empty string if key has no value.

=cut

sub getParam($) {
    my ($self, $key) = @_;
    return '' unless $key;
    my $params = $self->getParams();
    return '' unless $params && ref($params) eq 'HASH';
    return $params->{$key};
}


=head2 setParam

 Description: Sets tool input parameter value.
 Arguments:   $key      Parameter key.
              $value    Parameter value.
 Returns:     -

=cut

sub setParam($$) {
    my ($self, $key, $value) = @_;
    return unless $key;
    my $params = $self->getParams();
    return unless $params && ref($params) eq 'HASH';
    $params->{$key} = $value;
}


=head2 removeParam

 Description: Removes tool input parameter.
 Arguments:   $key      Parameter key.
 Returns:     -

=cut

sub removeParam($) {
    my ($self, $key) = @_;
    return unless $key;
    my $params = $self->getParams();
    return unless $params && ref($params) eq 'HASH';
    delete $params->{$key};
}


=head2 getParams

 Description: Returns tool's input parameters.
 Arguments:   -
 Returns:     Reference to a hash.

=cut

sub getParams($) {
    my $self = shift;
    return $self->{params};
}


=head2 setParams

 Description: Sets tool parameters.
 Arguments:   $params    Reference to a hash.
 Returns:     -

=cut

sub setParams($) {
    my ($self, $params) = @_;
    $self->{params} = $params;
}


=head2 getConfigValue

 Description: Returns tool configuration value.
 Arguments:   $key    Configuration key.
 Returns:     Value or empty string if key has no value.

=cut

sub getConfigValue($) {
    my ($self, $key) = @_;
    return '' unless $key;
    my $config = $self->{config};

    if ($config && ref($config) eq 'Dispatcher::Config') {
        my $value = $config->get($key);
        return $value if $value;
    }

    # try default configuration
    $config = $self->{defaultconfig};

    unless ($config) {
	$config = new Dispatcher::Config();
	my ($res, $msg) = $config->parse('tooldefault', 1);
	unless ($res) {
	    print STDERR "$0: could not parse tool default configuration: $msg\n";
	    return '';
	}
	$self->{defaultconfig} = $config;
    }

    return $self->{defaultconfig}->get($key);
}


=head2 setConfigValue

 Description: Sets tool configuration value.
 Arguments:   $key      Configuration key.
              $value    Parameter value.
 Returns:     -

=cut

sub setConfigValue($$) {
    my ($self, $key, $value) = @_;
    return unless $key;
    my $config = $self->{config};
    return unless $config && ref($config) eq 'Dispatcher::Config';
    $config->set($key, $value);
}


=head2 removeConfigValue

 Description: Removes tool configuration key-value pair.
 Arguments:   $key      Configuration key.
 Returns:     -

=cut

sub removeConfigValue($) {
    my ($self, $key) = @_;
    return unless $key;
    my $config = $self->{config};
    return unless $config && ref($config) eq 'Dispatcher::Config';
    $config->remove($key);
}


=head2 getConfig

 Description: Returns tool configuration.
 Arguments:   -
 Returns:     Reference to Dispatcher::Config object

=cut

sub getConfig() {
    my $self = shift;
    return $self->{config};
}


=head2 setConfig

 Description: Sets tool configuration.
 Arguments:   $config    Reference to Dispatcher::Config object.
 Returns:     -

=cut

sub setConfig($) {
    my ($self, $config) = @_;
    return unless $config && ref($config) eq 'Dispatcher::Config';
    $self->{config} = $config;
}

=head2 mapParamToDescription

 Description: Returns parameter description.
              NOTE: this method uses param2desc hash reference inside the object.
              Each tool that wishes to support descriptions should set it in the
              object constructor.
 Arguments:   $key    Parameter key.
 Returns:     Description or empty string.

=cut

sub mapParamToDescription($) {
    my ($self, $key) = @_;
    return '' unless $key;
    return '' unless $self->{param2desc};
    return '' unless ref($self->{param2desc}) eq 'HASH';
    return $self->{param2desc}->{$key};
}


=head2 parseConfig

 Description: Parses tool configuration and sets it inside the tool object.
              This is a convenience method for all tools derived from
              Dispatcher::Tool. It sets the YYYY, MM, DD, hh, mm and ss tags
              automatically by parsing the jobid.
 Arguments:   Same as Dispatcher::Config::parse, excluding the first argument.
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub parseConfig(;$$$$) {
    my ($self, $expand, $defaults, $checkuf, $file) = @_;
    return (0, "could not parse config, tool name not defined") unless $self->getToolName();

    my $jobid = $defaults->{jobid} if $defaults;
    if ($jobid && $jobid =~ /\-([\d]{4})(\d\d)(\d\d)\-(\d\d)(\d\d)(\d\d)/) {
        # set date and time because they may be used in configuration (paths and URLs)
	$defaults->{YYYY} = $1;
	$defaults->{MM} = sprintf("%.2d", $2);
	$defaults->{DD} = sprintf("%.2d", $3);
	$defaults->{hh} = sprintf("%.2d", $4);
	$defaults->{mm} = sprintf("%.2d", $5);
	$defaults->{ss} = sprintf("%.2d", $6);
    }

    my $config = new Dispatcher::Config();
    my ($res, $msg) = $config->parse($self->getToolName(), $expand, $defaults, $checkuf, $file);
    return (0, $msg) unless $res;

    $self->setConfig($config);

    # this is a perfect place to set the umask value for files
    my $umask = $self->getConfigValue('umask');
    if ($umask && $umask =~ /^\d\d\d$/) {
        umask(oct(sprintf("%.d", $umask)));
    }

    return (1, '');
}


=head2 isTool

 Description: Validates tool name.
 Arguments:   $name    Tool name.
 Returns:     1 on success
              0 on failure

=cut

sub isTool($) {
    my ($self, $name) = @_;
    return 0 unless $name;
    return 1;
}


=head2 createConfirmationPage

 Description: Creates an HTML job confirmation page.

              NOTE 1: HTTP headers must be added by the caller, because it is
              impossible to know when they should be added (caller may have
              started the page and this call finishes it).

 Arguments:   $in        Input sequence (optional).
              $out       Output (default: stdout).
              $top       Page top (default: getHtmlTop).
              $bottom    Page bottom (default: getHtmlBottom).

 Returns:     1, ''  on success
              0, msg on failure

=cut

sub createConfirmationPage(;$$$$) {
    my ($self, $in, $out, $top, $bottom) = @_;
    $out = *STDOUT unless $out;

    my $params = $self->getParams() || { };
    my ($res, $ifh, $ofh, $iclose, $oclose, $tmpfile);

    if ($in) {
	($res, $iclose, $ifh) = $self->getReadHandle($in);
	return (0, $ifh) unless $res;
    }

    if ($self->isFileHandle($out)) {
	($res, $oclose, $ofh) = $self->getWriteHandle($out);
    }
    else {
        $tmpfile = "$out.tmp";
	($res, $oclose, $ofh) = $self->getWriteHandle($tmpfile);
    }

    unless ($res) {
        close $ifh if $iclose;
        return (0, $ofh);
    }

    if ($top) {
        print $ofh $top;
    }
    else {
	print $ofh $self->getHtmlTop();
    }

    print $ofh <<ENDHTML;
    <span class="plargebold">Thanks for your submission.</span>
    <br><br>
    <span class="text">
    An email will be sent to you after your search has started
    containing the submission parameters or any errors that may
    have been encountered.
    When the job finishes, you should receive an email containing
ENDHTML

    if ($self->getConfigValue('mailresult')) {
        print $ofh "the results.";
    }
    else {
        print $ofh "link to the result page.";
    }

    print $ofh <<ENDHTML;
    </span>
    <br><br>
    <div align="center">
    <table width="460" cellSpacing="0" cellPadding="4" border="1" bordercolor="#999999" bgcolor="#eeeeee" class="tabletop">
    <tr class="tablebody">
      <td colspan="2" align="center">
      <span class="pbold">SUBMISSION PARAMETERS</span>
      </td>
    </tr>
ENDHTML

    my @confparams = split /,/, $self->getConfigValue('confparams');
    if (@confparams > 0) {
        my ($key, $value, $header);
        foreach $key (@confparams) {
            $key = $self->trim($key);
            $value = $params->{$key};
            if ($key && defined $value) {
                $header = $self->mapParamToDescription($key) || $key;
                print $ofh "<tr>\n<td nowrap><span class=\"pbold\">$header</span></td>\n";
                print $ofh "<td class=\"text\" nowrap>$value&nbsp;</td>\n</tr>\n";
            }
        }
    }

    if ($params->{warnings}) {
        print $ofh "<tr>\n<td nowrap><span class=\"pbold\">PLEASE NOTE!</span></td>\n";
        my $warnings = $params->{warnings};
        $warnings =~ s/\n/<br>/g;
	print $ofh "<td class=\"text\" nowrap>$warnings</td>\n</tr>\n";
    }

    if ($ifh && $self->getConfigValue('mailconfseq') &&	-s $ifh < MAX_CONFIRM_SEQUENCE_SIZE) {
        print $ofh <<ENDHTML;
        </table>
        <br>
        <table width="460" cellSpacing="0" cellPadding="4" border="1" bordercolor="#999999" bgcolor="#eeeeee" class="tabletop">
        <tr bgcolor="#cecfce">
          <td colspan="2" align="center"><span class="plargebold">SEQUENCE</span></td>
        </tr>
        <tr>
          <td colspan="2">
          <pre>
ENDHTML

        while (<$ifh>) {
            print $ofh $_;
	}

        print $ofh "</pre>\n</td>\n";
    }
    elsif ($params->{seqlen}) {
        my $header = $self->mapParamToDescription('seqlen') || 'Sequence length';
        print $ofh <<ENDHTML;
        <tr>
          <td nowrap><b>$header</b></td>
          <td class="text" nowrap>$params->{seqlen}</td>
ENDHTML
    }

    my $submiturl = $self->getConfigValue('submiturl');

    print $ofh <<ENDHTML;
        </tr>
        <form method="get" action="$submiturl">
        <tr bgcolor="#cecfce">
          <td colspan="2" align="right">
          <input type="submit" value="SUBMIT ANOTHER JOB">
          </td>
        </tr>
        </form>
        </table>
        </div>
ENDHTML

    if ($bottom) {
        print $ofh $bottom;
    }
    else {
	print $ofh $self->getHtmlBottom();
    }

    close $ifh if $iclose;
    close $ofh if $oclose;

    if ($tmpfile) {
        unless (rename($tmpfile, $out)) {
            return (0, "could not rename temporary file $tmpfile: $!");
        }
    }

    return (1, '');
}


=head2 createErrorPage

 Description: Creates error HTML page to stdout or file.
              NOTE 1: HTTP headers must be sent by the caller.
              NOTE 2: This method does not return errors, it is "best effort".

 Arguments:   $out       Output (default: stdout).
              $emsg      Error message string (optional).
              $efile     Error file or file handle (optional).
              $top       Page top (default: getHtmlTop).
              $bottom    Page bottom (default: getHtmlBottom).

 Returns:     -

=cut

sub createErrorPage(;$$$$$) {
    my ($self, $out, $emsg, $efile, $top, $bottom) = @_;
    $out = *STDOUT unless $out;

    my ($res, $ofh, $efh, $oclose, $eclose, $tmpfile);

    if ($efile && -s $efile) {
	($res, $eclose, $efh) = $self->getReadHandle($efile);
	return unless $res;
    }

    if ($self->isFileHandle($out)) {
	($res, $oclose, $ofh) = $self->getWriteHandle($out);
    }
    else {
        $tmpfile = "$out.tmp";
	($res, $oclose, $ofh) = $self->getWriteHandle($tmpfile);
    }

    unless ($res) {
        close $efh if $eclose;
	return;
    }

    if ($top) {
        print $ofh $top;
    }
    else {
	print $ofh $self->getHtmlTop();
    }

    my $header = $self->getConfigValue('htmlerror') || 'Error';

    print $ofh <<ENDHTML;
    <div align="center">
    <table width="100%" cellSpacing="0" cellPadding="5" border="1" bordercolor="#999999" class="tabletop">
    <tr class="tablebody">
      <td align="center"><span class="pbold">$header</span></td>
    </tr>
    <tr>
    <td class="text">
ENDHTML

    print $ofh ucfirst $emsg . "\n" if $emsg;

    if ($efh) {
        while (<$efh>) {
	    print $ofh $_;
	}
    }

    print $ofh "</td>\n</tr>\n";

    my $submiturl = $self->getConfigValue('submiturl');
    if ($submiturl) {
        print $ofh <<ENDHTML;
        <tr>
        <td>
          <table border="0"><tr><td>
          <form method="GET" action="$submiturl">
          <input type="submit" value="SUBMIT ANOTHER JOB">
          </form>
          </td></tr></table>
        </td>
        </tr>
ENDHTML
    }
   
    print $ofh "</table></div>\n";

    if ($bottom) {
        print $ofh $bottom;
    }
    else {
        print $ofh $self->getHtmlBottom();
    }

    close $efh if $eclose;
    close $ofh if $oclose;

    if ($tmpfile) {
        unless (rename($tmpfile, $out)) {
            return (0, "could not rename temporary file $tmpfile: $!");
        }
    }
}


=head2 createPollPage

 Description: Creates HTML page that polls the result page.
 Arguments:   $url    URL to be polled.
              $out    Output (default: stdout).
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub createPollPage($;$) {
    my ($self, $url, $out) = @_;
    return (0, 'no URL for poll page defined') unless $url;
    $out = *STDOUT unless $out;

    my ($res, $ofh, $oclose, $tmpfile);

    if ($self->isFileHandle($out)) {
	($res, $oclose, $ofh) = $self->getWriteHandle($out);
    }
    else {
        $tmpfile = "$out.tmp";
	($res, $oclose, $ofh) = $self->getWriteHandle($tmpfile);
    }

    return (0, $ofh) unless $res;

    my @parts = split('/', $url);
    my $title = $parts[$#parts];
    my $imagepath = $self->getConfigValue('imgbase');

    print $ofh <<ENDHTML;
<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML//EN\">
<html>
<head>
<META HTTP-EQUIV="REFRESH" CONTENT="40;URL=$url">
<META HTTP-EQUIV="PRAGMA" CONTENT="NO-CACHE">
<script language="JavaScript1.2">
  var bookmarkurl="$url"
  var bookmarktitle="$title"
  function addbookmark(){
    if(document.all)
      window.external.AddFavorite(bookmarkurl,bookmarktitle)
  }
  function reLoad(){
    window.location='$url'
    setTimeout ("reLoad()",50000);
  }
</script>

<title>Job running: $url</title>
</head>
<body bgcolor="#FFFFFF" text="#000000" onload="setTimeout('reLoad()',10000)">
<table width="100%" border="0" cellspacing="0" cellpadding="0" height="100%">
  <tr>
    <td align="center" valign="middle">
      <table width="650" border="1" cellspacing="0" cellpadding="0" bordercolor="#999999">
        <tr>
          <td>
            <table cellspacing="0" cellpadding="0" border="0" bordercolor="#999999" bgcolor="#FFFFFF" width="100%">
              <tr bgcolor="#eeeeee">
                <td align="center">
                  <img src="$imagepath/results_ebitop.jpg" width="442" height="86">&nbsp;
                </td>
                <td valign="top">
                  <div align="right">
                  <img src="$imagepath/trans.gif" width="203" height="15">
                  <br>
                  <b><font face="Arial, Helvetica, sans-serif" color="#000000" size="2">
                  <nobr>Your job is currently running...</nobr><br></font></b>
                  <font face="Arial, Helvetica, sans-serif" color="#000000" size="2">
                  ...please be patient</font>
                  </div>
                </td>
                <td valign="middle">
                  <img src="$imagepath/trans.gif" width="5" height="10">
                </td>
              </tr>
            </table>
            <table cellspacing="0" cellpadding="0" border="0" bordercolor="#999999" bgcolor="#FFFFFF" width="650">
              <tr>
                <td align="center" valign="top" colspan="2" nowrap>
                  <font size="3" face="Arial,Helvetica,Univers,Zurich BT"><b>
                  <img src="$imagepath/trans.gif" width="620" height="10">
                  <br>
                  </b></font>
                  <img src="$imagepath/iprs.png"><br>
                  <font face="Arial, Helvetica, sans-serif" size="3" color="#000000">
                  <b>The results of your job will appear in this browser window.</b></font><br>
                  <br>
                  <font face="Arial, Helvetica, sans-serif" size="2">Your Job output:
                  <a href="$url">$url</a></font>
                  <br><br>
                </td>
                <td align="center" width="10" rowspan="2">
                  <img src="$imagepath/trans.gif" width="10" height="50">
                </td>
                <td align="center" width="10" rowspan="2">
                <img src="$imagepath/red_ani.gif" width="10" height="263">
                </td>
                <td align="center" width="10" rowspan="2">
                  <img src="$imagepath/trans.gif" width="10" height="50">
                </td>
              </tr>
              <tr>
                <td width="600" valign="top" bgcolor="#eeeeee">
                  <div align="center">
                  <font size="3" face="Arial,Helvetica,Univers,Zurich BT"><b>
                  <img src="$imagepath/trans.gif" width="600" height="10">
                  <br>
                  </b></font>
                  <font face="Arial, Helvetica, sans-serif" size="2">
                  <b>Please Note the Following:</b></font>
                  </div>
                  <ul>
                    <li><font face="Arial, Helvetica, sans-serif" size="2">
                    You may press Shift+Refresh or Reload on your browser at any time to check
                    if results are ready. Should this window go blank please press the
                    Shift+Refresh or Reload button on your browser.</font></li>
                  </ul>
                  <ul>
                    <li><font face="Arial, Helvetica, sans-serif" size="2">
                    You may bookmark this page to view your results later if you wish.<br>
                    <b><font color="#000000">Netscape users:</font></b> Use Bookmark -
                    Add Bookmark or CTRL-D | Alt-K to bookmark this page.<br>
                    <b><font color="#000000">IE users:</font></b> Click ->
                    <a href="javascript:addbookmark()">BookMark</a> to bookmark this page.
                    </font></li>
                  </ul>
                </td>
                <td align="center" width="20" valign="top" bgcolor="#eeeeee">
                  <img src="$imagepath/trans.gif" width="20" height="10">
                </td>
              </tr>
            </table>
          </td>
        </tr>
      </table>
      <table width="650" border="0" cellspacing="0" cellpadding="0" align="center">
        <tr>
          <td>
            <img src="$imagepath/trans.gif" width="50" height="5">
          </td>
        </tr>
        <tr>
          <td>
            <div align="center">
            <font face="Arial, Helvetica, sans-serif" size="1" color="#999999\">
            &copy; Copyright European Bioinformatics Institute 2002. All Rights and
            Trademarks Reserved.
            </font></div>
          </td>
        </tr>
      </table>
     </td>
  </tr>
</table>
</body>
</html>
ENDHTML

    close $ofh if $oclose;

    if ($tmpfile) {
        unless (rename($tmpfile, $out)) {
            return (0, "could not rename temporary file $tmpfile: $!");
        }
    }

    return (1, '');
}


=head2 createConfirmationMail

 Description: Creates confirmation mail to stdout or file.
 Arguments:   $in     Input sequence (optional).
              $out    Output (default: stdout).
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub createConfirmationMail(;$$) {
    my ($self, $in, $out) = @_;
    $out = *STDOUT unless $out;

    my $params = $self->getParams();
    my ($res, $ifh, $ofh, $iclose, $oclose, $tmpfile);

    if ($in) {
	($res, $iclose, $ifh) = $self->getReadHandle($in);
	return (0, $ifh) unless $res;
    }

    ($res, $oclose, $ofh) = $self->getWriteHandle($out);
    unless ($res) {
        close $ifh if $iclose;
        return (0, $ofh);
    }

    my $cmsg = $self->getConfigValue('mailconfmsg');
    if ($cmsg && -r $cmsg) {
        unless (open(CMSG, "< $cmsg")) {
            close CMSG;
            close $ifh if $iclose;
            close $ofh if $oclose;
            return (0, "could not read file $cmsg: $!");
        }

        while (<CMSG>) {
            print $ofh $_;
        }
        close CMSG;

        print $ofh "\n";
    }

    my $mailsub = $params->{submission} eq 'email' ? 1 : 0;

    my @confparams = split /,/, $self->getConfigValue('confparams');
    if (@confparams > 0) {
        my ($key, $value, $header);
        foreach $key (@confparams) {
            $key = $self->trim($key);
            $value = $params->{$key};
            if ($key && defined $value) {
                if ($mailsub) {
                    # use mail commands
                    $header = $self->mapParamToMailCommand($key) || $key;
                    print $ofh sprintf("%s %s\n", $header, $value);
                }
                else {
                    # use descriptions
                    $header = $self->mapParamToDescription($key) || $key;
                    print $ofh sprintf("%-31s%s\n", $header, $value);
                }
            }
        }
    }

    if ($ifh) {
        # Include the input sequence when it is relatively small.
        if ($self->getConfigValue('mailconfseq') && -s $ifh < MAX_CONFIRM_SEQUENCE_SIZE) {
            if ($mailsub) {
                print $ofh "SEQ\n";
            }
            else {
                print $ofh "Sequence\n\n";
            }

	    my $line = <$ifh>;
	    $line = <$ifh> if $mailsub;  # skip fasta header

	    while ($line) { 
		print $ofh $line;
		$line = <$ifh>;
	    }

            print $ofh "END\n" if $mailsub;
        }
        elsif ($mailsub) {
            print $ofh sprintf("SEQLEN %s\n", $params->{seqlen});
        }
        else {
	    my $header = $self->mapParamToDescription('seqlen') || 'Sequence length';
            print $ofh sprintf("%-31s%s\n", $header, $params->{seqlen});
        }

	close $ifh if $iclose;
    }

    close $ofh if $oclose;

    return (1, '');
}


=head2 hasAccess

 Description: Checks whether remote host is allowed to access the tool or more
              specifically the group that the tool belongs to (like EPO tools).
              Host name is retrieved from environment variables REMOTE_HOST or
              REMOTE_ADDR.
 Arguments:   $group    Group name (optional).
 Returns:     1, host on access allowed
              0, msg  on access denied

=cut

sub hasAccess(;$) {
    my ($self, $group) = @_;

    my $host = $ENV{'REMOTE_HOST'} || $ENV{'REMOTE_ADDR'};

    $group = $self->getConfigValue('usergroup') unless $group;
    return (1, $host) unless $group;

    my $msg;

    unless ($host) {
	 $msg  = "Dear user, you are not allowed to access these pages. ";
	 $msg .= "Please refer to the following list of public services ";
	 $msg .= "from the EBI: <a href=\"http://www.ebi.ac.uk/Tools/\">";
	 $msg .= "http://www.ebi.ac.uk/Tools/";
	 return (0, $msg);
    }

    return(1, ''); 

## Uncomment following lines if you had host access list and comment line above(return(1, '');).
#
#    my $hosts = $Groups{$group}{hostlist};
#    for (@$hosts) {
#	  return (1, $host) if $host =~ /$_/;
#    }
#
#    $msg  = "Dear user, you are not allowed to access these pages from '$host'. ";
#    $msg .= "Please refer to the following list of public services ";
#    $msg .= "from the EBI: <a href=\"http://www.ebi.ac.uk/Tools/\">";
#    $msg .= "http://www.ebi.ac.uk/Tools/";
#    return (0, $msg);
}


=head2 parseRequest

 Description: Reads the request parameters to a hash table. Parameters must
              be given as a CGI query object. Uploaded files are written to
              disk with the configured paths uploadfile.1, uploadfile.2, etc.
 Arguments:   $query    CGI query object.
              $check    1: check parameters
                        0: no check
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub parseRequest($;$) {
    my ($self, $query, $check) = @_;
    return (0, "no input query given") unless $query;

    my $params = { };
    my @names = $query->param;
    my $nseq = 1;  # file handle counter

    for (@names) {
        my $key = $_;
	my $value = $query->param($key);
	my $fh = $query->upload($key);
        if ($fh) {
	    my $file = $self->getConfigValue('uploadfile.' . $nseq);
	    return (0, "uploadfile.$nseq not configured, please check tooldefault.conf") unless $file;
	    $nseq++;

	    unless (open(FILE, "> $file")) {
		close FILE;
		return (0, "could not write file $file: $!");
	    }

	    my ($res, $msg) = $self->safeWrite($fh, \*FILE, MAX_INPUT_SEQUENCE_SIZE);
	    close(FILE);
	    return (0, $msg) unless $res;
	}
        else {
	    my @values = $query->param($key);
	    my $value;

	    if (@values == 1) {
		$value = $values[0];
	    }
	    elsif (@values > 1) {
		$value = join('+', @values);
	    }

            $value = $self->trim($value) if defined $value;
	    $params->{$key} = $value if defined $value;
        }
    }

    $params->{submission} = 'cgi';
    $self->setParams($params);

    return $self->checkParams() if $check;
    return (1, '');
}


=head2 readParamFile

 Description: Reads parameters from file (handle) and creates a hash.
 Arguments:   $in        Input parameters.
              $params    Reference to a hash.
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub readParamFile($$) {
    my ($self, $in, $params) = @_;
    return (0, "could not read parameters, no parameter file given") unless $in;
    return (0, "could not read parameters, no parameter hash given") unless $params;

    my ($res, $iclose, $ifh) = $self->getReadHandle($in);
    return (0, $ifh) unless $res;

    while (<$ifh>) {
        chomp;
        my ($key, @values) = split /=/, $_;
        next unless $key;
        $key =~ s|^\s+||;  # no leading white
        $key =~ s|\s+$||;  # trailing white
        $params->{$key} = join('=', @values);

    }

    close $ifh if $iclose;

    return (1, '');
}


=head2 writeParamFile

 Description: Writes submission parameters to a file (handle).
 Arguments:   $out       Output parameters file.
              $params    Parameter hash.
 Returns:     1, path on success
              0, msg  on failure

=cut

sub writeParamFile($$) {
    my ($self, $out, $params) = @_;
    return (0, "parameter file not defined") unless $out;
    return (0, "parameter hash not defined") unless $params && ref($params) eq 'HASH';

    my ($res, $oclose, $ofh) = $self->getWriteHandle($out);
    return (0, $ofh) unless $res;

    my ($key, $value);
    foreach $key (keys %$params) {
        $value = $params->{$key};
        print $ofh "$key=$value\n" if defined $value;
    }
    close $ofh if $oclose;

    return (1, '');
}


=head2 getDate

 Description: Returns current date.
 Arguments:   -
 Returns:     yyyy, mm, dd

=cut

sub getDate() {
    my $self = shift;
    my ($sec, $min, $hour, $mday, $mon, $year, $wday) = localtime(time);
    $mon++;         # localtime starts months from 0
    $year += 1900;  # localtime starts years from 100
    return ($year, sprintf("%.2d", $mon), sprintf("%.2d", $mday));
}


=head2 getDateFromJobID

 Description: Returns date extracted from job ID.
 Arguments:   $jobid    Job identifier containing "-YYYYMMDD-",
 Returns:     yyyy, mm, dd

=cut

sub getDateFromJobID($) {
    my ($self, $jobid) = @_;

    if ($jobid =~ /\-([\d]{4})([\d]{2})([\d]{2})\-/) {
        return ($1, $2, $3);
    }

    return ('', '', '');
}


=head2 parseDate

 Description: Returns date in format YYYYMMDD (numerical).
 Arguments:   $date    Date in format in of the following formats:
                       DD-MMM-YYYY (01-Oct-2003)
                       DD-MMM-YY   (01-Oct-03)
                       DD-MM-YYYY  (01-10-2003)
                       DD-MM-YY    (01-10-03)
                       MM/DD/YYYY  (10/01/2003)
                       MM/DD/YY    (10/01/03)

 Returns:     YYYYMMDD

=cut

sub parseDate($) {
    my ($self, $date) = @_;
    return 0 unless $date;

    my ($day, $month, $year, $i);

    if ($date =~ /([\d]{1,2})-([a-zA-Z]{3})-([\d]{4})/ ||
        $date =~ /([\d]{1,2})-([a-zA-Z]{3})-([\d]{2})/)
    {
        # format DD-MMM-YYYY and DD-MMM-YY
        $day = $1;
        $month = $2;
        $year = $3;

        for($i=0; $i < @Months; $i++) {
            if (lc($Months[$i]) eq lc($2)) {
                $month = $i+1;
            }
        }
    }
    elsif ($date =~ /([\d]{1,2})-([\d]{1,2})-([\d]{4})/ ||
           $date =~ /([\d]{1,2})-([\d]{1,2})-([\d]{2})/)
    {
        # format DD-MM-YYYY or DD-MM-YY
        $day = $1;
        $month = $2;
        $year = $3;
    }
    elsif ($date =~ /([\d]{1,2})\.([\d]{1,2})\.([\d]{4})/ ||
           $date =~ /([\d]{1,2})\.([\d]{1,2})\.([\d]{2})/)
    {
        # format DD.MM.YYYY or DD.MM.YY
        $day = $1;
        $month = $2;
        $year = $3;
    }
    elsif ($date =~ /([\d]{1,2})\/([\d]{1,2})\/([\d]{4})/ ||
           $date =~ /([\d]{1,2})\/([\d]{1,2})\/([\d]{2})/)
    {
        # format MM/DD/YYYY or MM/DD/YY
        $month = $1;
        $day = $2;
        $year = $3;
    }

    # transform two digit year to four digits
    if ($year < 100) {
        # FIXME: this will break in 2030.
        my $prefix = ($year < 30) ? '20' : '19';
 	$year = sprintf("%s%.2d%", $prefix, $year);
    }

    return sprintf("%.4d%.2d%.2d", $year, $month, $day);
}


=head2 protect

 Description: Replaces unsafe characters in a string.
 Arguments:   $s    String.
 Returns:     string

=cut

sub protect($) {
    my ($self, $s) = @_;
    $s =~ s/[^-a-zA-Z0-9\.@\-\+]/_/go;
    return $s;
}


=head2 removeIllegalChars

 Description: Removes < > " ' characters from string.
 Arguments:   $s    String.
 Returns:     string

=cut

sub removeIllegalChars($) {
    my ($self, $s) = @_;
    return '' unless $s;
    $s =~ s/[<>"']//g;
    return $s;
}

=head2 systemCall

 Description: Execute a command as a system call
              (see Programming Perl system function for details).

 Arguments:   $cmd    Command (with args) to execute.

 Returns:     1, ''  on success
              0, msg on error

=cut

sub systemCall ($) {
    my ($self, $cmd) = @_;

    my $rc = 0xffff & system($cmd);
    my $msg;

    if ($rc == 0) {
        # command exited normally
    }
    elsif ($rc == 0xff00) {
        $msg = "$cmd: $!";
    } 
    elsif (($rc & 0xff) == 0) {
        $rc >>= 8;
        $msg = "$cmd: exit status $rc";
    }  
    else {
        $msg = "$cmd: ";
        if ($rc &   0x80) {
            $rc &= ~0x80;
            $msg .= "coredump from ";
        } 
        $msg .= "signal $rc";
    }
    return ($rc == 0, $msg);
}


=head2 sendMail

 Description: Sends mail message. If both message body string and file are
              given, string is set first, then file.
 Arguments:   $to          To field.
              $subject     Subject field.
              $bodys       Body passed in a string (optional).
              $bodyf       Body passed in a file or file handle (optional).
              $from        From field (optional).
              $errorsTo    Errors-To field (optional).
              $chunk       Split file to N line chunks (optional). This adds
                           " (part N)" strings also to subject lines.
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub sendMail($$;$$$$$) {
    my ($self, $to, $subject, $bodys, $bodyf, $from, $errorsTo, $chunk) = @_;
    return (0, "no destination address") unless $to;
    return (0, "no subject")             unless $subject;
    return (0, "no message body")        unless $bodys || $bodyf;

    my $msg = new Mail::Send;
    $msg->set('To', $to);
    $msg->set('Subject', $subject);
    $msg->set('From', $from) if $from;
    $msg->set('Errors-To', $errorsTo) if $errorsTo;

    my ($res, $ifh, $iclose);

    if ($bodyf) {
	($res, $iclose, $ifh) = $self->getReadHandle($bodyf);
        return (0, $ifh) unless $res;
    }

    my $mfh = $msg->open;
    print $mfh $bodys if $bodys;

    if ($ifh) {
	my $line = 1;
        my $part = 1;

	if ($chunk && $chunk < 1000) {
	    $chunk = 1000;
	}

        while (<$ifh>) {
            print $mfh $_;
            $line++;

	    if ($chunk && $line > $chunk) {
		# complete and send this part
		$mfh->close;

                # prepare next part
                $line = 1;
                $part++;

		$msg = new Mail::Send;
		$msg->set('To', $to);
		$msg->set('Subject', "$subject (part $part)");
		$msg->set('From', $from) if $from;
		$msg->set('Errors-To', $errorsTo) if $errorsTo;
		$mfh = $msg->open;
            }
        }
    }

    close $ifh if $iclose;

    # complete and send the message
    $mfh->close;

    return (1, '');
}


=head2 checkDirectory

 Description: Checks if given directory exists and tries to create it.
 Arguments:   $dir     Directory.
              $mode    Numeric mode (optional).
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub checkDirectory($;$) {
    my ($self, $dir, $mode) = @_;
    return (0, 'no directory given') unless $dir;

    $mode = $self->getConfigValue('dirmode') unless $mode;

    unless (-d $dir) {
        unless (mkpath($dir, 0, oct(sprintf("%.4d", $mode)))) {
            return (0, "could not create directory $dir: $!");
        }
    }

    return (1, '');
}


=head2 generateJobID

 Description: Generates new job identifier.
 Arguments:   $group    Tool group, e.g., 'fasta' for all fastas.
 Returns:     jobid in format "toolgroup-YYYYMMDD-rand"

=cut

sub generateJobID($) {
    my ($self, $group) = @_;
    $group = 'job' unless $group;
    srand(time + $PROCESS_ID);
    my $n = int(rand(100000000));
    return sprintf("%s-%.4d%.2d%.2d-%.8d", $group, $self->getDate(), $n);
}


=head2 checkJobID

 Description: Checks if given job ID is in correct format.
 Arguments:   $jobid    Job identifier.
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub checkJobID($) {
    my ($self, $jobid) = @_;
    return (0, 'no job identifier given') unless $jobid;

    if ($jobid =~ /^[A-Za-z]+-[\d]{8}-[\d]{8}$/) {
        return (1, '');
    }

    my $err = "Unrecognized job identifier '$jobid'.\n\n";
    $err .= "Supported format is TOOLNAME-YYYYMMDD-8DIGITS, where ";
    $err .= "TOOLNAME is the tool name, YYYYMMDD is the submission date ";
    $err .= "and 8DIGITS is a 8 digit number.\n\n";

    return (0, $err);
}


=head2 trim

 Description: Trims leading and trailing whitespace.
 Arguments:   $s    String.
 Returns:     trimmed string

=cut

sub trim($) {
    my ($self, $s) = @_;
    return '' unless defined $s;
    $s =~ s/^\s+//;
    $s =~ s/\s+$//;
    return $s;
}


=head2 errorExit

 Description: Reports errors and exits.
 Arguments:   $error    Error message, OR
              $efile    Error file.
              $html     1: create HTML error page
                        0: email error to user
 Returns:     -

=cut

sub errorExit(;$$$) {
    my ($self, $error, $efile, $html) = @_;
    my $tname = $self->getToolName() || $self->getParam('tool');
    my $jobid = $self->getConfigValue('jobid') || $self->getParam('jobid');

    exit unless $error || $efile;

    my $params = $self->getParams() || { };

    my $admin   = $self->getConfigValue('adminaddr');
    my $subject = $self->getConfigValue('mailerrorsubject') || "Error in $tname job $jobid";
    $subject =~ s/\$jobid/$jobid/;
    $error = '' unless defined $error;

    if ($efile && -f $efile && -s $efile > 0) {
	if (open(ERRORS, "< $efile")) {
	    local $INPUT_RECORD_SEPARATOR;  # undefine newline limit
	    $error .= <ERRORS>;
	}
	close ERRORS;
    }

    my $date = localtime;
    my $report = sprintf("Tool:  %s\nJobID: %s\nTime:  %s\nError:\n\n%s\n\n",
			 $tname || 'unknown' , $jobid || 'not available', $date, $error);

    if ($params && keys %$params > 0) {
        $report .= "Job submission parameters:\n";
        my ($key, $value);
        foreach $key (sort keys %$params) {
            $value = $params->{$key} || '';
            if ($key && $value) {
                $value =~ s/\n/\\/g;
                $value =~ s/\r/\\/g;
                $report .= sprintf("%-16s%-.200s\n", $key, $value);
            }
        }
        $report .= "\nNOTE: the newlines have been replaced with backslash.\n\n";
    }

    # always report to administrator
    my ($res, $msg) = $self->sendMail($admin, $subject, $report, undef,
				      $self->getConfigValue('mailfromaddr'),
				      $self->getConfigValue('mailerrorto'));
    unless ($res) {
	print STDERR "$0: failed to send error to administrator: $msg\n";
	print STDERR "$report\n";
    } 

    if ($html) {
        print header(-type => 'text/html');
        $self->createErrorPage(undef, $error);
    }
    elsif ($params->{email}) {
        ($res, $msg) = $self->sendMail($params->{email}, $subject, $error, undef,
				       $self->getConfigValue('mailfromaddr'),
				       $self->getConfigValue('mailerrorto'));
	unless ($res) {
	    print STDERR "$0: failed to send error to user: $msg\n";
	    print STDERR "$report\n";
	}
    }

    # check the temporary upload files and delete them
    my $upload = 1;
    my $upfile = $self->getConfigValue("uploadfile.$upload");
    while (-e $upfile) {
	unlink($upfile);
	$upload++;
	$upfile = $self->getConfigValue("uploadfile.$upload");
    }

    my $config = $self->getConfigValue('errorlogconf');

    exit(1);
}


=head2 sendForbidden

 Description: Sends HTTP 403 Forbidden.
              NOTE! Does not exit!
 Arguments:   $error     Error message (optional).
 Returns:     -

=cut

sub sendForbidden(;$) {
    my ($self, $error) = @_;
    print <<END_OF_ERROR
Status: 403 Forbidden
Content-type: text/html

<html>
<head>
<title>403 Forbidden</title>
</head>
<body>
<h1>403 Forbidden</h1>
<p>$error</p>
</body>
</html>
END_OF_ERROR
}


=head2 findFile

 Description: Tries to find a file. This is a convenience method for situations
              when job files can be in different directories. At the moment
              following work directories set in the tool configuration are
              searched: workdir, aworkdir and sworkdir.
 Arguments:   $file    filename, path or key in tool configuration.
 Returns:     Full path or $file.

=cut

sub findFile($) {
    my ($self, $file) = @_;
    return '' unless $file;
    return $file if -f $file;

    my $v = $self->getConfigValue($file);
    $file = $v if $v;

    my ($dir, @dirs);

    if ($self->getConfigValue('workdir')) {
	push @dirs, $self->getConfigValue('workdir');
    }

    if ($self->getConfigValue('aworkdir')) {
	push @dirs, $self->getConfigValue('aworkdir');
    }

    if ($self->getConfigValue('sworkdir')) {
	push @dirs, $self->getConfigValue('sworkdir');
    }

    my ($name, $path, $ext) = fileparse($file);

    foreach $dir (@dirs) {
        $path = catfile($dir, $name);
        return $path if -f $path;
    }

    # give up!    
    return $file;
}


=head2 safeWrite

 Description: Reads from a file handle and writes to another file handle. Checks
              the number of bytes written and stops writing if the given limit is
              exceeded.
              NOTE! Caller is responsible for closing the file handles!
 Arguments:   $readFH     Read file handle.
              $writeFH    Write file handle.
              $limit      Limit in bytes (optional).
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub safeWrite($$;$) {
    my ($self, $readFH, $writeFH, $limit) = @_;
    return (0, "no file handle for reading") unless $readFH;
    return (0, "no file handle for writing") unless $writeFH;
    $limit = 10240000 unless $limit || $limit < 1;  # 10MB by default

    my $blksize = (stat $readFH)[11] || 16384;  # preferred block size?
    $blksize = $limit if $limit < $blksize;

    my ($len, $buf, $written);
    my $total = 0;

    while ($len = sysread($readFH, $buf, $blksize)) {
        unless (defined $len) {
            next if $! =~ /^Interrupted/;
            return (0, "system read error: $!");
        }

        my $offset = 0;
        while ($len) {          # Handle partial writes.
            $written = syswrite($writeFH, $buf, $len, $offset);
            unless (defined $written) {
                return (0, "system write error: $!");
            }
            $len -= $written;
            $offset += $written;
            $total += $written;

#Uncomment if you want to have a size limitation for the input sequence file.
#            if ($total > $limit) {
#                return (0, "size exceeds limit $limit bytes");
#            }
        }
    }

    return (1, '');
}


=head2 getUniqueID

 Description: Generates unique identifier.
 Arguments:   -
 Returns:     id string

=cut

sub getUniqueID($) {
    my $self = shift;
    my ($sec, $min, $hour, $mday, $mon, $year, $wday) = localtime(time);
    $mon++;         # localtime starts months from 0
    $year += 1900;  # localtime starts years from 100
    my $time = sprintf("%d%.2d%.2d%.2d%.2d%.2d",
                       $year, $mon, $mday, $hour, $min, $sec);
    my $rand = sprintf("%.6d", int(rand(1000000)));
    return hostname() . "-$time-$rand";
}


=head2 getSequenceInfo

 Description: Returns information about sequence. The sequence must be
              passed in raw or fasta format.

 Arguments:   $in    Input sequence file or file handle.

 Returns:     1, dna|rna|protein, length
              0, msg

=cut

sub getSequenceInfo($) {
    my ($self, $in) = @_;
    return (0, "no sequence given") unless $in;

    my ($res, $iclose, $ifh) = $self->getReadHandle($in);
    return (0, $ifh) unless $res;

    # follows the same logic as in Bio::PrimarySeq _guess_alphabet,
    # except we read the sequence line by line to save memory
    my ($u, $atgc, $type);
    my $total = 0;

    while (<$ifh>) {
        $_ = $self->trim($_);
        next if $_ =~ /^>/;  # skip fasta header line
        $u     += ($_ =~ tr/Uu//);
        $atgc  += ($_ =~ tr/ATGCNatgcn//);
        $total += length($_);
    }

    close $ifh if $iclose;
    if ($total == 0) {
        my $msg = "Could not determine sequence type, sequence length is zero.\n\n";
	if ($self->getParam('submission') eq 'email') {
	    $msg .= "If your sequence is in Fasta format, please make sure you have a ";
	    $msg .= "newline after the > header line. For example:\n\n";
            $msg .= ">Header\n";
            $msg .= "MRCSISLVLGLLALEVALARNLQEHVFNSVQSMCSDDSFSEDTECINCQTNEECAQNDMC\n";
            $msg .= "CPSSCGRSCKTPVNIEVQKAGRCPWNPIQMIAAGPCPKDNPCSIDSDCSGTMKCCKNGCI\n";
            $msg .= "MSCMDPEPKSPTVISFQ\n";
        }
        else {
            $msg .= "<br><br>";
	    $msg .= "If your sequence is in Fasta format, please make sure you have a ";
	    $msg .= "newline after the > header line. For example:<br>\n<pre>";
            $msg .= ">Header\n";
            $msg .= "MRCSISLVLGLLALEVALARNLQEHVFNSVQSMCSDDSFSEDTECINCQTNEECAQNDMC\n";
            $msg .= "MSCMDPEPKSPTVISFQ\n</pre>";
        }
        return (0, $msg);
    }

    if(($atgc / $total) > 0.85) {
        $type = 'dna';
    }
    elsif((($atgc + $u) / $total) > 0.85) {
        $type = 'rna';
    }
    else {
       $type = 'protein';
    }

    return (1, $type, $total);
}


=head2 checkBoolean

 Description: Checks boolean.
 Arguments:   $value      Value string (optional).
              $default    Default value (optional).
 Returns:     value

=cut

sub checkBoolean($;$) {
    my ($self, $value, $default) = @_;

    return $default unless $value;

    unless ($value =~ /^0$|^1$|^y$|^n$|^yes$|^no$|^t$|^f$|^true$|^false$|^on$|^off$/i) {
        return $default;
    }

    return $value;
}


=head2 checkInteger

 Description: Checks integer.
 Arguments:   $value      Value string (optional).
              $default    Default value (optional).
              $lower      Lower limit (optional).
              $upper      Upper limit (optional).
 Returns:     1, value on success
              0, msg   on failure

=cut

sub checkInteger(;$$$$) {
    my ($self, $value, $default, $lower, $upper) = @_;

    unless ($value =~ /^[+-]?\d+$/) {
        $value = $default;
    }

    return undef unless defined $value;

    if (defined $lower && $value < $lower) {
        return $lower;
    }

    if (defined $upper && $value > $upper) {
	return $upper;
    }

    return $value;
}


=head2 checkDecimal

 Description: Checks decimal.
 Arguments:   $value      Value string (optional).
              $default    Default value (optional).
              $lower      Lower limit (optional).
              $upper      Upper limit (optional).
 Returns:     value

=cut

sub checkDecimal(;$$$$) {
    my ($self, $value, $default, $lower, $upper) = @_;

    unless ($value =~ /^-?(?:\d+(?:\.\d*)?|\.\d+)$/) {
        $value = $default;
    }

    return undef unless defined $value;

    if (defined $lower && $value < $lower) {
        return $lower;
    }

    if (defined $upper && $value > $upper) {
	return $upper;
    }

    return $value;
}


=head2 checkRange

 Description: Checks range.
 Arguments:   $value      Value string (optional).
              $default    Default value (optional).
              $lower      Lower limit (optional).
              $upper      Upper limit (optional).
 Returns:     value

=cut

sub checkRange(;$$$$) {
    my ($self, $value, $default, $lower, $upper) = @_;

    $value =~ s/ //g;  # remove spaces
    my ($n1, $n2);

    if ($value =~ /^(\d*)-(\d*)$/) {
        $n1 = $1 if defined $1;
        $n2 = $2 if defined $2;
    }
    else {
        return $default;
    }

    if (defined $lower && defined $n1 && $n1 < $lower) {
        $n1 = $lower;
    }

    if (defined $upper && defined $n2 && $n2 > $upper) {
	$n2 = $upper;
    }

    return "$n1-$n2";
}


=head2 checkString

 Description: Checks string value.
 Arguments:   $value      Value string (optional).
              $default    Default value (optional).
              $values     Allowed values as array reference (optional).
 Returns:     value

=cut

sub checkString(;$$$) {
    my ($self, $value, $default, $values) = @_;
    return $default unless $value;

    my $ok = 0;

    if ($values && ref($values) eq 'ARRAY') {
        for (@$values) {
            if ($value eq $_) {
                $ok = 1;
                last;
            }
        }
    } # else: no way to check

    return $default unless $ok;
    return $value;
}


=head2 checkParams

 Description: Checks input parameters. Each tool must override this method.
 Arguments:   -
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub checkParams() {
    my $self = shift;
    return (1, '');
}


=head2 isFileHandle

 Description: Checks whether parameter is a file handle.
 Arguments:   $fh    File handle.
 Returns:     1 for FileHandle reference or FileHandle glob.
              0 otherwise

=cut

sub isFileHandle($) {
    my ($self, $fh) = @_;

    my $is_handle = (ref($fh) ? (ref($fh) eq 'GLOB'
  	 	     || UNIVERSAL::isa($fh, 'GLOB')
  	 	     || UNIVERSAL::isa($fh, 'IO::Handle'))
          	     : (ref(\$fh) eq 'GLOB'));

    return $is_handle;
}


=head2 getReadHandle

 Description: Returns reference to a file handle for reading. The caller
              must close the file handle if close flag is 1.
              Examples:

                # pass file as string, caller must close file handle
                my ($res, $close, $fh) = $self->getReadHandle('file');

                # pass a file handle glob
                my ($res, $close, $fh) = $self->getReadHandle(\*STDIN);

                # pass reference to a FileHandle object
                my $input = new FileHandle("< file");
                my ($res, $close, $fh) = $self->getReadHandle($input);

 Arguments:   $file    Path, FileHandle reference or Filehandle glob.
 Returns:     1, close flag, file handle on success
              0, 0,          message     on failure

=cut

sub getReadHandle($) {
    my ($self, $file) = @_;

    if ($self->isFileHandle($file)) {
        return (1, 0, $file);
    }

    my $fh = new FileHandle;
    unless ($fh->open("< $file")) {
	$fh->close;
	return (0, 0, "could not read file $file: $!");
    }

    return (1, 1, $fh);
}


=head2 getWriteHandle

 Description: Returns reference to a file handle for writing. The caller
              must close the file handle if close flag is 1.
              Examples:

                # pass file as string, caller must close file handle
                my ($res, $close, $fh) = $self->getWriteHandle('file');

                # pass a file handle glob
                my ($res, $close, $fh) = $self->getWriteHandle(\*STDOUT);

                # pass reference to a FileHandle object
                my $input = new FileHandle("> file");
                my ($res, $close, $fh) = $self->getWriteHandle($input);

 Arguments:   $file    Path, FileHandle reference or Filehandle glob.
 Returns:     1, close flag, file handle on success
              0, 0,          message     on failure

=cut

sub getWriteHandle($) {
    my ($self, $file) = @_;

    if ($self->isFileHandle($file)) {
        return (1, 0, $file);
    }

    my $fh = new FileHandle;
    unless ($fh->open("> $file")) {
	$fh->close;
	return (0, 0, "could not write file $file: $!");
    }

    return (1, 1, $fh);
}


1;

