=head1 NAME

FingerPrintScan - FingerPrintScan utilities.

=head1 SYNOPSIS

  # examples here
  use Dispatcher::Tool::FingerPrintScan;

=head1 DESCRIPTION

=head1 VERSIONS

$Id: FingerPrintScan.pm,v 1.1.1.1 2005/08/18 13:18:26 hunter Exp $

Copyright (c) European Bioinformatics Institute 2002

=head1 AUTHORS / ACKNOWLEDGEMENTS

Ville Silventoinen <vsi@ebi.ac.uk>
Emmanuel Quevillon <tuco@ebi.ac.uk>

=cut

package Dispatcher::Tool::FingerPrintScan;

BEGIN {
  unless($ENV{IPRSCAN_HOME}){
    print STDERR "$0 : IPRSCAN_HOME environment variable not set!\n";
    exit();
  }
  unless($ENV{IPRSCAN_LIB}){
    $ENV{IPRSCAN_LIB} = "$ENV{IPRSCAN_HOME}/lib";
  }
}

use lib $ENV{IPRSCAN_LIB};
use Dispatcher::Tool;
@ISA = qw( Dispatcher::Tool );


use Data::Dumper;
use English;
use File::Spec::Functions;
use POSIX qw(strftime);
use Sys::Hostname;
use strict;


=head2 new

 Description: Constructor allocates an anonymous hash, which
              is tied to the class.
 Arguments:   $name        FingerPrintScan tool name (optional). If given, sets
                           the Dispatcher::Config object automatically inside
                           the object. If not given, the Dispatcher::Config
                           must be set later with setConfig.
              $defaults    Default values for configuration, hash reference
                           (optional).
 Returns:     $self object

=cut

sub new(;$$) {
    my ($class, $name, $defaults) = @_;

    my $self = { };
    bless $self, $class;

    if ($name) {
        $self->setToolName($name);
        my ($res, $msg) = $self->parseConfig(1, $defaults);
        die "failed to parse $name configuration: $msg\n" unless $res;
    }

    return $self;
}


=head2 getCmdLine

 Description: Returns FingerPrintScan command line.
 Arguments:   -
 Returns:     1, cmd on success
              0, msg on failure

=cut

sub getCmdLine() {
    my $self = shift;
    return $self->getConfigValue('cmdline');
}


=head2 parseResult

 Description: Reads a fingerprintscan result file and creates a hash table:

              'seqname'  => 'WAP_RAT',
              'seqinfo'  => 'Fasta sequence',
              'besthit'  => { 'id'     => '4DISULPHCORE',
                              'evalue' => '3.528459e-08',
                              'desc'   => '4-disulphide core signature',
                              'pr'     => 'PR00003'
			    },
              'nhits'    => '4',
              'nmotifs'  => '10',
              'hits'     => { '1' => { 'id'        => '4DISULPHCORE',
                                       'nmotifs'   => '4  of  4',
                                       'sumid'     => '2e+02',
                                       'aveid'     => '49',
                                       'profscore' => '1112',
                                       'ppvalue'   => '4.4e-13',
                                       'evalue'    => '3.5e-08',
                                       'graphscan' => 'IIII' },
                              '2' => { ... },
                              ...
			    },
              'motifs'  => { '1' => { 'id'         => '4DISULPHCORE',
                                      'nmotifs'    => '1  of  4',
                                      'idscore'    => '46.67',
                                      'pfscore'    => '239',
                                      'pvalue'     => '9.35e-04',
                                      'sequence'   => 'EVQKAGRCPW',
                                      'len'        => '10',
                                      'low'        => '7',
                                      'pos'        => '76',
                                      'high'       => '1379' },
                             '2' => { ... },
                             ...
			   }

 Arguments:   $href          Reference to a hash.
              $result        Path to a FingerPrintScan result file.
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub parseResult($$$) {
    my ($self, $href, $result, $inx) = @_;
    return (0, "no fingerprintscan input file defined") unless $result;

    $href = { } unless $href && ref($href) eq 'HASH';
    $inx  = { } unless $inx && ref($inx) eq 'HASH';

    unless (open(FILE, "< $result")) {
        return (0, "could not read fingerprintscan result file $result: $!");
    }

    my $hit = 1;
    my $fpos = 0;
    my $seen = 0;

    while (<FILE>) {
        chomp;

        # 
        # Example result:
        #
        # WAP_RAT2&%&PR00003&%&4DISULPHCORE&%&76&%&85&%&3.5e-08
        # WAP_RAT2&%&PR00003&%&4DISULPHCORE&%&101&%&108&%&3.5e-08
        # WAP_RAT2&%&PR00003&%&4DISULPHCORE&%&108&%&117&%&3.5e-08
        # WAP_RAT2&%&PR00003&%&4DISULPHCORE&%&118&%&126&%&3.5e-08
        # //
        # WAP_RAT3&%&PR00003&%&4DISULPHCORE&%&76&%&85&%&3.5e-08
        # WAP_RAT3&%&PR00003&%&4DISULPHCORE&%&101&%&108&%&3.5e-08
        # WAP_RAT3&%&PR00003&%&4DISULPHCORE&%&108&%&117&%&3.5e-08
        # WAP_RAT3&%&PR00003&%&4DISULPHCORE&%&118&%&126&%&3.5e-08
        # //
        #
        if ($_ =~ /(.*)&%&(.*)&%&(.*)&%&(\d+)&%&(\d+)&%&(.*)/) {
            $href->{$hit} = { } unless $href->{$hit};
	    $inx->{$1} = $fpos unless($seen);
            $href->{$hit}->{seqid}  = $1;
            $href->{$hit}->{methid} = $2;
            $href->{$hit}->{prdesc} = $3;
            $href->{$hit}->{start}  = $4;
            $href->{$hit}->{end}    = $5;
            $href->{$hit}->{evalue} = $6;
	    $seen = 1;
        }
	if($_ =~ /^\/\//){
	  $fpos = tell(FILE);
	  $seen = 0;
	}

	$hit++;
    }
    $href->{nhits} = $hit - 1;

    close FILE;

    return (1, '');
}


=head2 getRawResult

 Description: Returns raw formatted result string.
 Arguments:   $href      Reference to a result hash.
              $seqs      Sequence hash (see InterProScan::checkSequences).
              $iprscan   Reference to Dispatcher::Tool::InterProScan object.
                         Required for IPR fields and GO terms.
              $ipr       1: add IPR fields
                         0: no (default)
              $go        1: add GO terms
                         0: no (default)
 Returns:     1, str on success
              0, msg on failure

=cut

sub getRawResult($$;$$$) {
    my ($self, $href, $seqs, $iprscan, $ipr, $go) = @_;
    return (0, "no result hash defined")    unless $href && ref($href) eq 'HASH';
    return (0, "no sequences hash defined") unless $seqs && ref($seqs) eq 'HASH';

    if (!$iprscan && ($ipr || $go)) {
	return (0, "no InterProScan object defined");
    }
    elsif ($iprscan && ref($iprscan) ne 'Dispatcher::Tool::InterProScan') {
	return (0, "InterProScan object has wrong class: " . ref($iprscan));
    }

    my $nhits = $href->{nhits};
    return (0, "number of hits not defined in result hash") unless defined $nhits;
    return (1, '') if $nhits == 0;

    my $appl = $self->getConfigValue('applname');
    my $date = POSIX::strftime("%d-%b-%Y", localtime);
    my ($hit, $raw, $seqid, $seqcrc, $seqlen, $res, $iprac, $iprname, $goterms);

    for (my $i = 1; $i <= $nhits; $i++) {
	$hit = $href->{$i};
	next unless $hit;  # should not happen

	$seqid = $hit->{seqid};
	next unless $seqid;  # should not happen

	$seqcrc = $seqs->{$seqid}->{crc64};
	$seqlen = $seqs->{$seqid}->{len};

	$raw .= sprintf("%s\t%s\t%s\t", $seqid, $seqcrc, $seqlen);
        $raw .= sprintf("%s\t%s\t%s\t", $appl, $hit->{methid}, $hit->{prdesc});
	$raw .= sprintf("%d\t%d\t", $hit->{start}, $hit->{end});
	$raw .= sprintf("%s\tT\t", $hit->{evalue});
	$raw .= sprintf("%s", $date);

	if ($ipr) {
	    ($res, $iprac, $iprname, $goterms) = $iprscan->getInterProFields($hit->{methid}, $go);
	    return (0, $iprac) unless $res;

	    $raw .= sprintf("\t$iprac\t$iprname") if $ipr;
	    $raw .= sprintf("\t$goterms") if $go;
        }

	$raw .= "\n";
    }

    return (1, $raw);
}



1;
__END__

