=head1 NAME

InterProScanParser - parses InterProScan XML results.

IMPORTANT! On alpha platform LD_LIBRARY_PATH must contain path to the
expat library. At EBI expat is installed in /ebi/extserv/bin/expat/lib.

=head1 SYNOPSIS

  # examples here
  use Dispatcher::Tool::InterProScanParser;
  my $parser = new Dispatcher::Tool::InterProScanParser();

=head1 DESCRIPTION

=head1 VERSIONS

$Id: InterProScanParser.pm,v 1.2 2005/10/26 09:25:30 hunter Exp $

Copyright (c) European Bioinformatics Institute 2002

=head1 AUTHORS / ACKNOWLEDGEMENTS

Ville Silventoinen <vsi@ebi.ac.uk>

=cut


package Dispatcher::Tool::InterProScanParser;

use strict;
use Data::Dumper;
use English;
use XML::Parser;
use utf8;

use vars qw($VERSION);
$VERSION = '0.1';

# element names
use constant E_CATEGORY         => 'category';
use constant E_CHILD_LIST       => 'child_list';
use constant E_CLASSIFICATION   => 'classification';
use constant E_CONTAINS         => 'contains';
use constant E_DESCRIPTION      => 'description';
use constant E_FOUND_IN         => 'found_in';
use constant E_INTERPRO         => 'interpro';
use constant E_INTERPRO_MATCHES => 'interpro_matches';
use constant E_LOCATION         => 'location';
use constant E_MATCH            => 'match';
use constant E_PROTEIN          => 'protein';
use constant E_REL_REF          => 'rel_ref';

# attribute names
use constant A_CLASS_TYPE => 'class_type';
use constant A_CRC64      => 'crc64';
use constant A_DBNAME     => 'dbname';
use constant A_END        => 'end';
use constant A_EVIDENCE   => 'evidence';
use constant A_ID         => 'id';
use constant A_IPR_REF    => 'ipr_ref';
use constant A_LENGTH     => 'length';
use constant A_NAME       => 'name';
use constant A_PARENT_ID  => 'parent_id';
use constant A_SCORE      => 'score';
use constant A_START      => 'start';
use constant A_STATUS     => 'status';
use constant A_TYPE       => 'type';


=head1 CONSTRUCTORS


=head2 new

 Description: Constructor allocates an anonymous hash, which
              is tied to the class.
 Arguments:   -
 Returns:     $self object

=cut

sub new() {
    my $class = shift;

    my $self = { parsed   => 0,     # 0=not parsed, 1=parsed
                 state    => [ ],   # parser state, stack of elements
                 errors   => [ ],   # list of parse errors
                 warnings => [ ],   # list of parse warnings
		 proteins => [ ] }; # protein data

    bless $self, $class;
    $self->_init();
    return $self;
}


=head1 PUBLIC METHODS


=head2 parse

 Description: Parses InterProScan XML result file.
 Arguments:   file    Path to the result.
 Returns:     1, ''  on success
              1, msg on success with warnings
              0, msg on failure

=cut

sub parse($) {
    my ($self, $file, $entry) = @_;
    return (0, "no XML file/entry defined for parser")  unless $file;
    unless($entry){
	return (0, "XML file '$file' does not exist") unless -f $file;
    }

    my $parser = new XML::Parser(ErrorContext     => 2,
                                 ProtocolEncoding => 'UTF-8');

    # Note: Piggypack reference to this object inside XML::Parser object.
    # Otherwise we cannot access the instance variables inside $self or
    # call $self methods.
    $parser->{REF2SELF} = $self;

    $parser->setHandlers(Init  => \&_initParser,
                         Final => \&_finishParser,
                         Start => \&_startElement,
                         End   => \&_endElement,
                         Char  => \&_handleString);

    if($entry){
	eval { $parser->parse($file);};
    }else{
	eval { $parser->parsefile($file);};
    }
    return (0, "parser: $@") if $@;

    if (@{$self->{errors}}) {
        my $msg = @{$self->{errors}} > 1 ? 
                  "There were " . @{$self->{errors}} . " errors:\n" :
                  "There was one error:\n";
        my $error;

        foreach $error (@{$self->{errors}}) {
            $msg .= "$error\n";
        }
        return (0, $msg);
    }

    if (@{$self->{warnings}}) {
        my $msg = @{$self->{warnings}} > 1 ?
                  "There were " . @{$self->{warnings}} . " warnings:\n" :
                  "There was one warning:\n";
        my $warn;

        foreach $warn (@{$self->{warnings}}) {
            $msg .= "$warn\n";
        }
        return (1, $msg);
    }

    return (1, '');
}


=head2 getProteins

 Description: Returns list of proteins.
 Arguments:   -
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub getProteins() {
    my $self = shift;
    return (0, "please parse a result first") unless $self->{parsed};
    return (1, $self->{proteins});
}


=head2 getProtein

 Description: Returns protein hash.
 Arguments:   $id    protein ID
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub getProtein($) {
    my ($self, $id) = @_;
    return (0, "please parse a result first") unless $self->{parsed};
    my $proteins = $self->{proteins};
    my $protein;

    foreach $protein (@$proteins) {
	my $pid = $protein->{id};
	if (lc($id) eq lc($pid)) {
	    return (1, $protein);
	}
    }

    return (0, "protein with id $id not found");
}


# PRIVATE METHODS


#
# Description: Callback handler for XML::Parser Init event.
# Arguments:   $xp   XML::Parser::Expat object.
# Returns:     -
#
sub _initParser {
    my $xp = shift;
    my $self = $xp->{REF2SELF};

    # initialize global variables
    $self->{parsed}   = 0;
    $self->{state}    = [ ];
    $self->{errors}   = [ ];
    $self->{warnings} = [ ];
    $self->{proteins} = [ ];
}


#
# Description: Callback handler for XML::Parser Final event.
# Arguments:   $xp   XML::Parser::Expat object.
# Returns:     -
#
sub _finishParser {
    my $xp = shift;
    my $self = $xp->{REF2SELF};
    $self->{parsed} = 1;
}


#
# Description: Callback handler for XML::Parser Start event.  Stores element
#              and/or attribute values to Main object instance variables.
# Arguments:   $xp       XML::Parser::Expat object.
#              $elem     Element name.
#              %attrs    Attributes in name-value hash.
# Returns:     -
#
sub _startElement {
    my ($xp, $elem, %attrs) = @_;
    my $self = $xp->{REF2SELF};

    $self->_addState($elem);

    if ($elem eq E_INTERPRO_MATCHES) {

    }
    elsif ($elem eq E_PROTEIN) {
        unless ($self->_getState(1) eq E_INTERPRO_MATCHES) {
            $self->_invalidScope($xp, E_PROTEIN, E_INTERPRO_MATCHES);
            return;
        }

        my $id     = $attrs{&A_ID};
        my $name   = $attrs{&A_NAME};    # optional
        my $length = $attrs{&A_LENGTH};
        my $crc64  = $attrs{&A_CRC64};

        unless ($id && $length && $crc64) {
            # ignore this protein, converter is just behaving strangely:
            #
            #   <protein id="" length="" crc64="" >
            #   </protein>
            #
            #$self->_missingAttrs($xp, E_PROTEIN, A_ID, A_LENGTH, A_CRC64);
            #
            return;
        }

	my $protein = { &A_ID     => $id,
                        &A_LENGTH => $length,
                        &A_CRC64  => $crc64 };

	$protein->{&A_NAME} = $name if $name;
	my $proteins = $self->{proteins};
        push @$proteins, $protein;
    }
    elsif ($elem eq E_INTERPRO) {
        unless ($self->_getState(1) eq E_PROTEIN) {
            $self->_invalidScope($xp, E_INTERPRO, E_PROTEIN);
            return;
        }

        my $id     = $attrs{&A_ID};
        my $name   = $attrs{&A_NAME};
        my $type   = $attrs{&A_TYPE};
        my $parent = $attrs{&A_PARENT_ID};  # optional

	$id   = "no identifier" unless $id;
	$name = "no name"       unless $name;
	$type = "no type"       unless $type;

        #unless ($id) {
        #    $self->_missingAttrs($xp, E_INTERPRO, A_ID);
        #    return;
        #}

	my $interpro = { &A_ID   => $id,
			 &A_NAME => $name,
			 &A_TYPE => $type };

	$interpro->{&A_PARENT_ID} = $parent if $parent;

	my $proteins  = $self->{proteins};
	my $protein   = pop @$proteins;
        my $interpros = $protein->{interpros};
	$interpros = [ ] unless $interpros;
        $protein->{interpros} = $interpros;
	push @$interpros, $interpro;
	push @$proteins, $protein;
    }
    elsif ($elem eq E_CHILD_LIST) {
        unless ($self->_getState(1) eq E_INTERPRO) {
            $self->_invalidScope($xp, E_CHILD_LIST, E_INTERPRO);
            return;
        }
    }
    elsif ($elem eq E_CONTAINS) {
        unless ($self->_getState(1) eq E_INTERPRO) {
            $self->_invalidScope($xp, E_CONTAINS, E_INTERPRO);
            return;
        }
    }
    elsif ($elem eq E_FOUND_IN) {
        unless ($self->_getState(1) eq E_INTERPRO) {
            $self->_invalidScope($xp, E_FOUND_IN, E_INTERPRO);
            return;
        }
    }
    elsif ($elem eq E_REL_REF) {
	my $parent = $self->_getState(1);

        unless ($parent eq E_CHILD_LIST || $parent eq E_CONTAINS || $parent eq E_FOUND_IN) {
            $self->_invalidScope($xp, E_REL_REF, E_CHILD_LIST, E_CONTAINS, E_FOUND_IN);
            return;
        }

	my $ref = $attrs{&A_IPR_REF};
	unless ($ref) {
            $self->_missingAttrs($xp, E_REL_REF, A_IPR_REF);
            return;
	}

	my $proteins  = $self->{proteins};
	my $protein   = pop @$proteins;
        my $interpros = $protein->{interpros};
	my $interpro  = pop @$interpros;
	my $refs  = $interpro->{$parent};
	$refs = [ ] unless $refs;
	push @$refs, $ref;
	$interpro->{$parent} = $refs;
	push @$interpros, $interpro;
	push @$proteins, $protein;
    }
    elsif ($elem eq E_MATCH) {
        unless ($self->_getState(1) eq E_INTERPRO || $self->_getState(1) eq E_PROTEIN) {
            $self->_invalidScope($xp, E_MATCH, E_INTERPRO, E_PROTEIN);
            return;
        }

	my $id     = $attrs{&A_ID};
        my $name   = $attrs{&A_NAME};
        my $dbname = $attrs{&A_DBNAME};

	$id     = "no identifier"    unless $id;
	$name   = "no name"          unless $name;
	$dbname = "no database name" unless $dbname;

	#unless ($id) {
        #    $self->_missingAttrs($xp, E_MATCH, A_ID);
        #    return;
        #}

	my $match = { &A_ID     => $id,
                      &A_NAME   => $name,
                      &A_DBNAME => $dbname };

	my $proteins = $self->{proteins};
	my $protein  = pop @$proteins;

	if ($self->_getState(1) eq E_INTERPRO) {
	    my $interpros = $protein->{interpros};
	    my $interpro = pop @$interpros;
	    my $matches  = $interpro->{matches};
	    $matches = [ ] unless $matches;
	    push @$matches, $match;
	    $interpro->{matches} = $matches;
	    push @$interpros, $interpro;
        }
        else {
	    # unintegrated matches are within protein tag
	    my $matches = $protein->{matches};
	    $matches = [ ] unless $matches;
	    push @$matches, $match;
	    $protein->{matches} = $matches;
        }

	push @$proteins, $protein;
    }
    elsif ($elem eq E_LOCATION) {
        unless ($self->_getState(1) eq E_MATCH) {
            $self->_invalidScope($xp, E_LOCATION, E_MATCH);
            return;
        }

	my $start    = $attrs{&A_START};
        my $end      = $attrs{&A_END};
        my $score    = $attrs{&A_SCORE};  # optional
        my $status   = $attrs{&A_STATUS};
        my $evidence = $attrs{&A_EVIDENCE};

	# NOTE: start, end and score can have zero value
	unless (defined $start && defined $end && $status && $evidence) {
            $self->_missingAttrs($xp, E_LOCATION, A_START, A_END, A_STATUS, A_EVIDENCE);
            return;
        }

	my $location = { &A_START    => $start,
			 &A_END      => $end,
			 &A_STATUS   => $status,
			 &A_EVIDENCE => $evidence };

	$location->{&A_SCORE} = $score if defined $score;

	my $proteins  = $self->{proteins};
	my $protein   = pop @$proteins;

	if ($self->_getState(2) eq E_INTERPRO) {
	    my $interpros = $protein->{interpros};
	    my $interpro = pop @$interpros;
	    my $matches  = $interpro->{matches};
	    my $match    = pop @$matches;
	    my $locations = $match->{locations};
	    $locations = [ ] unless $locations;
	    push @$locations, $location;
	    $match->{locations} = $locations;
	    push @$matches, $match;
	    push @$interpros, $interpro;
        }
        elsif ($self->_getState(2) eq E_PROTEIN) {
	    # unintegrated matches are within protein tag
	    my $matches = $protein->{matches};
	    my $match    = pop @$matches;
	    my $locations = $match->{locations};
	    $locations = [ ] unless $locations;
	    push @$locations, $location;
	    $match->{locations} = $locations;
	    push @$matches, $match;
        }
	else {
            $self->_invalidScope($xp, E_LOCATION, E_INTERPRO, E_PROTEIN);
            return;
	}

	push @$proteins, $protein;
    }
    elsif ($elem eq E_CLASSIFICATION) {
        unless ($self->_getState(1) eq E_INTERPRO) {
            $self->_invalidScope($xp, E_CLASSIFICATION, E_INTERPRO);
            return;
        }

	my $id    = $attrs{&A_ID};
        my $type  = $attrs{&A_CLASS_TYPE};

	unless ($id && $type) {
            $self->_missingAttrs($xp, E_CLASSIFICATION, A_ID, A_CLASS_TYPE);
            return;
        }

	my $classification = { &A_ID         => $id,
                               &A_CLASS_TYPE => $type };

	my $proteins = $self->{proteins};
	my $protein  = pop @$proteins;
	my $interpros = $protein->{interpros};
	my $interpro = pop @$interpros;
	my $classifications = $interpro->{classifications};
	$classifications = [ ] unless $classifications;
	push @$classifications, $classification;
	$interpro->{classifications} = $classifications;
	push @$interpros, $interpro;
	push @$proteins, $protein;
    }
    elsif ($elem eq E_CATEGORY) {
        unless ($self->_getState(1) eq E_CLASSIFICATION) {
            $self->_invalidScope($xp, E_CATEGORY, E_CLASSIFICATION);
            return;
        }
    }
    elsif ($elem eq E_DESCRIPTION) {
        unless ($self->_getState(1) eq E_CLASSIFICATION) {
            $self->_invalidScope($xp, E_DESCRIPTION, E_CLASSIFICATION);
            return;
        }
    }
}


#
# Description: Callback handler for XML::Parser End event.
# Arguments:   $xp      XML::Parser::Expat object.
#              $elem    Element name.
# Returns:     -
#
sub _endElement {
    my ($xp, $elem) = @_;
    my $self = $xp->{REF2SELF};

    $self->_removeState();
}


#
# Description: Callback handler for XML::Parser Char event.
# Arguments:   $xp       XML::Parser::Expat object.
#              $string   String.
# Returns:     -
#
sub _handleString {
    my ($xp, $string) = @_;
    my $self = $xp->{REF2SELF};

    return unless $string;

    my $state = $self->_getState(0);

    if ($state eq E_CATEGORY || $state eq E_DESCRIPTION) {
        #
	# Example of classification:
        #
        # <classification id="GO:0019079" class_type="GO">
	#   <category>Biological Process</category>
	#   <description>viral genome replication</description>
	# </classification>
        #
	my $proteins = $self->{proteins};
	my $protein  = pop @$proteins;
	my $interpros = $protein->{interpros};
	my $interpro = pop @$interpros;
	my $classifications = $interpro->{classifications};
	my $classification = pop @$classifications;
	$classification->{$state} = $string;
	push @$classifications, $classification;
	$interpro->{classifications} = $classifications;
	push @$interpros, $interpro;
	push @$proteins, $protein;
    }
}


#
# Description: Adds element name to state stack array.
# Arguments:   $elem    Element name.
# Returns:     -
#
sub _addState($) {
    my ($self, $elem) = @_;
    return unless $elem;
    push @{$self->{state}}, $elem;
}


#
# Description: Removes last element name from state stack array.
# Arguments:   -
# Returns:     Element name.
#
sub _removeState() {
    my $self = shift;
    return pop @{$self->{state}};
}


#
# Description: Retrieves element names from state stack array.
#              Does not modify the stack.            
# Arguments:   $i   Offset index from stack top: 0 is the element on top,
#                   1 is the next element from top, etc.
# Returns:     Element name.
#
sub _getState($;$) {
    my ($self, $i) = @_;
    $i = 0 unless $i;

    my $aref = $self->{state};
    my $len  = @$aref;
    return '' unless $len > $i;  # underflow
    return $aref->[$len-$i-1];
}


#
# Description: Adds new message to the error list.
# Arguments:   $xp     XML::Parser::Expat object.
#              $msg    Error message.
# Returns:     -
#
sub _addError($$) {
    my ($self, $xp, $msg) = @_;
    push @{$self->{errors}}, "$self->{file} line " . $xp->current_line .
                             " column " . $xp->current_column . ": $msg";
}


#
# Description: Adds new message to the warning list.
# Arguments:   $xp     XML::Parser::Expat object.
#              $msg    Error message.
# Returns:     -
#
sub _addWarning($$) {
    my ($self, $xp, $msg) = @_;
    push @{$self->{warnings}}, "$self->{file} line " . $xp->current_line .
                               " column " . $xp->current_column . ": $msg";
}


#
# Description: Convenience method for missing attributes errors.
# Arguments:   $xp       XML::Parser::Expat object.
#              $elem     Element name.
#              @attrs    Attribute names (one or more).
# Returns:     -
#
sub _missingAttrs {
    my ($self, $xp, $elem, @attrs) = @_;

    my ($msg, $attr);
    my $delim = '';

    $msg  = "element '$elem' requires ";
    $msg .= @attrs > 1 ? "attributes " : "attribute ";

    foreach $attr (@attrs) {
        $msg .= "$delim'$attr'";
        $delim = ', ';
    }
    $self->_addError($xp, $msg);
}


#
# Description: Convenience method for invalid attribute value errors.
# Arguments:   $xp       XML::Parser::Expat object.
#              $attr     Attribute name.
#              $value    Attribute value.
# Returns:     -
#
sub _invalidAttrValue {
    my ($self, $xp, $attr, $value) = @_;
    my $msg = "attribute '$attr' has invalid value '$value'";
    $self->_addError($xp, $msg);
}


#
# Description: Convenience method for invalid element scope errors.
#              This can be replaced later by defining a DTD.
# Arguments:   $xp         XML::Parser::Expat object.
#              $child      Child element name.
#              @parents    Valid parent element names (one or more).
# Returns:     -
#
sub _invalidScope {
    my ($self, $xp, $child, @parents) = @_;
    my $msg  = "element '$child' must be inside element ";
    my $delim = '';
    my $parent;

    foreach $parent (@parents) {
        $msg .= "$delim'$parent'";
        $delim = ' or ';
    }
    $self->_addError($xp, $msg);
}


#
# Description: Initializes the object.
# Arguments:   -
# Returns:     -
#
sub _init() {
    my $self = shift;
}


1;
__END__
