=head1 NAME

SignalP - SignalP utilities.

=head1 SYNOPSIS

  # examples here
  use Dispatcher::Tool::SignalP;

=head1 DESCRIPTION

=head1 VERSIONS

$Id: SignalP.pm,v 1.3 2005/11/16 16:31:45 hunter Exp $

Copyright (c) European Bioinformatics Institute 2002

=head1 AUTHORS / ACKNOWLEDGEMENTS

Ville Silventoinen <vsi@ebi.ac.uk>
Emmanuel Quevillon <tuco@ebi.ac.uk>

=cut

package Dispatcher::Tool::SignalP;

BEGIN {
  unless($ENV{IPRSCAN_HOME}){
    print STDERR "$0 : IPRSCAN_HOME environment variable not set!\n";
    exit();
  }
  unless($ENV{IPRSCAN_LIB}){
    $ENV{IPRSCAN_LIB} = "$ENV{IPRSCAN_HOME}/lib";
  }
}

use lib $ENV{IPRSCAN_LIB};
use Dispatcher::Tool;
@ISA = qw( Dispatcher::Tool );

use Data::Dumper;
use English;
use File::Spec::Functions;
use POSIX qw(strftime);
use Sys::Hostname;
use strict;


=head2 new

 Description: Constructor allocates an anonymous hash, which
              is tied to the class.
 Arguments:   $name        SignalP tool name (optional). If given, sets
                           the Dispatcher::Config object automatically inside
                           the object. If not given, the Dispatcher::Config
                           must be set later with setConfig.
              $defaults    Default values for configuration, hash reference
                           (optional).
 Returns:     $self object

=cut

sub new(;$$) {
    my ($class, $name, $defaults) = @_;

    my $self = { };
    bless $self, $class;

    if ($name) {
        $self->setToolName($name);
        my ($res, $msg) = $self->parseConfig(1, $defaults);
        die "failed to parse $name configuration: $msg\n" unless $res;
    }

    return $self;
}


=head2 getCmdLine

 Description: Returns SignalP command line.
 Arguments:   -
 Returns:     1, cmd on success
              0, msg on failure

=cut

sub getCmdLine() {
    my $self = shift;
    return $self->getConfigValue('cmdline');
}


=head2 parseResult

  Description: Reads a signalp result file and creates a hash table:

  Arguments:   $href          Reference to a hash.
               $result        Path to a SignalP result file.
               $inx           Reference to a hsah.
  Returns:     1, ''  on success
               0, msg on failure

=cut

sub parseResult($$) {
    my ($self, $href, $result, $inx) = @_;
    return (0, "no signalp input file defined") unless $result;

    $href = { } unless $href && ref($href) eq 'HASH';
    $inx = { } unless $inx && ref($inx) eq 'HASH';

    my $version = $self->getConfigValue('signalp.version');
    return(0, "Signalp version not configured") unless($version);
    
    my ($res, $msg);
    my $sub = '_parseVersion' . $version;

    ($res, $msg) = $self->$sub($href, $result, $inx);
    return(0, $msg) unless $res;

    return(1, '');
  }


=head2 _parseVersion1

  Description: Reads a signalp 1.x result file and creates a hash table:
	      'nhits'   => '2',
              '1'       => { 'seqid'   => 'swall|Q9LIP3|C72Y_ARATH',
                             'seqdesc' => 'Cytochrome P450 71B37 (EC 1.14.-.-).',
                             'start'   => '1',
                             'end'     => '23'
                           },
              '2'       => { 'seqid'   => 'swall|Q9LIP4|C72X_ARATH',
                             'seqdesc' => 'Cytochrome P450 71B36 (EC 1.14.-.-).',
                             'start'   => '1',
                             'end'     => '19'
                           }

 Arguments:   $href          Reference to a hash.
              $result        Path to a SignalP result file.
	      $inx           Reference to a hash.
 Returns:     1, ''  on success
              0, msg on failure

=cut


sub _parseVersion1 {

  my ($self, $href, $result, $inx) = @_;

    unless (open(FILE, "< $result")) {
        return (0, "could not read signalp result file $result: $!");
    }

    my ($seqid, $seqdesc);
    my $hit = 1;

    my $fpos = 0;

    while (<FILE>) {
        chomp;

        #
        # Example result:
        #
        # >swall|Q9LIP3|C72Y_ARATH Cytochrome P450 71B37 (EC 1.14.-.-).
        # %score BG 419.439995 (4.194400)
        # %score FW 412.062700 (4.120627) FW-BG 7.377294 (0.073773)
        # %score N0 414.928068 (4.149281) N0-BG 4.511927 (0.045119) N0/FW 0.137227
        # %pred N0: S 1 1, n 2 5, H 6 17, c 18 22, C 23 23, M 24 100
        #    MATIWFLPLL FLSCLLLAAL RLKKRRQHQR KPPSPPGFPI IGNLHQLGEL PHQSLWSLSK 
        # ?0 SnnnnHHHHH HHHHHHHccc ccCMMMMMMM MMMMMMMMMM MMMMMMMMMM MMMMMMMMMM 
        # 
        #    KYGPVMLLKF GSIPTVVVSS SETAKQALKI HDLNCCSRPS 
        # ?0 MMMMMMMMMM MMMMMMMMMM MMMMMMMMMM MMMMMMMMMM 
        # 
        if ($_ =~ /^>(.*)/) {
	    $seqid = $self->trim($1);
	    $inx->{$seqid} = $fpos;
	    if ($seqid =~ /^([^\s]+)\s+(.+)/) {
		$seqid   = $1;
		$seqdesc = $2;
	    }
        }
	elsif ($_ =~ /^\%pred.*C\s+\d+\s+(\d+)/) {
            $href->{$hit}->{seqid}   = $seqid   if $seqid;
            $href->{$hit}->{seqdesc} = $seqdesc if $seqdesc;
	    $href->{$hit}->{methid}  = "";
            $href->{$hit}->{start}   = '1';
            $href->{$hit}->{end}     = $1-1;
	    $hit++;
	  }
	elsif($_ =~ /^\s*$/){
	  $fpos = tell(FILE);
	}
      }
  
  $href->{nhits} = $hit - 1;
  
  close FILE;
  
  return (1, '');
}

=head2 _parseVersion2

 Description: Reads a signalp 2.0 output and creates a hash table.
	      'nhits'   => '2',
              '1'       => { 'seqid'   => 'swall|Q9LIP3|C72Y_ARATH',
                             'seqdesc' => 'Cytochrome P450 71B37 (EC 1.14.-.-).',
                             'start'   => '1',
                             'end'     => '23'
                           },
              '2'       => { 'seqid'   => 'swall|Q9LIP4|C72X_ARATH',
                             'seqdesc' => 'Cytochrome P450 71B36 (EC 1.14.-.-).',
                             'start'   => '1',
                             'end'     => '19'
                           }

 Arguments:   $href     Reference to a hash.
              $result   Path to SignalP result file.
              $inx      Reference to a hash.

 Returns:     1, ''     on success
              0, $msg   on failure

=cut

  sub _parseVersion2 ($$$) {

    my ($self, $href, $result, $inx) = @_;

    unless(open(FILE, "<$result")){
      return(0, "could not read signalp result file $result: $!");
    }

    my $method = $self->getConfigValue('method');
    return(0, "Method for Signalp not configured") unless($method =~ /hmm|nn/);

    my $regexp = ($method eq 'nn' ? '\# Most likely .+pos\. (\d+)' : 'Max cleavage .+pos\. (\d+)');

    my ($seqid, $seqdesc, $seen, $secreted);
    my $hit = 1;

    my $fpos = 0;

    while (<FILE>) {
        chomp;

        # Example positive and negative results for HMM method
        #************************* SignalP predictions *************************
        #Using hidden Markov models (HMM) trained on eukaryotes
        #
        #----------------------------------------------------------------------
        #>sp|P07024|USHA_ECOLI Protein ushA precursor [Includes: UDP-sugar hydrolase (EC 3.6.1.45) (UDP-sugar diphosphatase) (UDP-sugar pyrophosphatase)
        #
        #SignalP-HMM result:
        #>sp|P07024|USHA_ECOLI
        #Prediction: Signal peptide
        #Signal peptide probability: 0.998
        #Signal anchor probability: 0.001
        #Max cleavage site probability: 0.780 between pos. 25 and 26
        #
        #----------------------------------------------------------------------
        #>test
        #
        #SignalP-HMM result:
        #>test
        #Prediction: Non-secretory protein
        #Signal peptide probability: 0.012
        #Signal anchor probability: 0.091
        #Max cleavage site probability: 0.008 between pos. 17 and 18
        #

        # Example results for NN method
        #************************* SignalP predictions *************************
        #Using neural networks (NN) trained on eukaryotes
        #
        #----------------------------------------------------------------------
        #>sp|P07024|USHA_ECOLI Protein ushA precursor [Includes: UDP-sugar hydrolase (EC 3.6.1.45) (UDP-sugar diphosphatase) (UDP-sugar pyrophosphatase)
        #
        #SignalP-NN result:
        #>sp|P07024|U           length = 100
        ## Measure  Position  Value  Cutoff  signal peptide?
        #  max. C    20       0.811   0.33   YES
        #  max. Y    20       0.748   0.32   YES
        #  max. S    11       0.984   0.82   YES
        #  mean S     1-19    0.930   0.47   YES
        ## Most likely cleavage site between pos. 19 and 20: TLA-SE
        #
        #----------------------------------------------------------------------
        #>FAEE
        #
        #SignalP-NN result:
        #>FAEE                  length = 100
        ## Measure  Position  Value  Cutoff  signal peptide?
        #max. C    18       0.480   0.32   YES
        #max. Y    18       0.227   0.33   NO
        #max. S    41       0.538   0.87   NO
        #mean S     1-17    0.194   0.48   NO
        #D     1-17    0.210   0.43   NO
        #


        if($_ =~ /^-+/){
            $fpos = tell(FILE);
            $seen = 1;
        }
        if($_ =~ />(\S+)/ && $seen){
            $seen = 0;
            $seqid = $1;
        }

        # If the output is from HMM, need to check if prediction is "Signal peptide"
        # or "Non-secretory protein" and flag as such (otherwise, will report all
        # results).  NN does not report any cleavage site if no signal peptide present
        # but flag "secreted" anyway :-)
        if (($method eq 'hmm' && /Prediction\: Signal/) || $method eq 'nn') {
            $secreted=1;
        }

        if($_ =~ /^$regexp/ && $secreted) {
            $href->{$hit}->{end} = $1;
            $href->{$hit}->{start} = '1';
            $href->{$hit}->{seqid} = $seqid;
            $href->{$hit}->{methid}  = "";
            $inx->{$seqid} = $fpos;
            $hit++;
            $secreted=0;
        }
    }

    close(FILE);

    $href->{nhits} = $hit -1;

    return(1, '');
}


=head2 getRawResult

 Description: Returns raw formatted result string.
 Arguments:   $href      Reference to a result hash.
              $seqs      Sequence hash (see InterProScan::checkSequences).
              $iprscan   Reference to Dispatcher::Tool::InterProScan object.
                         Required for IPR fields and GO terms.
              $ipr       1: add IPR fields
                         0: no (default)
              $go        1: add GO terms
                         0: no (default)
 Returns:     1, str on success
              0, msg on failure

=cut

sub getRawResult($$;$$$) {
    my ($self, $href, $seqs, $iprscan, $ipr, $go) = @_;
    return (0, "no result hash defined")     unless $href && ref($href) eq 'HASH';
    return (0, "no sequences hash defined") unless $seqs && ref($seqs) eq 'HASH';

    if (!$iprscan && ($ipr || $go)) {
	return (0, "no InterProScan object defined");
    }
    elsif ($iprscan && ref($iprscan) ne 'Dispatcher::Tool::InterProScan') {
	return (0, "InterProScan object has wrong class: " . ref($iprscan));
    }

    my $nhits = $href->{nhits};
    return (0, "number of hits not defined in result hash") unless defined $nhits;
    return (1, '') if $nhits == 0;

    my $appl = $self->getConfigValue('applname');
    my $date = POSIX::strftime("%d-%b-%Y", localtime);
    my ($hit, $raw, $seqid, $seqcrc, $seqlen, $res, $iprac, $iprname, $goterms);

    for (my $i = 1; $i <= $nhits; $i++) {
	$hit = $href->{$i};
	next unless $hit;  # should not happen

	$seqid = $hit->{seqid};
	next unless $seqid;  # should not happen

	$seqcrc = $seqs->{$seqid}->{crc64};
	$seqlen = $seqs->{$seqid}->{len};

	# SignalP result is very minimal, there are no identifiers or evalues
	$raw .= sprintf("%s\t%s\t%s\t", $seqid, $seqcrc, $seqlen);
        $raw .= sprintf("%s\tsignalp\tsignal-peptide\t", $appl);
	$raw .= sprintf("%d\t%d\t", $hit->{start}, $hit->{end});
	$raw .= sprintf("NA\t?\t");  # there is no evalue
	$raw .= sprintf("%s", $date);

	if ($ipr) {
	    # not possible to query SignalP matches from InterPro (no ID)
	    $raw .= sprintf("\tNULL\tNULL") if $ipr;
        }

	$raw .= "\n";
    }

    return (1, $raw);
}



1;
__END__

