=head1 NAME

TMHMM - TMHMM utilities.

=head1 SYNOPSIS

  # examples here
  use Dispatcher::Tool::TMHMM;

=head1 DESCRIPTION

=head1 VERSIONS

$Id: TMHMM.pm,v 1.1.1.1 2005/08/18 13:18:26 hunter Exp $

Copyright (c) European Bioinformatics Institute 2002

=head1 AUTHORS / ACKNOWLEDGEMENTS

Ville Silventoinen <vsi@ebi.ac.uk>
Emmanuel Quevillon <tuco@ebi.ac.uk>

=cut

package Dispatcher::Tool::TMHMM;

BEGIN {
  unless($ENV{IPRSCAN_HOME}){
    print STDERR "$0 : IPRSCAN_HOME environment variable not set!\n";
    exit();
  }
  unless($ENV{IPRSCAN_LIB}){
    $ENV{IPRSCAN_LIB} = "$ENV{IPRSCAN_HOME}/lib";
  }
}

use lib $ENV{IPRSCAN_LIB};
use Dispatcher::Tool;
@ISA = qw( Dispatcher::Tool );

use Data::Dumper;
use English;
use File::Spec::Functions;
use POSIX qw(strftime);
use Sys::Hostname;
use strict;


=head2 new

 Description: Constructor allocates an anonymous hash, which
              is tied to the class.
 Arguments:   $name        TMHMM tool name (optional). If given, sets
                           the Dispatcher::Config object automatically inside
                           the object. If not given, the Dispatcher::Config
                           must be set later with setConfig.
              $defaults    Default values for configuration, hash reference
                           (optional).
 Returns:     $self object

=cut

sub new(;$$) {
    my ($class, $name, $defaults) = @_;

    my $self = { };
    bless $self, $class;

    if ($name) {
        $self->setToolName($name);
        my ($res, $msg) = $self->parseConfig(1, $defaults);
        die "failed to parse $name configuration: $msg\n" unless $res;
    }

    return $self;
}


=head2 getCmdLine

 Description: Returns TMHMM command line.
 Arguments:   -
 Returns:     1, cmd on success
              0, msg on failure

=cut

sub getCmdLine() {
    my $self = shift;
    return $self->getConfigValue('cmdline');
}


=head2 parseResult

 Description: Reads a tmhmm result file and creates a hash table:

              'nhits'   => '2',
              '1'       => { 'seqid'   => 'swall|Q92455|ABF1_TRIRE',
                             'seqdesc' => 'Alpha-L-arabinofuranosidase precursor (EC 3.2.1.55) (Arabinosidase).',
                             'start'   => '1',
                             'end'     => '23'
                           },
              '2'       => { 'seqid'   => 'swall|Q9LIP4|C72X_ARATH',
                             'seqdesc' => 'Cytochrome P450 71B36 (EC 1.14.-.-).',
                             'start'   => '1',
                             'end'     => '19'
                           }

 Arguments:   $href          Reference to a hash.
              $result        Path to a TMHMM result file.
 Returns:     1, ''  on success
              0, msg on failure

=cut

sub parseResult($$) {
    my ($self, $href, $result, $inx) = @_;
    return (0, "no tmhmm input file defined") unless $result;

    $href = { } unless $href && ref($href) eq 'HASH';
    $inx = { } unless $href && ref($inx) eq 'HASH';

    unless (open(FILE, "< $result")) {
        return (0, "could not read tmhmm result file $result: $!");
    }

    my ($version, $seqid, $seqdesc);
    my $hit = 1;

    my $fpos = 0;

    while (<FILE>) {
        chomp;

        #
        # Example result:
        #
        # >swall|Q9LIP3|C72Y_ARATH Cytochrome P450 71B37 (EC 1.14.-.-).
        # %score BG 419.439995 (4.194400)
        # %score FW 412.062700 (4.120627) FW-BG 7.377294 (0.073773)
        # %score N0 414.928068 (4.149281) N0-BG 4.511927 (0.045119) N0/FW 0.137227
        # %pred N0: S 1 1, n 2 5, H 6 17, c 18 22, C 23 23, M 24 100
        #    MATIWFLPLL FLSCLLLAAL RLKKRRQHQR KPPSPPGFPI IGNLHQLGEL PHQSLWSLSK 
        # ?0 SnnnnHHHHH HHHHHHHccc ccCMMMMMMM MMMMMMMMMM MMMMMMMMMM MMMMMMMMMM 
        # 
        #    KYGPVMLLKF GSIPTVVVSS SETAKQALKI HDLNCCSRPS 
        # ?0 MMMMMMMMMM MMMMMMMMMM MMMMMMMMMM MMMMMMMMMM 
        # 
        if ($_ =~ /^>(.*)/) {
	    $seqid = $self->trim($1);
	    $inx->{$seqid} = $fpos;
	    if ($seqid =~ /^([^\s]+)\s+(.+)/) {
		$seqid   = $1;
		$seqdesc = $2;
	    }
        }
	elsif ($_ =~ /^\%pred.*M\s+/) {
	    my $loc = $_;

	    while ($loc =~ /M\s+(\d+)\s+(\d+),/) {
		$href->{$hit} = { } unless $href->{$hit};
		$href->{$hit}->{seqid}   = $seqid   if $seqid;
		$href->{$hit}->{seqdesc} = $seqdesc if $seqdesc;
		$href->{$hit}->{methid}  = "";
		$href->{$hit}->{start}   = $1;
		$href->{$hit}->{end}     = $2;
		$hit++;
		$loc = $POSTMATCH;  # there can be multiple locations on one line
	    }
	  }
	elsif($_ =~ /^\s*$/){
	  $fpos = tell(FILE);
	}
    }

    $href->{nhits} = $hit - 1;

    close FILE;

    return (1, '');
}


=head2 getRawResult

 Description: Returns raw formatted result string.
 Arguments:   $href      Reference to a result hash.
              $seqid     Sequence ID.
              $seqcrc    CRC64 checksum.
              $seqlen    Sequence length.
              $iprscan   Reference to Dispatcher::Tool::InterProScan object.
                         Required for IPR fields and GO terms.
              $ipr       1: add IPR fields
                         0: no (default)
              $go        1: add GO terms
                         0: no (default)
 Returns:     1, str on success
              0, msg on failure

=cut

sub getRawResult($$;$$$) {
    my ($self, $href, $seqs, $iprscan, $ipr, $go) = @_;
    return (0, "no result hash defined")     unless $href && ref($href) eq 'HASH';
    return (0, "no sequences hash defined") unless $seqs && ref($seqs) eq 'HASH';

    if (!$iprscan && ($ipr || $go)) {
	return (0, "no InterProScan object defined");
    }
    elsif ($iprscan && ref($iprscan) ne 'Dispatcher::Tool::InterProScan') {
	return (0, "InterProScan object has wrong class: " . ref($iprscan));
    }

    my $nhits = $href->{nhits};
    return (0, "number of hits not defined in result hash") unless defined $nhits;
    return (1, '') if $nhits == 0;

    my $appl = $self->getConfigValue('applname');
    my $date = POSIX::strftime("%d-%b-%Y", localtime);
    my ($hit, $raw, $seqid, $seqcrc, $seqlen, $res, $iprac, $iprname, $goterms);

    for (my $i = 1; $i <= $nhits; $i++) {
	$hit = $href->{$i};
	next unless $hit;  # should not happen

	$seqid = $hit->{seqid};
	next unless $seqid;  # should not happen

	$seqcrc = $seqs->{$seqid}->{crc64};
	$seqlen = $seqs->{$seqid}->{len};

	# TMHMM result is very minimal, there are no identifiers or evalues
	$raw .= sprintf("%s\t%s\t%s\t", $seqid, $seqcrc, $seqlen);
        $raw .= sprintf("%s\ttmhmm\ttransmembrane_regions\t", $appl);
	$raw .= sprintf("%d\t%d\t", $hit->{start}, $hit->{end});
	$raw .= sprintf("NA\t?\t");  # there is no evalue
	$raw .= sprintf("%s", $date);

	if ($ipr) {
	    # not possible to query TMHMM matches from InterPro (no ID)
	    $raw .= sprintf("\tNULL\tNULL") if $ipr;
        }

	$raw .= "\n";
    }

    return (1, $raw);
}



1;
__END__

