=head1 NAME

Index - Indexation of files tool

=head1 SYNOPSIS


  use Index;

=head1 DESCRIPTION

  This module can help you to index a file with key=seqid, value=position in the file
  You will need to have installed DB_File module (Berkley_DB) to use it.

  #e.g. Indexation of this file
  #>1433_OENHO Wap_rat
  #MATAPSPREENVYLAKLAEQAERYEEMVEFMEKVCAAADSEELTVEERNLLSVAYKNVIG
  #...
  #>41_HUMAN Wap_human
  #MTTEKSLVTEAENSQHQQKEEGEEAINSGQQEPQQEESCQTAAEGDNWCEQKLKASNGDT
  #...

  #keys=1433_OENHO & 41_HUMAN

  # examples here
  #Index module does not implement a 'new' method. So you need to call a specific module
  #Here we will show you how to index a Fasta file.
  use Index::Fasta;

  #For Fasta files:
  my ($res,$index) = new Index::Fasta($file);
  or
  my ($res, $index) = new Index::Fasta();
  die $index unless $res;
  ($res, $msg) = $index->setFile($file);

  # If you already have a hash table with key=word-value=position in the file
  ($res, $msg) = $index->indexOut(\%inx);
  die $msg unless $res;
  #otherwise
  my $r_array = ['ac', 'name'];
  ($res, $msg) = $index->buildIndex($r_array);  #array reference containing the list of key to index. Keys are defined in $Index::regexp hash table.
  die $msg unless $res;
  or
  my $r_hash = { 'id' => '>([^\s]+)', 'name' => '>[^\s]+\s+(\w+)' };
  ($res, $msg) = $index->buildIndex($r_hash);   #hash reference with key => regexp to index. Will index id and name as described in regexp.
  die $msg unless $res;

  ($res, $msg) = $index->indexOut($msg);
  die $msg unless $res;
  my $entry;

  #Case sensitive
  ($res, $entry) = $index->getEntry('1433_OENHO');
  die $entry unless $res;

  #Case insensitive
  ($res, $entry) = $index->getEntry('1433_OENHO', 1);
  die $entry unless $res;

  print $entry->[0];

  #should print :
  # >1433_OENHO Wap_rat
  #MATAPSPREENVYLAKLAEQAERYEEMVEFMEKVCAAADSEELTVEERNLLSVAYKNVIG 

  ($res, $msg) = $index->parseFields(\$entry);   #This parse the entry using the default rules ($Index::regexp) and keep in memory a hash table
  die $msg unless $res;                          #with keys equal keys of Index::regexp and values found in entry


  #Equivalent to 

  my($id, $name);
  ($res, $id) = $index->getField('id', \$entry); #Avoid user to call pasrFields first.
  die $id unless $res;

  #Getting fields...

  ($res, $id) = $index->getField('id');
  die $id unless $res;

  ($res, $name) = $index->getField('name');
  die $name unless $res;

  #equivalent to:
  ($res, $list) = $index->getField(['id', 'name']);  #Returns id and name values form entry in the same order than array of keys ['id', 'name']
  die $list unless $res;                             #$list will be equal to ['1433_OENHO', 'Wap_rat']


  #equivalent to following two lines
  ($res, $id) = $index->get_id();         #uses AUTOLOAD function to retrieve only id from entry previously parsed
  die $id unless $res;

  ($res, $name) = $index->get_name();     #uses AUTOLOAD function to retrieve only name from entry previously parsed
  die $name unless $res;

  #Can also be used as follow

  ($res, $name) = $index->get_name(\$entry);  #Same behavior as AUTOLOAD but also parse first the "new" $entry if $entry different than the old one.
  (die $name unless $res;


=head1 VERSIONS

$Id: Index.pm,v 1.1.1.1 2005/08/18 13:18:26 hunter Exp $

Copyright (c) European Bioinformatics Institute 2004

=head1 AUTHORS / ACKNOWLEDGEMENTS

Emmanuel Quevillon <tuco@ebi.ac.uk>

=cut


package Index;

BEGIN {
  unless($ENV{IPRSCAN_HOME}){
    print STDERR "$0 : IPRSCAN_HOME environment variable not set!\n";
    exit();
  }
  unless($ENV{IPRSCAN_LIB}){
    $ENV{IPRSCAN_LIB} = "$ENV{IPRSCAN_HOME}/lib";
  }
}

use strict;
use vars qw($AUTOLOAD);
use strict;
use DB_File;
use Fcntl;
use Data::Dumper;

#This package should be to index only the results of the jobs.
#To index the datafile, see specific packages in the Index directory (e.g. Index::Pfam)


=head2 new

 Description:  Create a new object Index. User needs to call inheriting module.

 Arguments:    

 Returns:      1, $self on success
               0, msg on failure

=cut


sub new (;$) {

    my($class, $in) = @_;

    my $mess = "Method 'new' not implemented in Index.pm module.
                You need to call the sub module method (e.g. : Index::Fasta)
                unless you give a file as an argument. See documentation for
                more details.\n";

    return(0, $mess) unless($in);

    my ($res, $msg, $self);

    $self = bless({ }, $class);

    ($res, $msg) = $self->checkFile($in);
    return(0, $msg) unless $res;

    ($res, $msg) = $self->setFile($in);
    return(0, $msg) unless $res;

    return(1, $self);
}



=head2 setFile

 Description:  Set the original file to be indexed and the name of the future 
               indexed file. Those name will be used later by other subroutines.

 Arguments:    $file Path to the file.
 Returns:      1, on success
               0, msg on failure

=cut


sub setFile ($) {

    my($self, $file) = @_;

    return(0, 'setFile: No file given') unless($file);

    $self->{file} = $file;
    $self->{inx} = $file . ".inx";

    return(1, '');
}

=head2 getFile

 Description:  Return the path to the file to be indexed.
               Check in the same time if the file exists and is set.

 Arguments:    
 Returns:      1, path to the file on success
               0, msg on failure

=cut

sub getFile {

    my($self) = @_;

    return '' unless($self->{file});
    return $self->{file};

}

=head2 setRecDel

 Description:  Set the input record delimiter. You can store 2 keys (dumper and/or building). Some times we need 2 differents records delimeter
               to first build (building) the indexed file and another one to delimit the entry (dumper).

 Arguments:    $key (supported keys: dumper (when uses getEntry), and building (when creating the index file).
               $recdel
 Returns:      1, '' on success
               0, msg on failure

=cut

sub setRecDel ($$) {

    my($self, $key, $recdel) = @_;

    return(0, "You must precise a key when you set a record delimeter, supported keys : (dumper, building).") unless $key;
    return(0, "Key '$key' unsupported. Supported keys: (delimeter, position).") unless ($key eq 'dumper' || $key eq 'building');

    unless($self->{recdel} && ref($self->{recdel}) eq 'HASH'){
	$self->{recdel} = { };
    }

    $self->{recdel}->{$key} = $recdel;

    return(1, '');

}

=head2 getRecDel

 Description:  Get the input record separator associated with a key (dumper or building).

 Arguments:    $key (supported keys: dumper (when uses getEntry), and building (when creating the index file)).

 Returns:      1, $recdel on success
               0  error  if no key given or key not known.

=cut

sub getRecDel ($) {

    my($self, $key) = @_;

    return (0, "getRecDel: No key given to retrieve record delemiter (dumper|building)") unless ($key && ($key eq 'dumper' || $key eq 'building'));

    return (1, $self->{recdel}->{$key});
}


=head2 checkFile

 Description:  Check if the file to index is set and exists
 Arguments:    $file path to file 
 Returns:      1, path to the file on success
               0, msg on failure

=cut


sub checkFile (;$) {

  my($self, $file) = @_;

  if($file){
      return(0, "checkFile: File $file does not exit : $!") unless(-f $file);
      return(0, 'checkFile: File $file is not readable') unless(-r $file);
  }else{
      return(0, "checkFile: No index file was set. Set it first using setFile().") unless($self->{file});
      return(0, "checkFile: File $self->{file} does not exit : $!") unless(-f $self->{file});
      return(0, 'checkFile: File $self->{file} is not readable') unless(-r $self->{file});
  }

  return (1, $file ? $file : $self->{file});
}


=head2 indexOut

 Description:  index the input sequence by descrption field

               Ex : >wap_rat blah balh
                    rthgensvcdawwq....
                    ......
                    >another sequence
                    ...
               Will index the position of the 'wap_rat' word.

 Arguments:    $inx reference to a hash table containing key=word, value=position in the file.
               To get position in the file you can use the perl build-in 'tell' function
 Returns:      1, on success
               0, msg on failure

=cut


sub indexOut ($) {

    my ($self, $inx) = @_;

    return(0, "indexOut: Could not create index. Need a reference to a hash table") unless(ref($inx) eq "HASH");
    return(0, "indexOut: No indexfile set!") unless(-f $self->{file});

    unless($self->{inx}){
	my($res, $msg) = $self->setFile($self->{file});
	return(0, "indexOut: $msg") unless($res);
    }

    #skip the building of the index file if it already exists
    if(-f $self->{inx}){
	if($self->checkIndex($self->{file})){
	    return(1, '');
	}
    }

    $inx->{'!!!check'} = (stat($self->{file}))[7];

    tie(my %index, 'DB_File', $self->{inx}, O_RDWR | O_CREAT, 0644, $DB_File::DB_BTREE) ;

    #We need to do it like that otherwise the BTREE does not work correclty
    foreach my $key (sort keys %$inx){
	$index{$key} = $inx->{$key};
    }

    untie(%index);                      # Close database
    return (1,'');
}

=head2 buildIndex

 Description:  Read a file and construct a hash table
               reflecting the index file.

 Arguments:    $r_list  reference to a hash table (key => regexp to index) or an array containing a list of keys present in the $regexp hash table.
               $file    file to index (optional). By default the file is set in the 'new' method.

 Returns:      1, ref to a hash on success
               0, msg on failure

=cut

sub buildIndex (;$$) {

    my($self, $r_list, $file) = @_;

    my($res, $msg, $rxp);

    ($res, $rxp) = $self->getRegExp();
    return(0, "builIndex: $rxp") unless $res;

    if($file){
	($res, $msg) = $self->checkFile($file);
	return(0, "builIndex: $msg") unless($res);

	($res, $msg) = $self->setFile($file);
	return(0, "builIndex: $msg") unless $res;
    }

    my @item = ();

    my $message = join("\n- ", sort keys %$rxp);

    if($r_list && ref($r_list) eq 'HASH'){
	@item = values(%$r_list);
    }elsif($r_list && ref($r_list) eq 'ARRAY'){
	for(@$r_list){
	    return(0, "buildIndex: your key '$_' is not defined/supported. Supported keys:\n- $message\n") unless (exists($rxp->{$_}));
	    push(@item, $rxp->{$_});
	}
    }else{
	#Default behavior, indexes all the keys in the $regexp hash table.
	($res, $msg) = $self->getRegExp();
	return(0, $msg) unless $res;
	@item = grep { !/^HASH/ } values(%$msg);   #We have to filter keys, some of them are hash table (see Index::InterPro::regexp).
    }

    return(0, "buildIndex: Could not find anything in the list. Nothing to index?") unless scalar(@item);

    my ($efile, $recdel);
    $efile = $self->getFile();
    return(0, "No file set!") unless $efile;

    ($res, $recdel) = $self->getRecDel('building');
    return(0, "builIndex: $recdel") unless $res;

    ($res, $recdel) = $self->getRecDel('dumper') unless $recdel;
    return(0, "buildIndex: $recdel") unless $res;

    return(0, "buildIndex: No recdel found or set") unless $recdel;

    #We need to remove the \n sign otherwise the regular expression is not found.
    #We double escape the carriage return as we use eval with $code.
    $recdel =~ s/%newline/\\n/;
    $recdel =~ s|/|\\/|g;

    my $code = "\n";
    $code .= "unless(open(INX, \"<$efile\")){\n";
    $code .= "    close INX;\n";
    $code .= "    die \"could not open $efile : \$!\";\n";
    $code .= "}\n";
    $code .= "my \$h_inx = { };\n";
    $code .= "my \$fpos = 0;\n";
    $code .= "my \$fasta = 0;\n\n";
    $code .= "while(<INX>){\n";
#    $code .= "    chomp;\n"; #It breaks the indexing when the INPUT_RECORD_DELEMETER is '\n'

    for(@item){
	next if $_ =~ /^\s*$/;
	my $reg = $_;
	
	$code .= "    if(\$_ =~ /^$reg/){\n";
	$code .= "        \$fasta = 1;\n";
	$code .= "        my \$exp = \$1;\n";
	$code .= "        if(exists(\$h_inx->{\$exp})){\n";
	#This filtering removes duplicates for positions but it slows down the indexing.
	$code .= "             unless(grep { \/^\$fpos\$\/ } split(\/\\+\/, \$h_inx->{\$exp})){\n";
	$code .= "                \$h_inx->{\$exp} .= '+' . \$fpos;\n";
	$code .= "             }\n";
	$code .= "        }else{\n";
	$code .= "            \$h_inx->{\$exp} = \$fpos;\n";
	$code .= "       }\n";
	$code .= "    }\n";
	$code .= "\n";
    }

    $code .= "     if(\$_ =~ /$recdel/){\n";
    $code .= "         \$fpos = tell(INX);\n";
    $code .= "     }\n";
    $code .= "}\n\n";
    $code .= "close(INX);\n";
    $code .= "die \"buildIndex: I read $efile until the end and I did not find anything \" unless(\$fasta);\n";
    $code .= "return \$h_inx;\n";

    my $h_inx = eval($code);
    return(0, "buildIndex: $@") if($@);

    return(1, $h_inx);
}


=head2 checkIndex

 Description:  Check if the index file is up to date comparing the actual
               size of the original file (not indexed) and the size stored
               in the indexed file. (see indexOut upper).

 Arguments:    $file  Name of the file to check (not the indexed one) (optional).
 Returns:      1, '', 'size & modification time' on success
               0, msg, 'size & modification time' on failure

=cut

sub checkIndex (;$) {

    my($self, $file) = @_;

    unless($file){
	return(0, "checkIndex: No such indexed file " . $self->{inx}) unless(-f $self->{inx});
	return(0, "checkIndex: No such file " . $self->{file}) unless(-f $self->{file});
    }else{
	return(0, "checkIndex: Indexed file does not exist") unless(-f "$file.inx");
	my ($res, $msg) = $self->setFile($file);
	return(0, "checkIndex: Unable to set File [$file] : $msg") unless($res);
    }

    my $size  = (stat($self->{file}))[7];
    my $mtime = (stat($self->{file}))[9];

    tie(my %index, 'DB_File', $self->{inx}, O_RDONLY, 0, $DB_BTREE);

    my $indsize = $index{'!!!check'};
    untie(%index);

    return (1, "", "$size|$mtime") if ($indsize == $size);
    return (0, "checkIndex: $self->{inx} index out of date?", "$size|$mtime");
}  # end checkIndex


=head2 getIndex

 Description:  Get the position in the flat file of the given word.

 Arguments:    $word to search, 
               $insensitive if to 1, perform search using case insensitive
               $file (optional to search in a specific file). This file
                     needs to be indexed first.

 Returns:      1, [ref to array of positions | position] on success
               0, msg on failure

=cut

sub getIndex ($;$$$) {

    my($self, $word, $insensitive, $file) = @_;

    return(0, "getIndex: Nothing to search?") unless(defined($word));
    my($res, $msg);

    if($file){
	($res, $msg) = $self->checkFile($file);
	return(0, "getIndex: $msg") unless($res);

	($res, $msg) = $self->setFile($file);
	return(0, "getIndex: $msg") unless($res);
    }

    ($res, $msg) = $self->checkIndex();
    return(0, $msg) unless($res);

    my $posid = undef;

    tie(my %index, 'DB_File', $self->{inx}, O_RDONLY, 0, $DB_BTREE);

    if($insensitive){
	#OK what we do here is: We retrieve all the keys matching $word (Case insensitive), then we get the corresponding values (positions in the file)
	#in the index file (tied by %index), and we join all the position with '+'.
	#From there, we split $pos with '+' and retrieve all individual positions in the file and create an entry in the hash table
	#to be sure that this position is unique, it avoids duplicates and display twice the same entry.

	my $seen = { };
	my $pos = join('+', map { $index{$_} } grep { /^$word$/i } keys %index);
	for(split/\+/, $pos){
	    $seen->{$_} = 1 unless(exists($seen->{$_}));
	}

	$posid = join('+', keys %$seen) if(keys %$seen);

    }else{
	$posid = $index{$word} if(exists($index{$word}));
    }

    untie(%index);

    return (1, $posid);

}  # end getIndex


=head2 getEntry

 Description:  Get the indexed entry from the original file.

 Arguments:    $word, what you want to retrieve
               $insensitive Case insensitive ?
               $recdel, Input record delimiter. If you want to precise newline,
               do it as '%newline>' for a recdel like '\n>'.
               $posid, position of this word in the file (optional)
               $remove, remove the record delimeter from the retrieved entry (optional)

 Returns:      1, array ref containing entr(y/ies)
               0, msg on failure

=cut


sub getEntry ($;$$$$$) {

    my($self, $word, $insensitive, $recdel, $posid, $remove) = @_;
    my $res;

    $remove = 0 unless($remove);

    unless(defined($posid)){ #Try to get a posid first with the word
	return(0, "getEntry: No word to search given!") unless($word);
	($res, $posid) = $self->getIndex($word, $insensitive);
	return(0, $posid) unless($res);
	return(1, '') unless(defined($posid));
    }

    return(0, 'getEntry: No input record delimeter set to retrieve entry') if(!$recdel && !$self->getRecDel('dumper'));

    unless($recdel){
	($res, $recdel) = $self->getRecDel('dumper');
	return(0, "getEntry: $recdel") unless $res;
    }

    $recdel =~ s/%newline/\n/g if($recdel =~ /%newline/);
    return(0, "getEntry: No recdel found or set") unless $recdel;

    local $/ = $recdel;

    unless(open(OUT , "<$self->{file}")){
	close(OUT); 
	return(0, "getEntry: Unable to open output file [" . $self->{file} . "] : $!");
    }

    my $onerec;
    my $start = 0;
    my $entries = [ ];

    if($posid =~ /\+/){ #Multiple entries indexed

	foreach my $position (sort { $a <=> $b } split(/\+/, $posid)){
	    unless(seek(OUT, $position, 0)){
		close (OUT);
		return(0, "getEntry: Unable to set the file pointer at position $position");
	    }

	    if ( eof(OUT) )
	      {
		  close (OUT);
		  last;
	      }

	    $onerec = <OUT>;
	    $onerec =~ s/$recdel$// if($remove);
	    push(@$entries, $onerec) if($onerec);
	}
    }
    else{
	unless(seek(OUT, $posid, 0)){
	    close (OUT);
	    return(0, "getEntry: Unable to set the file pointer at position $posid");
	}
	
	if ( eof(OUT) )
	  { #EOF
	      close (OUT);
	  }

	$onerec = <OUT>;
	$onerec =~ s/$recdel$// if($remove);
	push(@$entries, $onerec) if($onerec);
    }
    close (OUT);

    if(scalar(@$entries)){
	return(1, $entries);
    }else{
	return(1, '');
    }

  } #end getEntry


=head2 setRegExp

 Description: set a hash table with key -> regular expression to look for in the file to be indexed.

 Arguments:  $hash  A reference to a hash table.
 Returns:    1, '' on success
             0, msg on errors.

=cut

sub setRegExp ($) {

    my($self, $hash) = @_;

    return(0, "setRegExp: No reference to hash table given") unless($hash and ref($hash) eq 'HASH');
    $self->{regexp} = $hash;

    return(1, '');
}


=head2 getRegExp

 Description: Return the reference to a hash table containing the key/values pair for expression to look for in the file to be indexed.
 Arguments:
 Returns:     1, reference to hash table on success.
              0, msg on error

=cut

sub getRegExp () {

    my($self) = @_;

    return(0, "getRegExp: No reference to hash table regexp, set it first.") unless($self->{regexp});

    return(1, $self->{regexp});

}

=head2 setRules

 Description: set a hash table with key -> regular expression to look for in the file to be indexed.

 Arguments:  $hash  A reference to a hash table.
 Returns:    1, '' on success
             0, msg on errors.

=cut

sub setRules ($) {

    my($self, $hash) = @_;

    my($res, $msg, $rules);
    unless($hash){
	($res,$rules) = $self->getRegExp();
	return(0, "setRules: $rules") unless $res;
	$hash = $rules;
    }
    return(0, "setRules: No reference to hash table given") unless($hash and ref($hash) eq 'HASH');
    $self->{rules} = $hash;

    return(1, '');
}


=head2 getRules

 Description: Return the reference to a hash table containing the key/values pair for expression to look for in the file to be indexed.
 Arguments:
 Returns:     1, reference to hash table on success.
              0, msg on error

=cut

sub getRules () {

    my($self) = @_;

    unless($self->{rules}){
	#Default, if no $ules defined in the Index::XX package, we return Regexp.
	return $self->getRegExp();
    }

    return(1, $self->{rules});

}

=head2 setFields

 Description: set a hash table with key -> values parsed from entry using $rules

 Arguments:  $hash  A reference to a hash table.
 Returns:    1, '' on success
             0, msg on errors.

=cut

sub setFields ($;$$) {

    my($self, $hash, $entry) = @_;

    my($res, $msg);

    return(0, "setFields: Need a reference to a hash table.") unless($hash and ref($hash) eq 'HASH');
    return(0, "setFields: Need a reference to a scalar.") unless($entry and ref($entry) eq 'SCALAR');

    $self->{parsed} = { };
    $self->{parsed}->{done} = 1;
    $self->{parsed}->{ref} = $entry;
    $self->{fields} = $hash;

    return(1, '');
}


=head2 getFields

 Description: Return the reference to a hash table containing the key/values parsed from the entry.
 Arguments:
 Returns:     1, reference to hash table on success.
              0, msg on error

=cut

sub getFields () {

    my($self) = @_;

    return(0, "getFields: No reference to hash table fields, set it first.") unless($self->{fields} && $self->{parsed});

    return(1, $self->{fields});

}

=head2 removeFields

 Description: Remove the hash table containing the parsed entry from the memory and set the flag parsed to 0.
 Argument:    
 Returns:      

=cut

sub removeFields () {

    my ($self) = @_;

    $self->{parsed}->{done} = 0;
    $self->{parsed}->{ref}  = '';
    $self->{fields} = '';

}

=head2 parseFields

 Description:  Parse and entry (returned by getEntry) for all the keys known in $regexp.
               Keep in a hash table all the results and returns also the hash table.
               e.g. results from InterPro IPR000001 entry.
               {
		'short_name' => 'Kringle',
		'name' => 'Kringle',
		'taxonomy' => [
			       'Arthropoda',
			       'Bacteria',
			       'Caenorhabditis elegans',
			       'Chordata',
			       'Eukaryota',
			       'Fruit Fly',
			       'Green Plants',
			       'Human',
			       'Metazoa',
			       'Mouse',
			       'Nematoda',
			       'Plastid Group'
			      ],
		'ac' => 'IPR000001',
		'found' => [
			    'IPR003966',
			    'IPR008293',
			    'IPR011357',
			    'IPR011358',
			    'IPR011359'
			   ],
		'meth_ac' => [
			      'PR00018',
			      'PS50070',
			      'PS00021',
			      'PF00051',
			      'PD000395',
			      'SM00130'
			     ],
		'type' => 'Domain',
		'meth_name' => [
				'KRINGLE',
				'KRINGLE_2',
				'KRINGLE_1',
				'Kringle',
				'Kringle',
				'KR'
			       ]
	       }
               When use ask for a field (using getField) is pick the informations up from
               build hash table.

 Arguments:    $entry reference to a scalar containing the entry.

 Returns:      1, Reference to hash table on success
               0, msg on error

=cut


sub parseFields ($;$$) {

    my($self, $entry, $rules) = @_;

    return(0, "parseFields: No entry given for parsing.") unless $entry;
    return(0, "parseFields: Need a reference to a scalar.") unless (ref($entry) eq 'SCALAR');

    $self->removeFields();

    my($res, $msg);

    unless($rules && ref($rules) eq 'HASH'){
	($res, $rules) = $self->getRules();
	return(0, "parseFields: $rules") unless $res;	
    }

    my ($code, $loop, $var);

    $code .= "my \$hash = { };\n";

    foreach my $rule (sort keys %$rules){

	my($REstart, $REend, $reg) = (undef, undef, undef);

	if(ref($rules->{$rule}) && ref($rules->{$rule}) eq "HASH"){
	    return(0, "No start delimeter configured for rule $rule") unless(exists($rules->{$rule}->{start}));
	    return(0, "No end delimeter configured for rule $rule") unless(exists($rules->{$rule}->{end}));
	    return(0, "No rule delimeter configured for rule $rule") unless(exists($rules->{$rule}->{rule}));
	    $REstart = $rules->{$rule}->{start};
	    $REend   = $rules->{$rule}->{end};
	    $reg     = $rules->{$rule}->{rule};
	}

	$reg = $rules->{$rule} unless($reg);
	return(0, '$rule regular expresion not defined') unless($reg);

	my $rule_start = $rule."_start";
	my $rule_end   = $rule."_end";
	$var .= "my \$$rule_start = 0;\n";
	$var .= "my \$$rule_end   = 0;\n";

	if(defined($REstart)){
	    $loop .= "    if(\$_ =~ /$REstart/o){\n";
	    $loop .= "        \$$rule_start = 1;\n";
	    $loop .= "        \$$rule_end   = 0;\n";
	    $loop .= "    }\n";
	}

	if(defined($REstart) && defined($REend)){
	    $loop .= "    if(\$_ =~ /$reg/o && \$$rule_start && !\$$rule_end){\n";
	    $loop .= "        if(\$hash->{$rule}){\n";
	    $loop .= "            if(ref(\$hash->{$rule}) eq 'ARRAY'){\n";
	    $loop .= "                push(\@{\$hash->{$rule}}, \$1);\n";
	    $loop .= "            }\n";
	    $loop .= "            else{\n";
	    $loop .= "                my \$back = \$hash->{$rule};\n";
	    $loop .= "                \$hash->{$rule} = [ ];\n";
	    $loop .= "                push(\@{\$hash->{$rule}}, \$back, \$1);\n";
	    $loop .= "            }\n";
	    $loop .= "         }else{\n";
	    $loop .= "             \$hash->{$rule} = \$1;\n";
	    $loop .= "         }\n";
	    $loop .= "    }\n";
	}else{
	    $loop .= "    if(\$_ =~ /$reg/o){\n";
	    $loop .= "        if(exists(\$hash->{$rule})){\n";
	    $loop .= "            if(ref(\$hash->{$rule}) eq 'ARRAY'){\n";
	    $loop .= "                push(\@{\$hash->{$rule}}, \$1);\n";
	    $loop .= "            }\n";
	    $loop .= "            else{\n";
	    $loop .= "                my \$back = \$hash->{$rule};\n";
	    $loop .= "                \$hash->{$rule} = [ ];\n";
	    $loop .= "                push(\@{\$hash->{$rule}}, \$back, \$1);\n";
	    $loop .= "            }\n";
	    $loop .= "        }else{\n";
	    $loop .= "           \$hash->{$rule} = \$1;\n";
	    $loop .= "        }\n";
	    $loop .= "    }\n";
	}

	if(defined($REend)){
	    $loop .= "    if(\$_ =~ /$REend/o){\n";
	    $loop .= "        \$$rule_end   = 1;\n";
	    $loop .= "        \$$rule_start = 0;\n";
	    $loop .= "    }\n";
	}
    }

    $code .= $var;
    $code .= "for(split( /\\n/, \$\$entry)){\n";
    $code .= "$loop\n";
    $code .= "}\n";
    $code .= "return \$hash;\n";

    my $r_hash = eval($code);
    return(0, "parseFields: $@") if($@);

    ($res, $msg) = $self->setFields($r_hash, $entry);
    return(0, "parseFields: $msg") unless $res;

    return (1, $r_hash);
}


=head2 isParsed

 Description: Check if the entry is already parsed. It keeps in memory the last entry parsed.
 Argument:    $entry  Scalar reference.
 Returns:     1 if parsed
              0 if not parsed.

=cut

sub isParsed (;$) {

    my($self, $entry) = @_;

    if($entry){
	if($self->{parsed}->{done} && $self->{parsed}->{ref} eq $entry){
	    return 1;
	}else{
	    return 0;
	}
    }
    return $self->{parsed}->{done};

}

=head2 getField

 Description:  Retrieve a specific field from interpro.xml indexed file
               The fields need to be defined in the regexp hash table at the
               begining of the package (regular expression).

 Arguments:    $rule, field to search. May be a simple key like 'ac' or a reference to an array containing a list
	       of keys. In this case, the returned result is a reference to an array where the results of items are
	       in the same order than the key list. If there are multiple results for a key, then the result is a
	       reference to an array.
	       e.g.:  $interpro->getField(['ac', 'name', 'contains']); will return a array ref like:

	       [ 'IPR000001',
                 'Kringle', 
                 [ 'IPR003966',
                   'IPR008293',
                   'IPR011357',
                   'IPR011358',
                   'IPR011359'
		 ]
	       ]

               $entry A reference to a scalar containing the whole entry. Avoid user to call first parseFields method,
                      then it is done on the fly. Kept in memory is the last entry succefully parsed.
                      So the user must be carefull when using this subroutine without any reference.
 Returns:      1, entry for field or reference to an array if multiple results.
               0, msg on failure

=cut

sub getField ($;$$) {

    my ($self, $rule, $entry) = @_;

    if($self->isParsed($entry)){

	if($self->{fields} && ref($self->{fields}) eq 'HASH'){

	    my ($res, $hash) = $self->getFields();
	    return(0, "getField: $hash") unless $res;
	    my $ra_fields = [ ];

	    if($rule && ref($rule) eq 'ARRAY'){
		for(@$rule){
		    push(@$ra_fields, $hash->{$_}) if(exists($hash->{$_}) && $hash->{$_} ne "");
		}
	    }elsif(ref(\$rule) eq 'SCALAR'){
		push(@$ra_fields, $hash->{$rule}) if(exists($hash->{$rule}) && $hash->{$rule} ne "");
	    }else{
		return(0, "getField: rule must be a reference to an array or a field name (e.g: 'ac')");
	    }
	    return(1, $ra_fields) if(scalar(@$ra_fields));
	    return(1, '');
	}else{
	    return(0, "getField: parsed entry not stored as a reference to a hash table.");
	}

    }elsif($entry){

	return(0, "getField: Need a reference to a scalar.") unless (ref($entry) eq 'SCALAR');

	my($res, $msg) = $self->parseFields($entry);
	return(0, "getField: $msg") unless $res;
	
	return $self->getField($rule);
	
    }else{
	return(0, "getField: You cannot use rule '$rule' alone without giving a reference to an entry, unless you parsed the entiere entry with Index::parseFields first");
    }

}

=head2 AUTOLOAD

 Description: This method is generic and allow user to retrieve any field form the entry if it is described in $rules hash table
              of the module used.
              By default, $rules is loaded as the rules to be used to parse the entry. If no $rules hash table are defined, by
              default, it load $regexp instead.

 Argument:    You have to use this like that: get_id() or get_ac().... id or ac are the keys you want to retrieve.
              If those keys are not defined in one of rules or regexp hash tables, then it throws an error.
              One additional argument may be given like get_ac(\$entries). $entries is used to be parsed if not done
              already. Check getField subroutine. This flexibility allow the used not to call parseFields subroutine and
              can give the entry to parse directly to getField subroutine.
 Returns:     1, reference to an array on success.
              0, msg on error.

=cut

sub AUTOLOAD {
    my ($self, $arg) = @_;

    $AUTOLOAD =~ /.*::get_(\w+)/ || return(0 , "No such method implemented ($AUTOLOAD). Can only use get_XXX()");
    my $attr = $1;

    my ($res, $hash) = $self->getRules();
    return(0, $hash) unless $res;

    return(0, "Key '$attr' unknown in regexp/rule hash table. Allowed keys are:\n- " . join("\n- ", sort keys %$hash) . "\n") unless(grep { /^$attr$/ } keys %$hash);

    my $ra_fields;

    ($res, $ra_fields) = $self->getField($attr, $arg);
    return(0, $ra_fields) unless $res;
    return(1, '') unless $ra_fields;

    if(ref(($ra_fields->[0])) eq 'ARRAY'){
	return(1, $ra_fields->[0]);
    }else{
	return(1, $ra_fields);
    }

    return(1, '');
}

1;
