=head1 NAME

IprMatches - IprMatches index utilities.

=head1 SYNOPSIS

use Index::IprMatches;

=head1 DESCRIPTION

  This package is used to index match.xml file.
  It indexes the crc64 attribute and the id of the proteins (see below).
  It also creates a raw file if a crc64 of one of your sequence(s)
  matches the known crc64 in the match.xml file (which contain the known
  hit proteins in InterPro).

  # examples here
  use Index::IprMatches;

  my($res, $msg, $index);
  ($res, $index) = new Index::IprMatches($file); #xml file here for iprmatches
  die $index unless $res;

  ($res, $msg) = $index->setRecDel('dumper', '</protein>');
  die($msg) unless($res);

  ($res, $r_inx) = $index->buildIndex(['ac', 'name']); #list of entries you want to index. $regexp in this module for allowing rules.
  die($r_inx) unless($res);                            #if no argument given, will build the index with all the rules describe in $regexp.

  #or you can build the index on your own key-value pairs based on regular expression. Only the first match is taken into account.
  #so if you want to index multiple things based on the same line, you need to create another key-value pair.
  ($res, $r_inx) = $index->builIndex({ 'id' => '>(\S+)', 'name' => '^\s+\w+\s+(\S+)' });
  die $r_inx unless $res;

  ($res, $mess) = $index->indexOut($r_inx);               #need a reference to hash table.
  die $mess unless $res;

  my $id = 'RS16_ECOLI';
  my ($res, $pos) = $index->getIndex($id); #return the position in the file of the match.xml entry containing this $id.
  die $pos unless $res;

  #getEntry returns a reference to an array in case of multiple entries found.
  my ($res, $entry) = $index->getEntry($id); #return the complete entry.
  die $entry unless $res;
  $entry = $entry->[0] if($entry);

  #Either (parsing once the whole entry)
  my($res, $msg) = $index->parseFields(\$entry);
  die $msg unless $res;

  my ($res, $name) = $index->getField('name'); #return the name of this entry.
  $name = $name->[0] if $name;
  my ($res, $name) = $index->getField(['ac', 'name', 'desc']); #return the ac, name and description of this entry.
  if($name){
      my $nm = $name->[0];
      my $ac = $name->[1];
      my $de = $name->[2];
  }

  .... # see below for fields you can retrieve.
  my ($res, $name) = $index->get_name(); #return the name of this entry.
  $name = $name->[0] if $name;
  #or simplier
  my ($res, $name) = $index->getField('name', \$entry); #return the name of this entry by parsing it on the fly
  $name = $name->[0] if $name;

  #Specific function for this module.
  #==================================
  my $ofh = \*STDIN; #(File handle);
  my $appl = [qw(BlastProDom Coil Seg FingerPrintScan)];
  ($res, $msg) = $index->wrtie_raw($index->parse(\$entry, $appl), $seqid, $ofh, $ipr, $go);
  die $msg  unless $res;


=head1 VERSIONS

$Id: IprMatches.pm,v 1.1.1.1 2005/08/18 13:18:26 hunter Exp $

Copyright (c) European Bioinformatics Institute 2002

=head1 AUTHORS / ACKNOWLEDGEMENTS

Emmanuel Quevillon <tuco@ebi.ac.uk>

=cut

package Index::IprMatches;

BEGIN {
  unless($ENV{IPRSCAN_HOME}){
    print STDERR "$0 : IPRSCAN_HOME environment variable not set!\n";
    exit();
  }
  unless($ENV{IPRSCAN_LIB}){
    $ENV{IPRSCAN_LIB} = "$ENV{IPRSCAN_HOME}/lib";
  }
}

use lib $ENV{IPRSCAN_LIB};
use Index;
@ISA = qw(Index);
use strict;
use English;


#We define regular expression for parsing the entry
my $regexp = {
	      seqlen    => '\s*<protein +.*?length=.([^\"\']+)',
	      crc       => '\s*<protein +.*?crc64=.([^\"\']+)',
	      id        => '\s*<protein +id=.([^\"\']+)',
	      name      => '\s*<protein +.*?name=.([^\"\']+)',
	      iprid     => '\s*<interpro +id=.([^\"\']+)',
	      iprnm     => '\s*<interpro +.*name=.([^\"\']+)',
	      iprtype   => '\s*<interpro +.type=.([^\"\']+)',
	      matchid   => '\s*<match +id=.([^\"\']+)',
	      matchnm   => '\s*<match +.*name=.([^\"\']+)',
	      matchdbnm => '\s*<match +.*dbname=.([^\"\']+)',
	      locstart  => '\s*<location +start=.([^\"\']+)',
	      locend    => '\s*<location +.*end=.([^\"\']+)',
	      locstat   => '\s*<location +.*status=.([^\"\']+)',
	      locevid   => '\s*<location +.*evidence=.([^\"\']+)',
	      locscore  => '\s*<location +.*score=.([^\"\']+)'
	     };

#Used to retrieve fields in one go. If you need to put some delimeter in the parsing, use 'rule', 'start' and 'end'...
#In the new method you'll have to use setRules() method to put this hash table into memory
my $rules = {
             seqlen     => '\s*<protein +.*?length=.([^\"\']+)',
	      crc        => '\s*<protein +.*?crc64=.([^\"\']+)',
	      id         => '\s*<protein +id=.([^\"\']+)',
             name       => '\s*<protein +.*?name=.([^\"\']+)',
             iprid      => { start => '\s*<interpro',
                             end   => '\s*<\/interpro>',
                             rule  => '\s*<interpro +id=.([^\"\']+)'
                           },
             iprnm      => { start => '\s*<interpro',
                             end   => '\s*<\/interpro>',
                             rule  => '\s*<interpro +.*name=.([^\"\']+)'
                           },
             iprtype    => { start => '\s*<interpro',
                             end   => '\s*<\/interpro>',
                             rule  => '\s*<interpro +.*type=.([^\"\']+)'
                           },
             matchid    => { start => '\s*<match',
                             end   => '\s*<\/match>',
                             rule  => '\s*<match +id=.([^\"\']+)'
                           },
             matchnm    => { start => '\s*<match',
                             end   => '\s*<\/match>',
                             rule  => '\s*<match +.*name=.([^\"\']+)'
                           },
             matchdbnm  => { start => '\s*<match',
                             end   => '\s*<\/match>',
                             rule  => '\s*<match +.*dbname=.([^\"\']+)'
                           },
             locstart   => { start => '\s*<location',
                             end   => '\s*<location.*\/>',
                             rule  => '\s*<location +start=.([^\"\']+)'
                           },
             locend     => { start => '\s*<location',
                             end   => '\s*<location.*\/>',
                             rule  => '\s*<location +.*end=.([^\"\']+)'
                           },
             locevid    => { start => '\s*<location',
                             end   => '\s*<location.*\/>',
                             rule  => '\s*<location +.*evidence=.([^\"\']+)'
                           },
             locscore   => { start => '\s*<location',
                             end   => '\s*<location.*\/>',
                             rule  => '\s*<location +.*score=.([^\"\']+)'
                           },
             locstat    => { start => '\s*<location',
                             end   => '\s*<location.*\/>',
                             rule  => '\s*<location +.*status=.([^\"\']+)'
                           }
};

=head2 new

 Description:  Create a new object Index::IprMatches.

 Arguments:    $in a file to index
               $tool Do you want to use Dispatcher::Tool to use index.conf values? (optional)

 Returns:      1, $self on success
               0, msg on failure

=cut


sub new (;$$) {

    my($class, $file, $tool) = @_;

    my $self = bless({ } , $class);

    my($res, $msg);

    if($file){
	($res, $msg) = $self->checkFile($file);
	return(0, $msg) unless $res;

	($res, $msg) = $self->setFile($file);
	return(0, $msg) unless($res);
    }

    if($tool){
	($res, $msg) = $self->_init();
	return(0, $msg) unless $res;
    }

    ($res, $msg) = $self->setRegExp($regexp);
    return(0, $msg) unless $res;

    ($res, $msg) = $self->setRules($rules);
    return(0, $msg) unless $res;

    return (1, $self);
}

=head2 _init

 Description: Initialize record delimiter, file and parse configuration file from index configuration file.
 Argument:    
 Returns:     1, '' on success
              0, msg on error

=cut

sub _init {

    require Dispatcher::Tool;

    my ($self) = @_;

    my $tool = new Dispatcher::Tool();

    $tool->setToolName('index');

    my($res, $msg) = $tool->parseConfig(1);
    return(0, $msg) unless($res);

    $self->{tool} = $tool;

    my $recdel = $tool->getConfigValue('recdel.match');
    return(0, 'No tag input record delimeter [recdel.match] configured in index.conf') unless($recdel);

    ($res, $msg) = $self->setRecDel('dumper',$recdel);
    return(0, $msg) unless($res);

    #File already setted and passed as an argument to the new method
    unless($self->getFile()){
	($res, $msg) = $self->setFile($tool->getConfigValue('file.match'));
	return(0, $msg) unless($res);
    }

    return (1, '');
}


=head2 parse

 Description:  Parse the entry retrieved from the indexed file.

 Arguments:    $text, scalar reference to the complete entry.
               $r_appl, reference to an applications array.
	       $r_taxo, reference to a taxonomy list array (optional) when user filters output with taxonomy.
 Returns:      reference of three hash tables.
               0, msg on failure

=cut

sub parse ($$;$$$) {

    my ($self, $text, $r_appl, $r_taxos) = @_;

    if(!ref($text) or $$text eq ""){
	return(0, "parse: Unable to parse XML entry...! Empty entry or not e reference to a Scalar.");
    }
    if(!ref($r_appl) or ref($r_appl) ne "ARRAY"){
	return(0, "parse: You need to give a reference to an array...!");
    }
    if($r_taxos){
	if(!ref($r_taxos) or ref($r_taxos) ne 'HASH'){
	    return(0, "parse: You need to give a reference to a hash table... !");
	}
    }

    if(!scalar(@{$r_appl})){
	return(0, "parse: Array is empty. No applications selected...!");
    }
    my %prot = ();
    my %ipr = ();
    my %matches = ();
    my %convert = (
		   GENE3D   => 'Gene3D',
		   PANTHER  => 'HMMPanther',
		   PRODOM   => 'BlastProDom',
		   SSF      => 'superfamily',
		   PFAM     => 'HMMPfam',
		   PROSITE  => 'ProfileScan',
		   PREFILE  => 'ProfileScan',
		   PRINTS   => 'FPrintScan',
		   SMART    => 'HMMSmart',
		   PIRSF    => 'HMMPIR',
		   PROFILE  => 'ProfileScan',
		   TIGRFAMs => 'HMMTigr'
		  );

    #Eample here of what it looks like
    #
    #  <protein id="O00085" name="PHYA_ASPTE" length="466" crc64="21CDCB559C96AE66">
    #    <interpro id="IPR000560" name="Histidine acid phosphatase" type="Family">
    #      <match id="PF00328" name="Acid_phosphat_A" dbname="PFAM">
    #        <location start="71" end="460" status="T" evidence="HMMPfam" score="5.1E-105" />
    #      </match>
    #      <match id="PS00616" name="HIS_ACID_PHOSPHAT_1" dbname="PROSITE">
    #        <location start="73" end="87" status="T" evidence="AddProsite" score="8.0E-5" />
    #      </match>
    #      <match id="PS00778" name="HIS_ACID_PHOSPHAT_2" dbname="PROSITE">
    #        <location start="355" end="371" status="?" evidence="AddProsite" score="0.0" />
    #      </match>
    #    </interpro>
    #  </protein>

    my $dbid = "";
    my $iprid = "";
    my $acc = "";
    my $bool = 0;

    foreach my $line (split /\n/, $$text){

	if($line =~ /\s*<protein\s+id=.([^\"]+).\s+name=.([^\"]+).\s+length=.([^\"]+).\s+crc64=.([^\"]+).+>/){
	    $prot{$1} = {id => "$1", name => "$2", length => "$3", crc64 => "$4"};
	    $acc = $1;
	}
	if($line =~ /^\s+<interpro\s+id=.([^\"]+).\s+name=.([^\"]+)./){
	    #We have to filter this InterPro id, it is not ok for the taxonomy asked by the user.
	    $ipr{$acc}{$1} = {id => "$1", name => "$2"};
	    $iprid = $1;
	}
	if($line =~ /^\s+<match\s+id=.([^\"]+).\s+name=.([^\"]+).\s+dbname=.([^\"]+).+/){
	    $bool = 0;
	    next unless(grep {/$convert{$3}/i} @{$r_appl}); #to display results for the applications asked only.
	    delete($matches{$iprid}{$dbid}) unless($dbid ne "" and exists($matches{$iprid}{$dbid}) and $matches{$iprid}{$dbid}{loc} > 0);
	    $matches{$iprid}{$1} = { id => "$1", name => "$2", dbname => "$convert{$3}", loc => []};
	    $dbid = $1;
	    $bool = 1;;
	}
	if($line =~ /^\s+<location\s+start=.([^\"]+).\s+end=.([^\"]+).\s+status=.([^\"]+).\s+evidence=.([^\"]+).\s+(score=.([^\"]+))?/){
	    next unless($bool);
	    next unless($1 != 0);
	    if($5 ne ""){
		push(@{$matches{$iprid}{$dbid}{loc}}, {start => "$1", end => "$2", status => "$3", evidence => "$4", evalue => "$6"});
	    }else{
		push(@{$matches{$iprid}{$dbid}{loc}}, {start => "$1", end => "$2", status => "$3", evidence => "$4", evalue => "0"});
	    }
	}
	if($line =~ /^\s+<location\s+status=.([^\"]+).\s+/){
	    next unless($bool);
	    push(@{$matches{$iprid}{$dbid}{loc}}, {start => "0", end => "0", status => "?", evidence => "ProfileScan"});
	}
    }

    #If user uses a taxonomy filtering
    if(keys %$r_taxos){
	
	foreach my $iprid (keys %$r_taxos){
	    delete($matches{$iprid}) unless($r_taxos->{$iprid});
	}
    }

    return (\%prot, \%ipr, \%matches);
}


=head2 write_raw

 Description:  Creates raw file for entries retieved and parsed from match.xml

 Arguments:    $r_prot, reference to hash table containing protein name as key
               $r_ipr, reference to hash table containing interpro entries as key
               $r_matches, reference to hash table containing matches entries as key
               $seq_id, ID of the sequence returned this iprmatches entry.
               $fh, a reference to a file handle to write in.
               $ipr, iprlookup asked?
               $go, go terms asked?

 Returns:      1, '' on success.
               0, msg on failure

=cut

sub write_raw ($$$$$;$$$$$$$) {

  use Index::InterPro;

  my($res, $index) = new Index::InterPro(undef, 1);
  return(0, "write_raw: $index") unless $res;

  my($self, $r_prot, $r_ipr, $r_matches, $seq_id, $fh, $ipr, $go) = @_;
  if(ref($r_prot) ne "HASH" or ref($r_ipr) ne "HASH" or ref($r_matches) ne "HASH"){
    return(0, "Can not write the raw file, wrong references passed to the function, need Hash references...\n");
  }

  my $i = 0;
  my $date = `date +"%d-%b-%Y"`;
  $date =~ s/\n//g;

  foreach my $protid (keys %{$r_prot}){

      foreach my $iprid (keys %{$$r_ipr{$protid}}){

	  foreach my $dbid (keys %{$r_matches->{$iprid}}){

	      for($i = 0; $i <= $#{$$r_matches{$iprid}{$dbid}{loc}}; $i++){

		  my $go_entry = "";
		  #Perform the GO lookup.
		  if($go){
		      my $go_class;
		      ($res, $go_entry) = $index->getEntry($$r_matches{$iprid}{$dbid}{id});
		      return(0, "write_raw: $go_entry") unless $res;

		      $go_entry = $go_entry->[0]; #get_entry returns an array ref. A database member acc/id should be unique on all the interpro database.

		      ($res, $go_entry) = $index->parseFields(\$go_entry);
		      return(0, "write_raw: $go_entry") unless $res;

		      ($res, $go_class) = $index->get_goclass();
		      return(0, "write_raw: $go_class") unless $res;

		      $go_entry = (ref($go_class) eq 'ARRAY' ? join(', ', @{$go_class}) : '');

		  }

		  my $line = sprintf("%s\t%s\t%s\t", $seq_id, $$r_prot{$protid}{crc64},$$r_prot{$protid}{length});
		  $line   .= sprintf("%s\t%s\t%s\t", $$r_matches{$iprid}{$dbid}{dbname}, $$r_matches{$iprid}{$dbid}{id}, $$r_matches{$iprid}{$dbid}{name});
		  $line   .= sprintf("%s\t%s\t%s\t", $$r_matches{$iprid}{$dbid}{loc}[$i]{start}, $$r_matches{$iprid}{$dbid}{loc}[$i]{end}, $$r_matches{$iprid}{$dbid}{loc}[$i]{evalue});
		  $line   .= sprintf("%s\t%s", $$r_matches{$iprid}{$dbid}{loc}[$i]{status}, $date);
		  $line   .= sprintf("\t%s\t%s", $iprid, $$r_ipr{$protid}{$iprid}{name}) if($ipr);
		  $line   .= sprintf("\t%s", $go_entry) if($go and $ipr);

		  print $fh "$line\n";

	      } #end for
	  } #end $dbid
      } #end $iprid
  } #end $protid

  return(1, '');
} #end write_raw()

1;
