=head1 NAME

Logger - simple logging utility.

=head1 SYNOPSIS

  # logging to default log file
  use Logger;
  my $Log = new Logger();
  $Log->debug("message");

  # logging to specific log file
  use Logger;
  my $Log = new Logger("/path/to/logger.conf");
  $Log->info("started");

=head1 DESCRIPTION

Logger is based on Log::Dispatch::Config.  Example of an configration file,
which logs to a file and stdout:

  dispatchers = file screen
  file.class = Log::Dispatch::File
  file.min_level = debug
  file.filename = /path/to/logs/dispatcher.log
  file.mode = append
  file.format = [%d] [%p] %m at %F line %L%n
  screen.class = Log::Dispatch::Screen
  screen.min_level = info
  screen.stderr = 1
  screen.format = %m

Logger caches the created Logger objects to minimize the startup time.  By
instantiating Logger objects with different configuration files, it is
possible to log information for different purposes.

=head1 VERSIONS

$Id: Logger.pm,v 1.1.1.1 2005/08/18 13:18:26 hunter Exp $

Copyright (c) European Bioinformatics Institute 2002

=head1 AUTHORS / ACKNOWLEDGEMENTS

Ville Silventoinen <vsi@ebi.ac.uk>

=cut


package Logger;

use strict;
use English;
use File::Spec::Functions;
use Log::Dispatch::Config;

use vars qw($VERSION);
$VERSION = '0.1';

# global variables
my %Cache;   # cache Logger instances


=head1 CONSTRUCTORS


=head2 new

 Description: Constructor allocates an anonymous hash, which
              is tied to the class.
 Arguments:   $conf    Path to the configuration file (optional).
 Returns:     $self object

=cut

sub new($) {
    my($class, $conf) = @_;

    unless ($conf && -r $conf) {
        print STDERR "$0: Logger could not read configuration file $conf\n";
    }

    my $self = $Cache{$conf};

    if (!$self) {
        $self = { logger => undef,
                  conf   => $conf,
                  ok     => 0 };
        $Cache{$conf} = bless $self, $class;
        $self->_init();
    }

    return $self;
}


=head1 PUBLIC METHODS


=head2 emergency

 Description: Log emergency message.
 Arguments:   $msg    Message string.
 Returns:     -

=cut

sub emergency($) {
    my ($self, $msg) = @_;
    $self->_log('emergency', $msg);
}


=head2 alert

 Description: Log alert message.
 Arguments:   $msg    Message string.
 Returns:     -

=cut

sub alert($) {
    my ($self, $msg) = @_;
    $self->_log('alert', $msg);
}


=head2 critical

 Description: Log critical message.
 Arguments:   $msg    Message string.
 Returns:     -

=cut

sub critical($) {
    my ($self, $msg) = @_;
    $self->_log('critical', $msg);
}


=head2 error

 Description: Log error message.
 Arguments:   $msg    Message string.
 Returns:     -

=cut

sub error($) {
    my ($self, $msg) = @_;
    $self->_log('error', $msg);
}


=head2 warn

 Description: Log warning message.
 Arguments:   $msg    Message string.
 Returns:     -

=cut

sub warn($) {
    my ($self, $msg) = @_;
    $self->_log('warning', $msg);
}


=head2 notice

 Description: Log notice message.
 Arguments:   $msg    Message string.
 Returns:     -

=cut

sub notice($) {
    my ($self, $msg) = @_;
    $self->_log('notice', $msg);
}


=head2 info

 Description: Log info message.
 Arguments:   $msg    Message string.
 Returns:     -

=cut

sub info($) {
    my ($self, $msg) = @_;
    $self->_log('info', $msg);
}


=head2 debug

 Description: Log debug message.
 Arguments:   $msg    Message string.
 Returns:     -

=cut

sub debug($) {
    my ($self, $msg) = @_;
    $self->_log('debug', $msg);
}


=head2 logJob

 Description: Convenience method to log job information (info level).
 Arguments:   $jobid      Job identifier.
              $program    Program name, e.g., fastx.
              $search     interactive|email
              $db         Database name (optional).
              $email      Email address (optional).
              $host       Client hostname or IP address (optional).
 Returns:     -

=cut

sub logJob($$$;$$$) {
    my ($self, $jobid, $program, $search, $db, $email, $host) = @_;
    return unless $program && $search;

    $db    = '' unless $db;
    $email = '' unless $email;
    $host  = $host || $ENV{'REMOTE_HOST'} || $ENV{'REMOTE_ADDR'} || 'UNKNOWN';
    $self->_log('info', "$host $program $db $search $email $jobid");
}


=head1 PRIVATE METHODS


=head2 _log

 Description: Logs message with given level.
 Arguments:   $level    Log level.
              $msg      Message string (optional).
 Returns:     -

=cut

sub _log($;$) {
    my($self, $level, $msg) = @_;
    return unless $msg;
    return unless $self->{ok};

    $level = 'info' unless $level;  # we have to use some level
    local $Log::Dispatch::Config::CallerDepth = 2;
    $self->{logger}->$level($msg);
}


=head2 _init

 Description: Initializes the object.
 Arguments:   -
 Returns:     -

=cut

sub _init () {
    my $self = shift;

    Log::Dispatch::Config->configure($self->{conf});
    $self->{logger} = Log::Dispatch::Config->instance;

    # We can now start logging with this Logger.
    $self->{ok} = 1;
}


1;
__END__
